/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL )

/* The delayed relocator.  If the original attempt to relocate leaves us
 * in a silly place (some BIOSes do not initialise the INT 15 memory map
 * until after BCV time, where we need to hook INT 13), we can hook this
 * in to INT 13 / 15 / 19, and will try again on the first call to one of
 * these interrupts.
 */

#include <librm.h>
#include <config/general.h>

	.text
	.arch i386

#define SIZEOF_I386_REGS	32

/****************************************************************************
 * Delayed relocation common interrupt handler code
 *
 * This is called from the first INT 13 / 15 / 19 call after the hooks have
 * been installed.  We unhook ourselves from all three interrupts, and try
 * find a better home for gPXE by calling relocate() again.
 *
 * (If this doesn't put us at the end of memory either, we don't try again;
 * there's probably not much point.)
 *
 * Parameters:
 *  None
 * Returns:
 *  None
 * Corrupts:
 *  None
 ****************************************************************************
 */
	.section ".text16", "ax", @progbits
	.code16

	.globl	delayed_reloc_common
delayed_reloc_common:
	pushal
	pushw	%ds
	pushw	%es

	movw	%cs, %ax
	movw	%cs:rm_ds, %bx
	movw	%bx, %ds

	/* Uninstall the delayed relocation hooks */
	xorw	%dx, %dx
	movw	%dx, %es
	pushl	%cs:old_int15
	popl	%es:( 0x15 * 4 )
	pushl	%cs:old_int19
	popl	%es:( 0x19 * 4 )

	/* Try to relocate again */
	pushl	$relocate
	pushw	%cs
	call	prot_call
	popl	%edx /* discard */

	/* Copy code to new location */
	pushw	%ds
	pushal
	movl	$SIZEOF_I386_REGS, %ecx
	pushl	$1f
	jmp	real_to_prot
	.section ".text", "ax", @progbits
	.code32
1:	popal
	pushal
	movw	$PHYSICAL_DS, %ax
	movw	%ax, %ds
	movw	%ax, %es
	rep movsb
	movw	$VIRTUAL_DS, %ax
	movw	%ax, %ds
	movw	%ax, %es
	movl	$SIZEOF_I386_REGS, %ecx
	pushl	$2f
	jmp	prot_to_real
	.section ".text16", "ax", @progbits
	.code16
2:	popal
	popw	%ds

	/* Initialise librm at new location */
	pushw	%cs
	call	init_librm

	/* Let C code perform any required post-relocation actions */
	pushl	$post_delayed_reloc_hook
	pushw	%cs
	call	prot_call
	popl	%edx /* discard */

	/* Restore registers */
	popw	%es
	popw	%ds
	popal

	ret
	.size delayed_reloc_common, . - delayed_reloc_common

/****************************************************************************
 * Hooked interrupt handlers
 *
 * NOTE: we must re-read the interrupt vector to find out where to chain to.
 * The call to delayed_reloc_common has unhooked our handlers, but the call
 * to hide_post_delayed_reloc() may have hooked new handlers.
 *
 ****************************************************************************
 */
#define SMAP 0x534d4150

#ifdef EBDA_ALLOCATE
	.section ".prefix.lib", "awx", @progbits
delayed_reloc_int15_prefix:
	pushfw
	pushw	%cs
	pushw	$delayed_reloc_int15
	call	find_and_jump_text16
	
	.size delayed_reloc_int15_prefix, . - delayed_reloc_int15_prefix
	.previous
#endif
delayed_reloc_int15:
	pushfw
	cmpb	$0x88, %ah
	je	1f
	cmpw	$0xe801, %ax
	je	1f
	cmpw	$0xe820, %ax
	jne	2f
	cmpl	$SMAP, %edx
	jne	2f
1:	call	delayed_reloc_common
	pushw	%es
	pushw	%ax
	xorw	%ax, %ax
	movw	%ax, %es
	pushl	%es:( 0x15 * 4 )
	popl	%cs:old_int15
	popw	%ax
	popw	%es
2:	popfw
	ljmp	*%cs:old_int15
	.size delayed_reloc_int15, . - delayed_reloc_int15

#ifdef EBDA_ALLOCATE
	.section ".prefix.lib", "awx", @progbits
delayed_reloc_int19_prefix:
	pushfw
	pushw	%cs
	pushw	$delayed_reloc_int19
	call	find_and_jump_text16
	
	.size delayed_reloc_int19_prefix, . - delayed_reloc_int19_prefix
	.previous
#endif
delayed_reloc_int19:
	pushfw
	call	delayed_reloc_common
	pushw	%es
	pushw	%ax
	xorw	%ax, %ax
	movw	%ax, %es
	pushl	%es:( 0x19 * 4 )
	popl	%cs:old_int19
	popw	%ax
	popw	%es
	popfw
	ljmp	*%cs:old_int19
	.size delayed_reloc_int19, . - delayed_reloc_int19

/****************************************************************************
 * Install the delayed relocation interrupt hooks
 *
 * INT 15 should *not* be hooked before calling this routine; it should be
 * hooked by hide_post_delayed_reloc().
 *
 * INT 13 and/or INT 19 may already be hooked before calling this routine;
 * the correct chaining behaviour will happen automatically.
 *
 * Parameters:
 *  None
 * Returns:
 *  None
 * Corrupts:
 *  None
 ****************************************************************************
 */
	.section ".text16", "ax", @progbits
	.code16
	.globl	install_delayed_reloc
install_delayed_reloc:
	pushw	%es
	pushw	%ax
	xorw	%ax, %ax
	movw	%ax, %es

	pushl	%es:( 0x15 * 4 )
	popl	%cs:old_int15
#ifdef EBDA_ALLOCATE
	pushw	%cs:prefix_seg
	pushw	$delayed_reloc_int15_prefix
#else
	pushw	%cs
	pushw	$delayed_reloc_int15
#endif
	popl	%es:( 0x15 * 4 )

	pushl	%es:( 0x19 * 4 )
	popl	%cs:old_int19
#ifdef EBDA_ALLOCATE
	pushw	%cs:prefix_seg
	pushw	$delayed_reloc_int19_prefix
#else
	pushw	%cs
	pushw	$delayed_reloc_int19
#endif
	popl	%es:( 0x19 *4 )

	popw	%ax
	popw	%es
	ret
	.size install_delayed_reloc, . - install_delayed_reloc

	.section ".text16.data", "aw", @progbits
old_int15:	.long 0
old_int19:	.long 0

