#ifndef _GPXE_IBFT_H
#define _GPXE_IBFT_H

/*
 * SPDX-License-Identifier: BSD-2-Clause
 *
 * Copyright Fen Systems Ltd. 2007.  Portions of this code are derived
 * from IBM Corporation Sample Programs.  Copyright IBM Corporation
 * 2004, 2007.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

FILE_LICENCE ( BSD2 );

/** @file
 *
 * iSCSI boot firmware table
 *
 * The information in this file is derived from the document "iSCSI
 * Boot Firmware Table (iBFT)" as published by IBM at
 *
 * ftp://ftp.software.ibm.com/systems/support/system_x_pdf/ibm_iscsi_boot_firmware_table_v1.02.pdf
 *
 */

#include <stdint.h>
#include <gpxe/acpi.h>
#include <gpxe/in.h>
#include <gpxe/resolv.h>

/** iSCSI Boot Firmware Table signature */
#define IBFT_SIG "iBFT"

/** An offset from the start of the iBFT */
typedef uint16_t ibft_off_t;

/** Length of a string within the iBFT (excluding terminating NUL) */
typedef uint16_t ibft_size_t;

/** A string within the iBFT */
struct ibft_string {
	/** Length of string */
	ibft_size_t length;
	/** Offset to string */
	ibft_off_t offset;
} __attribute__ (( packed ));

/** An IP address within the iBFT */
struct ibft_ipaddr {
	/** Reserved; must be zero */
	uint16_t zeroes[5];
	/** Must be 0xffff if IPv4 address is present, otherwise zero */
	uint16_t ones;
	/** The IPv4 address, or zero if not present */
	struct in_addr in;
} __attribute__ (( packed ));

/**
 * iBFT structure header
 *
 * This structure is common to several sections within the iBFT.
 */
struct ibft_header {
	/** Structure ID
	 *
	 * This is an IBFT_STRUCTURE_ID_XXX constant
	 */
	uint8_t structure_id;
	/** Version (always 1) */
	uint8_t version;
	/** Length, including this header */
	uint16_t length;
	/** Index 
	 *
	 * This is the number of the NIC or Target, when applicable.
	 */
	uint8_t index;
	/** Flags */
	uint8_t flags;
} __attribute__ (( packed ));

/**
 * iBFT Control structure
 *
 */
struct ibft_control {
	/** Common header */
	struct ibft_header header;
	/** Extensions */
	uint16_t extensions;
	/** Offset to Initiator structure */
	ibft_off_t initiator;
	struct {
		/** Offset to NIC structure for NIC n */
		ibft_off_t nic;
		/** Offset to Target structure for target 0 */
		ibft_off_t target;
	} entry[2];
} __attribute__ (( packed ));

/** Structure ID for Control section */
#define IBFT_STRUCTURE_ID_CONTROL 0x01

/** Attempt login only to specified target
 *
 * If this flag is not set, all targets will be logged in to.
 */
#define IBFT_FL_CONTROL_SINGLE_LOGIN_ONLY 0x01

/**
 * iBFT Initiator structure
 *
 */
struct ibft_initiator {
	/** Common header */
	struct ibft_header header;
	/** iSNS server */
	struct ibft_ipaddr isns_server;
	/** SLP server */
	struct ibft_ipaddr slp_server;
	/** Primary and secondary Radius servers */
	struct ibft_ipaddr radius[2];
	/** Initiator name */
	struct ibft_string initiator_name;
} __attribute__ (( packed ));

/** Structure ID for Initiator section */
#define IBFT_STRUCTURE_ID_INITIATOR 0x02

/** Initiator block valid */
#define IBFT_FL_INITIATOR_BLOCK_VALID 0x01

/** Initiator firmware boot selected */
#define IBFT_FL_INITIATOR_FIRMWARE_BOOT_SELECTED 0x02

/**
 * iBFT NIC structure
 *
 */
struct ibft_nic {
	/** Common header */
	struct ibft_header header;
	/** IP address */
	struct ibft_ipaddr ip_address;
	/** Subnet mask
	 *
	 * This is the length of the subnet mask in bits (e.g. /24).
	 */
	uint8_t subnet_mask_prefix;
	/** Origin */
	uint8_t origin;
	/** Default gateway */
	struct ibft_ipaddr gateway;
	/** Primary and secondary DNS servers */
	struct ibft_ipaddr dns[2];
	/** DHCP server */
	struct ibft_ipaddr dhcp;
	/** VLAN tag */
	uint16_t vlan;
	/** MAC address */
	uint8_t mac_address[6];
	/** PCI bus:dev:fn */
	uint16_t pci_bus_dev_func;
	/** Hostname */
	struct ibft_string hostname;
} __attribute__ (( packed ));

/** Structure ID for NIC section */
#define IBFT_STRUCTURE_ID_NIC 0x03

/** NIC block valid */
#define IBFT_FL_NIC_BLOCK_VALID 0x01

/** NIC firmware boot selected */
#define IBFT_FL_NIC_FIRMWARE_BOOT_SELECTED 0x02

/** NIC global / link local */
#define IBFT_FL_NIC_GLOBAL 0x04

/** Origin enumeration */
enum ibft_prefix_origin {
	IBFT_PREFIX_ORIGIN_OTHER = 0,
	IBFT_PREFIX_ORIGIN_MANUAL = 1,
	IBFT_PREFIX_ORIGIN_WELL_KNOWN = 2,
	IBFT_PREFIX_ORIGIN_DHCP = 3,
	IBFT_PREFIX_ORIGIN_ROUTER_ADVERTISEMENT = 4,
	IBFT_PREFIX_ORIGIN_UNCHANGED = 16,
};

/**
 * iBFT Target structure
 *
 */
struct ibft_target {
	/** Common header */
	struct ibft_header header;
	/** IP address */
	struct ibft_ipaddr ip_address;
	/** TCP port */
	uint16_t socket;
	/** Boot LUN */
	uint64_t boot_lun;
	/** CHAP type
	 *
	 * This is an IBFT_CHAP_XXX constant.
	 */
	uint8_t chap_type;
	/** NIC association */
	uint8_t nic_association;
	/** Target name */
	struct ibft_string target_name;
	/** CHAP name */
	struct ibft_string chap_name;
	/** CHAP secret */
	struct ibft_string chap_secret;
	/** Reverse CHAP name */
	struct ibft_string reverse_chap_name;
	/** Reverse CHAP secret */
	struct ibft_string reverse_chap_secret;
	/** Name resolution interface */
	struct interface resolv;
} __attribute__ (( packed ));

/** Structure ID for Target section */
#define IBFT_STRUCTURE_ID_TARGET 0x04

/** Target block valid */
#define IBFT_FL_TARGET_BLOCK_VALID 0x01

/** Target firmware boot selected */
#define IBFT_FL_TARGET_FIRMWARE_BOOT_SELECTED 0x02

/** Target use Radius CHAP */
#define IBFT_FL_TARGET_USE_CHAP 0x04

/** Target use Radius rCHAP */
#define IBFT_FL_TARGET_USE_RCHAP 0x08

/* Values for chap_type */
#define IBFT_CHAP_NONE		0	/**< No CHAP authentication */
#define IBFT_CHAP_ONE_WAY	1	/**< One-way CHAP */
#define IBFT_CHAP_MUTUAL	2	/**< Mutual CHAP */

/**
 * iSCSI Boot Firmware Table (iBFT)
 */
struct ibft_table {
	/** ACPI header */
	struct acpi_description_header acpi;
	/** Reserved */
	uint8_t reserved[12];
	/** Control structure */
	struct ibft_control control;
} __attribute__ (( packed ));

/**
 * iSCSI string block descriptor
 *
 * This is an internal structure that we use to keep track of the
 * allocation of string data.
 */
struct ibft_string_block {
	/** The iBFT containing these strings */
	struct ibft_table *table;
	/** Offset of first free byte within iBFT */
	unsigned int offset;
};

/* Total size of nics, targets, and string table */
#define IBFT_PAD_SIZE 0x4d0

/**
 * An iBFT created by gPXE
 *
 */
struct gpxe_ibft {
	/** The fixed section */
	struct ibft_table table;
	/* Additional room for the entry[] array to grow into. Starts at
	 * 0x42, and should align to a page. */
	char entry_pad[0x1c]; /* additional 7, total of 9 adapters */
	/** The Initiator section */
	struct ibft_initiator initiator __attribute__ (( aligned ( 16 ) ));
	/** Remaining space for intiator, nics and targets */
	char pad[IBFT_PAD_SIZE] __attribute__ (( aligned ( 16 ) ));
} __attribute__ (( packed, aligned ( 16 ) ));

struct net_device;
struct iscsi_session;

extern int ibft_fill_iscsi ( struct iscsi_session *iscsi );

extern int ibft_fill_netdev ( struct net_device *netdev );

/* Invalidate ibftab, and restore any pre-existing iBFT */
extern void ibft_clear ( void );

#endif /* _GPXE_IBFT_H */
