/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Copyright Fen Systems Ltd. 2007.  Portions of this code are derived
 * from IBM Corporation Sample Programs.  Copyright IBM Corporation
 * 2004, 2007.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <byteswap.h>
#include <realmode.h>
#include <gpxe/pci.h>
#include <gpxe/acpi.h>
#include <gpxe/in.h>
#include <gpxe/netdevice.h>
#include <gpxe/dhcp.h>
#include <gpxe/iscsi.h>
#include <gpxe/ibft.h>
//#include <gpxe/resolv.h>
#include <gpxe/sf_settings.h>

/** @file
 *
 * iSCSI boot firmware table
 *
 * The information in this file is derived from the document "iSCSI
 * Boot Firmware Table (iBFT)" as published by IBM at
 *
 * ftp://ftp.software.ibm.com/systems/support/system_x_pdf/ibm_iscsi_boot_firmware_table_v1.02.pdf
 *
 */

/** The iBFT used by gPXE */
struct gpxe_ibft __data16 ( ibftab ) = {
	/* Table header */
	.table = {
		/* ACPI header */
		.acpi = {
			.signature = IBFT_SIG,
			.length = sizeof ( struct gpxe_ibft ),
			.revision = 1,
			.oem_id = "FENSYS",
			.oem_table_id = "gPXE",
		},
		/* Control block */
		.control = {
			.header = {
				.structure_id = IBFT_STRUCTURE_ID_CONTROL,
				.version = 1,
				.length = sizeof ( struct ibft_control ),
				.flags = 0,
			},
		},
	},
	/* iSCSI initiator information */
	.initiator = {
		.header = {
			.structure_id = IBFT_STRUCTURE_ID_INITIATOR,
			.version = 1,
			.length = sizeof ( struct ibft_initiator ),
			.flags = ( IBFT_FL_INITIATOR_BLOCK_VALID |
				   IBFT_FL_INITIATOR_FIRMWARE_BOOT_SELECTED ),
		},
	},
};

#define ibftab __use_data16 ( ibftab )
extern struct gpxe_ibft __data16 ( ibftab );

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/**
 * Fill in an IP address field within iBFT
 *
 * @v ipaddr		IP address field
 * @v in		IPv4 address
 */
static void
ibft_set_ipaddr ( struct ibft_ipaddr *ipaddr, struct in_addr in )
{
	memset ( ipaddr, 0, sizeof ( ipaddr ) );
	if ( in.s_addr ) {
		ipaddr->in = in;
		ipaddr->ones = 0xffff;
	}
}

/**
 * Fill in an IP address within iBFT from configuration setting
 *
 * If the setting doesn't exist, then the address is set to zero
 *
 * @v ipaddr		IP address field
 * @v settings		Configuration settings
 * @v setting		Configuration setting
 */
static void
ibft_set_ipaddr_option ( struct ibft_ipaddr *ipaddr,
			 struct settings *settings,
			 struct setting *setting )
{
	struct in_addr in = { 0 };

	( void ) fetch_ipv4_setting ( settings, setting, &in );

	ibft_set_ipaddr ( ipaddr, in );
}

/**
 * Allocate a string within iBFT
 *
 * @v strings		iBFT string block descriptor
 * @v string		String field to fill in
 * @v len		Length of string to allocate (excluding NUL)
 * @ret rc		Return status code
 */
static int ibft_alloc_string ( struct ibft_string_block *strings,
			struct ibft_string *string, size_t len ) {
	char *dest __unused;
	unsigned int remaining;

	dest = ( ( ( char * ) strings->table ) + strings->offset );
	remaining = ( strings->table->acpi.length - strings->offset );
	if ( len >= remaining )
		return -ENOSPC;

	string->offset = strings->offset;
	string->length = len;
	strings->offset += ( len + 1 );
	return 0;
}

/**
 * Fill in a string field within iBFT
 *
 * @v strings		iBFT string block descriptor
 * @v string		String field
 * @v data		String to fill in, or NULL
 * @ret rc		Return status code
 */
static int
ibft_set_string ( struct ibft_string_block *strings,
		  struct ibft_string *string,
		  const char *data, unsigned int length )
{
	char *dest;
	int rc;

	if ( ! data )
		return 0;

	if ( ( rc = ibft_alloc_string ( strings, string, length ) ) != 0 )
		return rc;
	dest = ( ( ( char * ) strings->table ) + string->offset );
	strncpy ( dest, data, length );

	return 0;
}

/**
 * Fill in a string field within iBFT from configuration setting
 *
 * If the string doesn't exist, then this function returns -ENOENT.
 * NB: This differs from upstream gPXE (at time of writing anyway)
 *
 * @v strings		iBFT string block descriptor
 * @v string		String field
 * @v setting		Configuration setting
 * @v settings		Settings block to use
 * @ret rc		Return error code
 */
static int
ibft_set_string_option ( struct ibft_string_block *strings,
			 struct ibft_string *string,
			 struct settings *settings,
			 struct setting *setting )
{
	char *dest = NULL;
	int len;
	int rc;

	len = fetch_setting_len ( settings, setting );
	if ( len < 0 )
		return -ENOENT;

	if ( ( rc = ibft_alloc_string ( strings, string, len ) ) != 0 )
		return rc;

	dest = ( ( ( char * ) strings->table ) + string->offset );
	fetch_string_setting ( settings, setting, dest, ( len + 1 ) );

	return 0;
}

/**
 * Fill in NIC portion of iBFT
 *
 * @v nic		NIC portion of iBFT
 * @v strings		iBFT string block descriptor
 * @v netdev		Network device
 * @ret rc		Return status code
 */
static int
ibft_fill_nic ( struct ibft_nic *nic, struct ibft_string_block *strings,
		struct net_device *netdev )
{
	struct in_addr netmask_addr = { 0 };
	unsigned int netmask_count = 0;
	struct settings *settings;
	char *hostname;
	int rc;

	settings = netdev_nvo_settings ( netdev );
	if ( settings == NULL )
		return -ENOENT;

	if ( sf_sett_get_intz ( settings, &initiator_dhcp_setting ) ) {
		nic->origin = IBFT_PREFIX_ORIGIN_DHCP;

		settings = netdev_dhcp_settings ( netdev );
		if ( settings == NULL ) {
			DBG ( "%s failed: %s has no dhcp settings\n",
			      __func__, netdev->name );
			return -ENOENT;
		}
	} else
		nic->origin = IBFT_PREFIX_ORIGIN_MANUAL;

	/* Extract values from DHCP configuration */
	ibft_set_ipaddr_option ( &nic->ip_address, settings, &ip_setting );
	ibft_set_ipaddr_option ( &nic->gateway, settings, &gateway_setting );
	ibft_set_ipaddr_option ( &nic->dns[0], settings, &dns_setting );
	/* dhcp_server will be wiped if we're using NVO, which is what we want */
	ibft_set_ipaddr_option ( &nic->dhcp, settings, &dhcp_server_setting);

	hostname = iscsi_initiator_hostname ( netdev );
	rc = ibft_set_string ( strings, &nic->hostname,
			       hostname, strlen ( hostname ) );
	free ( hostname );
	if ( rc != 0 )
		return rc;

	/* Derive subnet mask prefix from subnet mask */
	fetch_ipv4_setting ( settings, &netmask_setting, &netmask_addr );
	while ( netmask_addr.s_addr ) {
		if ( netmask_addr.s_addr & 0x1 )
			netmask_count++;
		netmask_addr.s_addr >>= 1;
	}
	nic->subnet_mask_prefix = netmask_count;

	/* Extract values from net-device configuration */
	memcpy ( nic->mac_address, netdev->ll_addr,
		 sizeof ( nic->mac_address ) );
	nic->pci_bus_dev_func = netdev->dev->desc.location;

	return 0;
}

/**
 * Fill in Initiator portion of iBFT
 *
 * @v initiator		Initiator portion of iBFT
 * @v strings		iBFT string block descriptor
 * @ret rc		Return status code
 */
static int
ibft_fill_initiator ( struct ibft_initiator *initiator,
		      struct ibft_string_block *strings,
		      struct net_device *netdev )
{
	char *initiator_iqn = iscsi_initiator_iqn ( netdev );
	int rc;

	rc = ibft_set_string ( strings, &initiator->initiator_name,
			       initiator_iqn, strlen ( initiator_iqn ) );
	free ( initiator_iqn );

	return rc;
}

/**
 * Fill in Target CHAP portion of iBFT from NVO
 *
 * This function uses the NVO context so it can be called both
 * from the iscsi login path and the dhcp only path.
 *
 * @v target		Target portion of iBFT
 * @v strings		iBFT string block descriptor
 * @v settings		NVO settings structure to use
 * @ret rc		Return status code
 */
static int
ibft_fill_target_chap ( struct ibft_target *target,
			struct ibft_string_block *strings,
			struct settings *settings )
{
	int rc;

	if ( sf_sett_get_intz ( settings, &chap_setting ) ) {
		if ( sf_sett_get_intz ( settings, &chap_target_setting ) )
			target->chap_type = IBFT_CHAP_MUTUAL;
		else
			target->chap_type = IBFT_CHAP_ONE_WAY;
	} else
		target->chap_type = IBFT_CHAP_NONE;

	if ( target->chap_type == IBFT_CHAP_NONE )
		return 0;

	rc = ibft_set_string_option ( strings, &target->chap_name,
				      settings, &username_setting );
	if ( rc < 0 )
		return rc;

	rc = ibft_set_string_option ( strings, &target->chap_secret,
				      settings, &password_setting );
	if ( rc < 0 )
		return rc;

	if ( target->chap_type != IBFT_CHAP_MUTUAL )
		return 0;

	rc = ibft_set_string_option ( strings, &target->reverse_chap_name,
				      settings, &reverse_username_setting );
	if ( rc < 0 )
		return rc;

	rc = ibft_set_string_option ( strings, &target->reverse_chap_secret,
				      settings, &reverse_password_setting );
	if ( rc < 0 )
		return rc;

	return 0;
}

/* Populate the iBFT target structure using iSCSI login details */
static int
ibft_fill_iscsi_target ( struct ibft_target *target, void *priv,
			 struct ibft_string_block *strings )
{
	struct iscsi_session *iscsi = priv;
	struct sockaddr_in *sin_target =
		( struct sockaddr_in * ) &iscsi->target_sockaddr;

	ibft_set_ipaddr ( &target->ip_address, sin_target->sin_addr );
	target->socket = ntohs ( sin_target->sin_port );
	/* iscsi->lun is a 64-bit host-endian representation of the SCSI SAM2
	 * LUN format.; a single level LUN of 0-255 (which is all we support)
	 * is 00,nn,00,00,00,00,00,00 */
	target->boot_lun = (iscsi->lun & 0xff00) >> 8;
	return ibft_set_string ( strings, &target->target_name,
				 iscsi->target_iqn, strlen ( iscsi->target_iqn ) );
}

/** Target address [it is global for ibft_resolv_done()] */
static struct sockaddr target_sockaddr = { 0, "" };
#if 0
static int target_resolv_rc;

static void ibft_resolv_done ( struct resolv_interface *resolv __unused,
			       struct sockaddr *sa, int rc )
{
	target_resolv_rc = rc;

	if (rc == 0)
		memcpy( &target_sockaddr, sa, sizeof ( *sa ) );
}

/** iBFT resolution interface operations */
static struct resolv_interface_operations ibft_resolv_ops = {
	.done = ibft_resolv_done,
};

/** iBFT resolution interface */
static struct resolv_interface ibft_resolv = {
	.intf = {
		.dest = &null_resolv.intf,
		.refcnt = NULL,
	},
	.op = &ibft_resolv_ops,
};
#endif

/* Populate the iBFT target structure using netdev NVO details */
static int
ibft_fill_netdev_target ( struct ibft_target *target, void *priv,
			  struct ibft_string_block *strings )
{
	struct net_device *netdev = priv;
	struct sockaddr_in *sin_target;
	struct settings *settings;
	char *target_iqn = NULL, *server_name = NULL;
	unsigned int target_port;
	char root_path[256];
	uint64_t lun;
	int rc;

	settings = netdev_nvo_settings ( netdev );
	if ( settings == NULL )
		return -ENOENT;

	if ( sf_sett_get_intz ( settings, &initiator_dhcp_setting ) ) {
		settings = netdev_dhcp_settings ( netdev );
		if ( settings == NULL ) {
			DBG ( "%s failed: %s has no dhcp settings\n",
			      __func__, netdev->name );
			return -ENOENT;
		}
	}

	rc = sf_root_path_fetch ( netdev, root_path, sizeof ( root_path ) );
	if ( rc <= 0 ) {
		DBG ( "%s failed: %s has no root_path setting: %s\n",
		      __func__, netdev->name, strerror ( rc ) );
		return -ENOENT;
	}

	rc = iscsi_parse_root_path ( root_path, &server_name,
				     &target_port, &lun, &target_iqn );
	if ( rc < 0 )
		goto out1;

#if 0
	/* Resolve the server_name */
	target_resolv_rc = -EINPROGRESS;
	rc = resolv ( &ibft_resolv, server_name, &target_sockaddr );
	if ( rc != 0 )
		goto out2;

	while ( target_resolv_rc == -EINPROGRESS )
		step ( );
	if ( target_resolv_rc != 0 )
		goto out3;
#endif

	sin_target = (struct sockaddr_in *)&target_sockaddr;
	ibft_set_ipaddr ( &target->ip_address, sin_target->sin_addr );
	target->socket = target_port;
	target->boot_lun = lun;
	rc = ibft_set_string ( strings, &target->target_name,
			       target_iqn, strlen ( target_iqn ) );

#if 0
out3:
out2:
#endif
	free ( server_name );
	free ( target_iqn );
out1:
	return rc;
}

/** Arrays of offsets (from the structure start) to strings within each structure */
static const off_t ibft_control_string_tbl[] = {
	0,
};

static const off_t ibft_initiator_string_tbl[] = {
	offsetof ( struct ibft_initiator, initiator_name ),
	0,
};

static const off_t ibft_nic_string_tbl[] = {
	 offsetof ( struct ibft_nic, hostname ),
	 0,
};

static const off_t ibft_target_string_tbl[] = {
	offsetof ( struct ibft_target, target_name ),
	offsetof ( struct ibft_target, chap_name ),
	offsetof ( struct ibft_target, chap_secret ),
	offsetof ( struct ibft_target, reverse_chap_name ),
	offsetof ( struct ibft_target, reverse_chap_secret ),
	0,
};

/** Table of information about each structure type */
struct ibft_table_info {
	const off_t *string_tbl;
	size_t min_size;
};

static const struct ibft_table_info ibft_struct_tbl[] = {
	[ IBFT_STRUCTURE_ID_CONTROL ] = {
		.string_tbl = ibft_control_string_tbl,
		.min_size = sizeof ( struct ibft_control ),
	},
	[ IBFT_STRUCTURE_ID_INITIATOR ] = {
		.string_tbl = ibft_initiator_string_tbl,
		.min_size = sizeof ( struct ibft_initiator ),
	},
	[ IBFT_STRUCTURE_ID_NIC ] = {
		.string_tbl = ibft_nic_string_tbl,
		.min_size = sizeof ( struct ibft_nic ),
	},
	[ IBFT_STRUCTURE_ID_TARGET ] = {
		.string_tbl = ibft_target_string_tbl,
		.min_size = sizeof ( struct ibft_target ),
	},
};

#define for_each_string_location( _loc, _id )				\
	for ( ( _loc ) = ibft_struct_tbl [ ( _id ) ].string_tbl ;	\
	      *( _loc ) != 0 ;						\
	      ++( _loc ) )						\

/** Structure size alignment */
#define IBFTAB_STRUCT_ALIGN( _s )					\
	( ( ( _s ) + 0xf ) & ~0xf )
#define IBFTAB_STRUCT_SIZE( _t )					\
	IBFTAB_STRUCT_ALIGN ( sizeof ( _t ) )

/** Initial used size of ibftab */
#define INITIAL_IBFTAB_SIZE						\
	IBFTAB_STRUCT_ALIGN ( offsetof ( struct gpxe_ibft, pad ) )

/** Offset into ibftab for new structures inserted into ibftab */
static unsigned int ibftab_next_structure_offset = INITIAL_IBFTAB_SIZE;

/** Description of string table */
static struct ibft_string_block ibftab_strings = {
	.table = NULL,
	.offset = INITIAL_IBFTAB_SIZE,
};

/** The combined size of aligned, adjacent nic, target structures */
#define IBFTAB_ENTRY_SIZE						\
	( IBFTAB_STRUCT_SIZE ( struct ibft_nic ) +			\
	  IBFTAB_STRUCT_SIZE ( struct ibft_target ) )

/** Pointer to an existing valid iBFT from another bootrom (gPXE or not) */
static struct ibft_table *prev_ibft = NULL;

/** The number of nics/targets currently in the iBFT */
static unsigned int ibft_entry_count = 0;

/* Validate that every string present in the structure lies entirely within prev_ibft */
static int
ibft_validate_strings ( struct ibft_header *header )
{
	struct ibft_string *string;
	const off_t *loc;

	for_each_string_location ( loc, header->structure_id ) {
		string = ( ( void * ) header + *loc );
		if ( string->offset != 0 &&
		     ( ( string->offset + string->length > prev_ibft->acpi.length ) ||
		       ( string->offset + string->length < string->offset ) ) )
			return -E2BIG;
	}

	return 0;
}

/* Validate the structure contained within the (foreign or gPXE generated) iBFT,
 * then call the callback() on the structure */
static int
ibft_parse_structure ( struct ibft_table *table,
		       int ( *callback ) ( struct ibft_header *header ),
		       uint8_t structure_id, ibft_off_t offset )
{
	struct ibft_header *header;

	/* All structures are optional apart from CONTROL, and
	 * ibft_parse_table() guarantees that offset != 0 for CONTROL */
	if ( offset == 0 )
		return 0;

	if ( offset > table->acpi.length ) {
		DBG ( "%s: offset %d > table length %d\n", __func__,
		      offset, table->acpi.length );
		return -E2BIG;
	}

	header = ( void * )table + offset;
	if ( header->structure_id != structure_id ) {
		DBG ( "%s: header structure %d not %d\n", __func__,
		      header->structure_id, structure_id );
		return -EFAULT;
	}

	if ( header->length < ibft_struct_tbl [ structure_id ].min_size ) {
		DBG ( "%s: header too short %d < %d\n",
		      __func__, header->length,
		      ( int ) ibft_struct_tbl [ structure_id ].min_size );
		return -EMSGSIZE;
	}

	if ( header->length + offset > table->acpi.length ) {
		DBG ( "%s: header overruns table\n", __func__ );
		return -ENOSPC;
	}

	/* Execute the callback if BLOCK_VALID */
	if ( structure_id != IBFT_STRUCTURE_ID_CONTROL &&
	     ( header->flags & IBFT_FL_INITIATOR_BLOCK_VALID ) )
		return callback ( header );

	return 0;
}

/* Parse the given (foreign or gPXE generated) iBFT, and execute the
 * given callback for each structure present.
 */
static int
ibft_parse_table ( struct ibft_table *table,
		   int ( *callback ) ( struct ibft_header *header ) )
{
	ibft_off_t *pos, *end;
	int rc;

	rc = ibft_parse_structure ( table, callback, IBFT_STRUCTURE_ID_CONTROL,
				    offsetof ( struct ibft_table, control ) );
	if (rc != 0)
		return rc;
	rc = ibft_parse_structure ( table, callback, IBFT_STRUCTURE_ID_INITIATOR,
				    table->control.initiator );
	if (rc != 0)
		return rc;

	pos = &table->control.entry[0].nic;
	end = ( void * )&table->control + table->control.header.length;
	while ( pos + 1 < end ) {
		rc = ibft_parse_structure ( table, callback,
					    IBFT_STRUCTURE_ID_NIC,
					    * ( pos++ ) );
		if (rc != 0)
			return rc;
		rc = ibft_parse_structure ( table, callback,
					    IBFT_STRUCTURE_ID_TARGET,
					    * ( pos++ ) );
		if (rc != 0)
			return rc;
	}

	return 0;
}

/* Number of bytes to move all strings down by */
static int ibft_string_move_delta;

/* Move all string pointers in ibftab down by %ibft_string_move_delta */
static int
ibft_header_move_strings ( struct ibft_header *header )
{
	struct ibft_string *string;
	const off_t *loc;

	for_each_string_location ( loc, header->structure_id ) {
		string = ( void * )header + *loc;
		/* do not move empty strings, ie. strings at 0 offset */
		if ( string->offset != 0 )
			string->offset += ibft_string_move_delta;
	}

	return 0;
}

/* Try and move the string heap down ( or up ) by the given amount */
static int
ibft_move_string_heap ( long delta )
{
	unsigned int length;
	char *strings;
	int rc;

	if ( ibftab_strings.offset + delta > ( int )sizeof ( ibftab ) )
		return -ENOSPC;

	ibft_string_move_delta = delta;
	rc = ibft_parse_table ( &ibftab.table, ibft_header_move_strings );
	assert ( rc == 0 );

	length = ibftab_strings.offset - ibftab_next_structure_offset;
	strings = ( void * )&ibftab + ibftab_next_structure_offset;
	memmove ( strings + delta, strings, length );

	ibftab_strings.offset += delta;
	ibftab_next_structure_offset += delta;

	return 0;
}

/* Insert NIC and Target tables, populate the NIC and initiator tables,
 * and use the given callback to populate the target table
 */
static int
ibft_fill_data ( struct net_device *netdev,
		 int ( *fill_target ) ( struct ibft_target *, void *priv,
					struct ibft_string_block *strings ),
		 void *priv )
{
	struct ibft_target *target;
	struct ibft_nic *nic;
	int rc = 0;

	/* Currently limited to 2 entries */
	if ( ibft_entry_count == 2 )
		return -ENOSPC;
	/* More generally */
	if ( offsetof ( struct gpxe_ibft,
			table.control.entry[ibft_entry_count + 1] ) >=
	     offsetof ( struct gpxe_ibft, initiator ) )
		return -ENOSPC;

	assert ( ibftab.table.control.entry[ibft_entry_count].nic == 0 );
	assert ( ibftab.table.control.entry[ibft_entry_count].target == 0 );

	if ( ibftab.table.control.initiator == 0 ) {
		rc = ibft_fill_initiator ( &ibftab.initiator, &ibftab_strings,
					   netdev );
		if ( rc != 0 )
			return rc;
		ibftab.table.control.initiator =
			offsetof ( struct gpxe_ibft, initiator );
	}

	/* Move all existing string pointers down to make room for the
	 * nic and target structures */
	nic = ( void * )&ibftab + ibftab_next_structure_offset;
	target = ( void * )nic + IBFTAB_STRUCT_SIZE ( struct ibft_nic );
	if ( ( rc = ibft_move_string_heap ( IBFTAB_ENTRY_SIZE ) ) != 0 )
		goto checksum;

	memset ( nic, '\0', IBFTAB_ENTRY_SIZE );
	nic->header.structure_id = IBFT_STRUCTURE_ID_NIC;
	nic->header.version = 1;
	nic->header.length = sizeof ( struct ibft_nic );
	nic->header.flags = ( IBFT_FL_NIC_BLOCK_VALID |
			      IBFT_FL_NIC_FIRMWARE_BOOT_SELECTED );
	nic->header.index = ibft_entry_count;
	if ( ( rc = ibft_fill_nic ( nic, &ibftab_strings, netdev ) ) != 0 )
		goto fail;

	target->header.structure_id = IBFT_STRUCTURE_ID_TARGET;
	target->header.version = 1;
	target->header.length = sizeof ( struct ibft_target );
	target->header.flags = ( IBFT_FL_TARGET_BLOCK_VALID |
				 IBFT_FL_TARGET_FIRMWARE_BOOT_SELECTED );
	target->nic_association = ibft_entry_count;
	target->header.index = ibft_entry_count;
	if ( ( rc = fill_target ( target, priv, &ibftab_strings ) ) != 0 )
		goto fail;
	if ( ( rc = ibft_fill_target_chap ( target, &ibftab_strings,
					    netdev_nvo_settings ( netdev ) ) ) != 0 )
		goto fail;

	/* All done */
	ibftab.table.control.entry[ibft_entry_count].nic =
		( char * )nic - ( char * )&ibftab;
	ibftab.table.control.entry[ibft_entry_count].target =
		( char * )target - ( char * )&ibftab;
	if (++ibft_entry_count > 2)
		ibftab.table.control.header.length += 4;

	goto checksum;

fail:
	/* Move the string heap back */
	( void ) ibft_move_string_heap ( -IBFTAB_ENTRY_SIZE );

	/* Wipe the trailing portion of the string heap
	 * so the checksum is valid again */
	memset ( ( char * )&ibftab + ibftab_strings.offset, '\0',
		 sizeof ( ibftab ) - ibftab_strings.offset );

	/* fall-thru */
checksum:
	acpi_fix_checksum ( &ibftab.table.acpi );

	return rc;
}

static unsigned int ibft_prev_lowest_target_index = -1U;
static struct ibft_target *ibft_prev_lowest_target = NULL;

/* Find the lowest target structure in prev_ibft with the lowest index */
static int
ibft_find_lowest_target ( struct ibft_header *header )
{
	if ( header->structure_id == IBFT_STRUCTURE_ID_TARGET &&
	     header->index <= ibft_prev_lowest_target_index ) {
		ibft_prev_lowest_target_index = header->index;
		ibft_prev_lowest_target =
			container_of ( header, struct ibft_target, header );
	}

	return 0;
}
		
/* Copy initiator, the target with the lowest index, and it's corresponding
 * nic out of prev_ibft into ibftab. The initiator will be copied directly
 * into ibftab.table.initator, whilst nic and target0 will be allocated
 * properly.
 */
static int
ibft_copy_prev_structures ( struct ibft_header *header )
{
	ibft_off_t offset = ( char * ) header - ( char * ) prev_ibft;
	struct ibft_header *new_header;
	const off_t *loc;
	int rc;

	if ( header->structure_id == IBFT_STRUCTURE_ID_NIC &&
	     header->index == ibft_prev_lowest_target->nic_association ) {
		offset = INITIAL_IBFTAB_SIZE;
		ibftab.table.control.entry[0].nic = offset;

	} else if ( header == &ibft_prev_lowest_target->header ) {
		offset = ( INITIAL_IBFTAB_SIZE +
			   IBFTAB_STRUCT_SIZE ( struct ibft_nic ) );
		ibftab.table.control.entry[0].target = offset;

	} else if ( header->structure_id == IBFT_STRUCTURE_ID_INITIATOR ) {
		offset = offsetof ( struct gpxe_ibft, initiator );
		ibftab.table.control.initiator = offset;

	} else
		/* Don't copy this structure */
		return 0;

	/* Push NIC and Target as Index=0 */
	new_header = ( void * )&ibftab + offset;
	new_header->structure_id = header->structure_id;
	new_header->version = 1;
	new_header->length = ibft_struct_tbl [ header->structure_id ].min_size;
	new_header->index = 0;
	new_header->flags = header->flags;

	/* Copy the rest of the structure up to the size we allocated,
	 * including string offsets (which will be wrong) */
	memcpy ( ( char * ) new_header + sizeof ( struct ibft_header ),
		 ( char * ) header + sizeof ( struct ibft_header ),
		 new_header->length - sizeof ( struct ibft_header ) );

	if ( header->structure_id == IBFT_STRUCTURE_ID_TARGET ) {
		struct ibft_target *target =
			container_of ( new_header, struct ibft_target, header );
		target->nic_association = 0;
	}

	/* And now copy the strings properly */
	for_each_string_location ( loc, header->structure_id ) {
		struct ibft_string *from = ( ( void * )header + *loc );
		struct ibft_string *to = ( ( void * )new_header + *loc );
		char *data = ( void * )prev_ibft + from->offset;

		if ( from->offset == 0 )
			continue;

		if ( ( rc = ibft_set_string ( &ibftab_strings, to,
					      data, from->length ) ) != 0 )
			return rc;
	}

	return 0;
}

/* Scan for an existing table, storing the result in prev_ibft */
static void
ibft_scan ( void )
{
	const unsigned start = 0x80000;
	const unsigned finish = 0x100000;
	const unsigned step = 0x10;
	unsigned phys_addr, i;
	void *virt_addr;
	uint8_t sum;
	int rc;

	for ( phys_addr = start; phys_addr < finish; phys_addr += step ) {
		virt_addr = phys_to_virt ( phys_addr );
		if ( virt_addr == &ibftab )
			continue;

		if ( memcmp ( virt_addr, IBFT_SIG, 4 ) == 0 ) {
			prev_ibft = virt_addr;

			DBG ( "Found possible iBFT @ %08x\n", phys_addr );

			/* Verify the checksum */
			sum = 0;
			for ( i = 0; i < prev_ibft->acpi.length; i++ )
				sum += ( ( uint8_t * ) prev_ibft ) [ i ];
			if (sum != 0) {
				DBG ( "  Invalid checksum '%x'\n", sum );
				continue;
			}

			/* Validate the entire table, including strings */
			rc = ibft_parse_table ( virt_addr, ibft_validate_strings );
			if ( rc != 0 ) {
				DBG ( "  Failed parsing %d - %s",
				      rc, strerror ( rc )  );
				continue;
			}

			return;
		}
	}

	prev_ibft = NULL;
}
	
static void ibft_initialize ( void )
{
	int rc;

	if ( ibftab_strings.table != NULL )
		return;

	/* ibftab is in data16, so the offset isn't known at link time */
	ibftab_strings.table = &ibftab.table;

	/* Scan for an existing valid iBFT, and if it's present, copy
	 * all or none of {initiator, nic0, and target0} into the new one */
	ibft_scan ( );
	if ( prev_ibft != NULL ) {
		/* Create room for the nic0 and target0 tables */
		ibftab_next_structure_offset = ibftab_strings.offset =
			INITIAL_IBFTAB_SIZE + IBFTAB_ENTRY_SIZE;

		/* Find the target (and therefore NIC) that we want */
		rc = ibft_parse_table ( prev_ibft, ibft_find_lowest_target );
		assert ( rc == 0 );

		rc = ibft_parse_table ( prev_ibft, ibft_copy_prev_structures );
		if ( rc != 0 ||
		     ibftab.table.control.initiator == 0 ||
		     ibftab.table.control.entry[0].nic == 0 ||
		     ibftab.table.control.entry[0].target == 0) {

			DBG ( "%s: Incomplete parse rc %d initiator %d "
			      "nic %d target %d\n", __func__, rc,
			      ibftab.table.control.initiator,
			      ibftab.table.control.entry[0].nic,
			      ibftab.table.control.entry[0].target );

			/* Wipe the new tables and empty the string heap */
			ibftab_next_structure_offset = ibftab_strings.offset =
				INITIAL_IBFTAB_SIZE;

			ibftab.table.control.initiator = 0;
			ibftab.table.control.entry[0].nic = 0;
			ibftab.table.control.entry[0].target = 0;
		} else
			ibft_entry_count = 1;

		prev_ibft->acpi.signature[0] = 'x';
	}
}

int
ibft_fill_iscsi ( struct iscsi_session *iscsi )
{
	ibft_initialize ( );

	return ibft_fill_data ( iscsi->netdev, ibft_fill_iscsi_target, iscsi );
}

int
ibft_fill_netdev ( struct net_device *netdev )
{
	ibft_initialize ( );

	return ibft_fill_data ( netdev, ibft_fill_netdev_target, netdev );
}

void ibft_clear ( void )
{
	if ( prev_ibft != NULL )
		prev_ibft->acpi.signature[0] = 'i';

	ibftab.table.acpi.signature[0] = 'x';
	return;
}
