/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <registers.h>
#include <gpxe/init.h>
#include <gpxe/instance.h>
#include <realmode.h>
#include <gpxe/clp.h>
#include <gpxe/if_ether.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

struct list_head clp_devices = LIST_HEAD_INIT(clp_devices);
static char response_string[CLP_MAX_STR_LEN];

void register_clp_device(struct clp_device *dev)
{
	list_add_tail(&dev->list_head, &clp_devices);
}

void unregister_clp_device(struct clp_device *dev)
{
	list_del(&dev->list_head);
}

/* The associated CLP device must match the netport number and the current
 * network adapter.
 */
struct clp_device *clp_device_from_netport(int netport)
{
	struct clp_device *dev;
	uint32_t pci_busdevfn_from_prefix = instance_pci_busdevfn();

	uint8_t our_bus = bus_from_busdevfn(pci_busdevfn_from_prefix);
	uint8_t our_dev = dev_from_busdevfn(pci_busdevfn_from_prefix);

	list_for_each_entry(dev, &clp_devices, list_head) {
		if (dev->pci_dev && dev->pci_dev->bus == our_bus &&
		    dev_from_busdevfn(dev->pci_dev->devfn) == our_dev &&
		    dev->netport == netport)
			return dev;
	}
	return NULL;
}


static inline void driver_load()
{
	initialise();
	startup();
}

static inline void driver_unload()
{
	shutdown_exit();
}

/* This structure must match with the clp_info definition in romprefix.S.
 */
struct clp_info {
	uint16_t es;
	uint32_t edi;
	uint16_t ds;
	uint32_t esi;
	uint32_t status;
} __attribute__((packed));

/**
 * convert a string to an integer
 *
 * @param     str
 * @return    The converted value.
 */
int atoi(char *str)
{
	int number = 0;
	char* p = str;

	while((*p >= '0') && (*p <= '9')) {
		number = number * 10 + (*p - '0');
		p++;
	}
	return number;
}

void netport_reset(struct clp_response *response, int netport)
{
	struct clp_device *clp_dev = clp_device_from_netport(netport);

	if (clp_dev == NULL || clp_dev->reset == NULL)
		goto error;

	clp_dev->reset(clp_dev, response);

	return;

error:
	response->string = CLP_STRING_INTERNAL_ERROR;
	response->status = CLP_STATUS_INTERNAL_ERROR;
}

static int byte_from_hex_char(char c)
{
	if ('0' <= c && c <= '9')
		return c - '0';
	if ('a' <= c && c <= 'f')
		return 10 + c - 'a';
	if ('A' <= c && c <= 'F')
		return 10 + c - 'A';
	return -1;
}

void netport_set_perm_mac(struct clp_response *response, int netport, char *arg)
{
	struct clp_device *clp_dev = clp_device_from_netport(netport);
	int start_quote = 0, end_quote = 0, require_quotes = 0;
	unsigned mac_index = 0, mac_byte_index = 0;
	uint8_t mac[ETH_ALEN];
	char c;
	int tmp;

	DBG("Setting permanent MAC address to %s for netport #%d\n", arg,
	    netport);

	if (clp_dev == NULL || clp_dev->set_perm_mac == NULL)
		goto internal_error;

	/* Parse the string; see HP BladeSystem HSS spec (SF-109836-PD-1) */
	if (*arg == '"') {
		start_quote = 1;
		require_quotes = 1;
		++arg;
	}

	for (; (c = *arg) && mac_index < ETH_ALEN; ++arg) {
		if (c ==  ' ' || c == '-' || c == '.' || c == ':' || c == ',') {
			require_quotes = 1;
			continue;
		}

		tmp = byte_from_hex_char(c);
		if (tmp < 0)
			goto syntax_error;

		if (mac_byte_index == 0) {
			mac[mac_index] = tmp << 4;
			++mac_byte_index;
		} else {
			mac[mac_index] += tmp;
			mac_byte_index = 0;
			++mac_index;
		}
	}

	if (*arg == '"') {
		end_quote = 1;
		require_quotes = 1;
		++arg;
	}

	if (require_quotes && (!start_quote || !end_quote))
		goto syntax_error;

	if (mac_index != ETH_ALEN)
		goto syntax_error;

	clp_dev->set_perm_mac(clp_dev, response, mac);

	return;

internal_error:
	response->string = CLP_STRING_INTERNAL_ERROR;
	response->status = CLP_STATUS_INTERNAL_ERROR;
	return;
syntax_error:
	response->string = CLP_STRING_SYNTAX_ERROR;
	response->status = CLP_STATUS_SYNTAX_ERROR;
}

void netport_set_boot(struct clp_response *response, int netport, char *arg)
{
	struct clp_device *clp_dev = clp_device_from_netport(netport);
	int enable;

	DBG("Enabling network booting on netport #%d (dev %p)\n", netport,
	       clp_dev);

	if (clp_dev == NULL || clp_dev->set_boot == NULL)
		goto error;

	if (strncmp(arg, "enabled", 7) == 0)
		enable = 1;
	else if (strncmp(arg, "disabled", 8) == 0)
		enable = 0;
	else if (strcmp(arg, "") == 0)
		goto out;
	else
		goto error;

	clp_dev->set_boot(clp_dev, response, enable);

	return;

out:
	response->string = CLP_STRING_SUCCESS;
	response->status = CLP_STATUS_SUCCESS;
	return;
error:
	response->string = CLP_STRING_INTERNAL_ERROR;
	response->status = CLP_STATUS_INTERNAL_ERROR;
}

/* The OEMHP_LF<n> command is documented in SF-109835-PD-1, section 3.3 */
void netport_set_hp_lf(struct clp_response *response, int netport, char *args)
{
	struct clp_device *clp_dev = clp_device_from_netport(netport);
	enum hp_lf_args arg_index = HP_LF_PROTOCOL;
	int error = 0;
	char *arg;

	DBG("Configuring the logical function on netport #%d (dev %p)\n",
	    netport, clp_dev);

	/* The string must be enclosed in quotes */
	if (*args != '"' || *(args + strlen(args) - 1) != '"')
		goto syntax_error;

	++args;
	*(args + strlen(args) - 1) = 0;

	while ((arg = strsep(&args, ";")) != NULL) {
		switch (arg_index) {
		case HP_LF_MAC:
			netport_set_perm_mac(response, netport, arg);
			if (response->status != CLP_STATUS_SUCCESS)
				error = 1;
			break;
		case HP_LF_BOOT:
			netport_set_boot(response, netport, arg);
			if (response->status != CLP_STATUS_SUCCESS)
				error = 1;
			break;
		case HP_LF_PROTOCOL:
		case HP_LF_PRIORITY:
		case HP_LF_VNI:
		case HP_LF_MINBW:
		case HP_LF_MAXBW:
		case HP_LF_ENABLE:
		case HP_LF_SRIOV:
			break;
		}
		++arg_index;
	}

	if (error) {
		response->string = CLP_STRING_INTERNAL_ERROR;
		response->status = CLP_STATUS_INTERNAL_ERROR;
	}

	return;

syntax_error:
	response->string = CLP_STRING_SYNTAX_ERROR;
	response->status = CLP_STATUS_SYNTAX_ERROR;
}

void clp_parse_set_netport(struct clp_response *response, int netport,
			   char *args)
{
	char *arg;

	while ((arg = strsep(&args, " ")) != NULL) {
		char *key = strsep(&arg, "=");
		if (strcmp(key, "default") == 0) {
			netport_reset(response, netport);
		} else if (strcmp(key, "PermanentAddress") == 0) {
			netport_set_perm_mac(response, netport, arg);
		} else if (strcmp(key, "OEMHP_BootEnable") == 0) {
			netport_set_boot(response, netport, arg);
		} else if (strcmp(key, "OEMHP_LF0") == 0) {
			netport_set_hp_lf(response, netport, arg);
		} else {
			response->string = CLP_STRING_UNSUPPORTED;
			response->status = CLP_STATUS_UNSUPPORTED;
			return;
		}

		/* The lower byte of this status contains the CLP command
		 * status; higher bits can be set to indicate support levels.
		 */
		if ((response->status & 0xff) != CLP_COMMAND_COMPLETED) {
			/* We've had a failure, so we can't reasonably parse
			 * the rest of the command.
			 */
			return;
		}
	}
}

void clp_parse_set(struct clp_response *response, char *args)
{
	if (strncmp(args, "netport", 7) == 0) {
		args += 7;
		char* netport_str = strsep(&args, " ");
		if (args == NULL) {
			response->string = CLP_STRING_UNSUPPORTED;
			response->status = CLP_STATUS_UNSUPPORTED;
			return;
		}
		clp_parse_set_netport(response, atoi(netport_str), args);
	}
}

void netport_show_perm_mac(struct clp_response *response, int netport)
{
	struct clp_device *clp_dev = clp_device_from_netport(netport);
	uint8_t mac[ETH_ALEN];
	int rc;

	DBG("Getting the MAC address on netport #%d (dev %p)\n", netport,
	    clp_dev);
	if (clp_dev == NULL) {
		response->string = CLP_STRING_INVALID_TARGET;
		response->status = CLP_STATUS_INVALID_TARGET;
		return;
	}

	if (!clp_dev->get_mac)
		goto error;
	rc = clp_dev->get_mac(clp_dev, mac);
	if (rc)
		goto error;

	response->status = CLP_STATUS_SUCCESS;
	response->string = response_string;
	if (response->flags & CLP_FORMAT_OEMHP_BINARY) {
		/* The first byte reports the number of MAC addresses being
		 * returned:  we only return one.
		 */
		memset(response_string, 1, 1);
		memcpy(response_string + 1, mac, ETH_ALEN);
		response->length = ETH_ALEN + 1;
	} else {
		response->length = snprintf(response_string, CLP_MAX_STR_LEN,
			"status=0\nstatus_tag=COMMAND COMPLETED\n"
			"PermanentAddress=%02x%02x%02x%02x%02x%02x\n"
			"endoutput\n",
			mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	}

	return;
error:
	response->string = CLP_STRING_INTERNAL_ERROR;
	response->status = CLP_STATUS_INTERNAL_ERROR;
}

void clp_parse_show_netport(struct clp_response *response, int netport,
			    char *args)
{
	char *arg;

	while ((arg = strsep(&args, " ")) != NULL) {
		char *key = strsep(&arg, "=");
		if (strcmp(key, "PermanentAddress") == 0) {
			netport_show_perm_mac(response, netport);
		}
	}
}

void clp_parse_show(struct clp_response *response, char *args) {
	response->flags = 0;
	if (strncmp(args, "-o format=OEMHP_binary ", 23) == 0) {
		args += 23;
		response->flags |= CLP_FORMAT_OEMHP_BINARY;
	}

	if (strncmp(args, "netport", 7) == 0) {
		args += 7;
		char* netport_str = strsep(&args, " ");
		if (args == NULL) {
			response->string = CLP_STRING_UNSUPPORTED;
			response->status = CLP_STATUS_UNSUPPORTED;
			return;
		}
		clp_parse_show_netport(response, atoi(netport_str), args);
	}
}

/**
 * Parses a SMASH CLP command string, placing the result of the command into the
 * response argument.
 *
 * @param input       the command string to be parsed.
 * @param response    the response to be returned to the caller.
 */
void clp_parse(struct clp_response *response, char *input)
{
	int	i;

	/* Sanity check the input string */
	DBG2("CLP command string: ");
	for (i=0; (i < CLP_INPUT_BUF_SIZE) && input[i]; i++) {
		if (!isprint(input[i])) {
			DBG2("\nNon-printable character 0x%x at index %d\n",
			     input[i] & 0xff, i);
			return;
		}
		DBG2("%c", input[i]);
	}
	DBG2("\n");

	if (strcmp(input, "exit") == 0) {
		/* In our case, exit is a no-op, we have no clean up to do. */
		response->string = CLP_STRING_SUCCESS;
		response->status = CLP_STATUS_SUCCESS;
	} else if (strncmp(input, "set ", 4) == 0) {
		char *args = input + 4;
		clp_parse_set(response, args);
	} else if (strncmp(input, "show ", 5) == 0) {
		char *args = input + 5;
		clp_parse_show(response, args);
	}
}


__asmcall void c_clp_entry(struct i386_all_regs *ix86)
{
	struct clp_info clp_info;
	struct clp_response clp_response;
	char input[CLP_INPUT_BUF_SIZE];
	int outlen;

	instance_set_pci_busdevfn(ix86);
	DBG("Configuring Solarflare network adapter using DMTF CLP for device %x\n",
	    instance_pci_busdevfn());

	copy_from_real(&clp_info, ix86->segs.ds, ix86->regs.edi,
		       sizeof(clp_info));

	memset(&clp_response, 0, sizeof(clp_response));
	clp_response.string = CLP_STRING_UNSUPPORTED;
	clp_response.status = CLP_STATUS_UNSUPPORTED;

	DBG("command at es:edi=%x:%x\n", clp_info.es, clp_info.edi);
	copy_from_real(input, clp_info.es, clp_info.edi, CLP_INPUT_BUF_SIZE);

	driver_load();
	clp_parse(&clp_response, input);
	driver_unload();

	if (strlen(clp_response.string) + 1 > CLP_OUTPUT_BUF_SIZE) {
		clp_response.status = CLP_STATUS_INTERNAL_ERROR;
		clp_response.string = CLP_STRING_INTERNAL_ERROR;
	}

	/* For binary output, clp_response.length should have been set. */
	if (clp_response.length)
		outlen = clp_response.length;
	else
		outlen = strlen(clp_response.string) + 1;

	copy_to_real(clp_info.ds, clp_info.esi, clp_response.string, outlen);
	clp_info.status = clp_response.status;
	DBG("CLP status=%d reponse(%d characters)=%s to ds:esi=%x:%x\n",
	    clp_response.status, outlen, clp_response.string,
	    clp_info.ds, clp_info.esi);
	copy_to_real(ix86->segs.ds, ix86->regs.edi, &clp_info,
		     sizeof(clp_info));
}
