/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>

/**
 * strlen - Find the length of a string
 * @s: The string to be sized
 */
size_t wcslen(const wchar_t * s)
{
	const wchar_t *sc;
	
	for (sc = s; *sc != L'\0'; ++sc)
		/* nothing */;
	return sc - s;
}


/**
 * strstr - Find the first substring in a %NUL terminated string
 * @s1: The string to be searched
 * @s2: The string to search for
 */
wchar_t * wcsstr(const wchar_t * s1, const wchar_t * s2)
{
	int l1, l2;
	
	l2 = wcslen(s2);
	if (!l2)
		return (wchar_t *) s1;
	l1 = wcslen(s1);
	while (l1 >= l2) {
		l1--;
		if (!memcmp(s1,s2,l2*2))
			return (wchar_t *) s1;
		s1++;
	}
	return NULL;
}


/**
 * Write a formatted wide string to a wide string buffer
 *
 * @v buf		Buffer into which to write the wide string
 * @v size		Size of buffer (wide characters)
 * @v fmt		Wide format string
 * @v args		Arguments corresponding to the format string
 * @ret len		Length in characters (not bytes) of formatted string
 *
 * If the buffer is too small to contain the string, the returned
 * length is the length that would have been written had enough space
 * been available.
 */
int vswprintf(wchar_t *wbuf, size_t size, const wchar_t *wfmt, va_list args)
{
	size_t fmt_len = wcslen ( wfmt );
	char *fmt, *buf;
	int ret;

	/* alloc and convert fmt */
	fmt = zalloc ( fmt_len * 2 + 1 );
	wcstombs ( fmt, wfmt, fmt_len );

	/* do the UTF8 printf */
	buf = zalloc( size * 2 + 1 );
	vsnprintf( buf, size, fmt, args );

	/* convert back to wide characters */
	ret = mbstowcs ( wbuf, buf, size );

	free ( fmt );
	free ( buf );
	return ret;
}

/**
 * Write a formatted wide string
 *
 * @v fmt		Format string
 * @v ...		Arguments corresponding to the format string
 * @ret	len		Length of formatted string
 */
int swprintf(wchar_t *wcs, size_t size, const wchar_t *wfmt, ...) {
	va_list args;
	int ret;
	
	va_start ( args, wfmt );
	ret = vswprintf ( wcs, size, wfmt, args );
	va_end ( args );
	return ret;
}


/**
 * Convert a multibyte string to a wide-character string
 */
size_t mbstowcs(wchar_t *out, const char *in, size_t outlen)
{
	size_t inlen = strlen ( in );
	return mbsnrtowcs( out, &in, inlen, outlen, NULL );
}


/**
 * Convert a multibyte string to a wide-character string
 * Always guaranteed a NULL at the end - even if not enough output buffer
 * 
 * Adpated from BSD licensed:
 *    <edk2>/MdeModulePkg/Universal/Console/TerminalDxe/Vtutf8.c
 *
 * @v in		UTF8 string (doesn't need to be NULL terminated)
 * @v inlen		Length of ASCII string to process
 * @v out		UCS2 unicode buffer
 * @v outlen		Size of UCS2 buffer *in characters*
 * @ret num		Number of *characters* copied (exc. NULL)
 */

/*
 *  Bits per Character | Unicode Character Range | Unicode Binary  Encoding | UTF8 Binary Encoding
 *       0-7	      |     0x0000 - 0x007F	|     00000000 0xxxxxxx	    | 0xxxxxxx
 *       8-11 	      |     0x0080 - 0x07FF	|     00000xxx xxxxxxxx     | 110xxxxx 10xxxxxx
 *      12-16	      |     0x0800 - 0xFFFF	|     xxxxxxxx xxxxxxxx	    | 1110xxxx 10xxxxxx 10xxxxxx
 */
size_t mbsnrtowcs(wchar_t *out, const char **_in, size_t inlen, size_t outlen,
		  mbstate_t *ps __unused)
{
	const unsigned char *in = (const unsigned char *)(*_in);
	int utf_remain = 0;
	wchar_t outval = L'\0';
	int ip = 0;
	size_t op = 0;
	
	while ( inlen > 0  && ( !out || ( outlen > 1 ) ) && in[ip] ) {

		switch (utf_remain) {
		case 0:
			if ( in[ip] < 0x80 ) {
				outval = in[ip];
			} else if ( ( in[ip] & 0xe0) == 0xc0 ) {
				outval = ( in[ip] & 0x1f ) << 6;
				utf_remain = 1;
			} else if ( (in[ip] & 0xf0) == 0xe0 ) {
				outval = ( in[ip] & 0x0f ) << 12;
				utf_remain = 2;
			} else
				return (size_t) -1;				
			break;
		case 1:
			if ( ( in[ip] & 0xc0 ) != 0x80 )
				return (size_t) -1;
			outval |= ( in[ip] & 0x3f );
			utf_remain = 0;
			break;
		case 2:
			if ( ( in[ip] & 0xc0 ) != 0x80 )
				return (size_t) -1;
			outval |= ( in[ip] & 0x3f ) << 6;
			utf_remain = 1;
			break;
		default:
			assert ( 0 );
		}

		if ( out )
			out[op] = outval;

		if ( utf_remain == 0 ) {
			outlen--;
			op++;
		}

		inlen--;
		ip++;
	}
	out[op] = 0;
	
	return op;	
}


/**
 * convert a wide-character string to a multibyte string
 *
 * Adpated from BSD licensed:
 *    <edk2>/MdeModulePkg/Universal/Console/TerminalDxe/Vtutf8.c
 *
 * @v out		UTF8 buffer
 * @v in		UCS2 unicode string
 * @v outlen		Length of dest buffer (bytes)
 * @ret num		Number of bytes output (exc. NULL)
 */
size_t wcstombs(char *out, const wchar_t *in, size_t outlen)
{
	int ip = 0, op = 0;
	
	while ( in[ip] ) {
		size_t outn;
		
		if ( in[ip] < 0x80 )
			outn = 1;
		else if ( in[ip] < 0x800 )
			outn = 2;
		else
			outn = 3;
		
		if ( out && ( outlen < outn ) )
			break;

		/* Write to output */
		if ( out && outn == 1 ) {
			out[op] = in[ip];
		} else if ( out && outn == 2 ) {
			out[op+0] = 0xc0 | ( (in[ip] & 0x03c0) >> 6 );
			out[op+1] = 0x80 | ( (in[ip] & 0x00bf) >> 0 );
		} else if ( out && outn == 3 ) {
			out[op+0] = 0xe0 | ( (in[ip] & 0xf000) >> 12 );
			out[op+1] = 0x80 | ( (in[ip] & 0x0fc0) >> 6 );
			out[op+2] = 0x80 | ( (in[ip] & 0x003f) >> 0 );
		}		
		ip++;
		op += outn;
		outlen -= outn;
	}
	if ( out )
		out[op] = 0;
	
	return op;
}
