/**************************************************************************
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver datapath common code for Solarflare network cards
 *
 * Written by Shradha Shah <sshah@solarflare.com>
 *
 * Copyright Fen Systems Ltd. 2005
 * Copyright Level 5 Networks Inc. 2005
 * Copyright 2006-2017 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 *
 ***************************************************************************/
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <byteswap.h>
#include <gpxe/io.h>
#include <gpxe/pci.h>
#include <gpxe/malloc.h>
#include <gpxe/iobuf.h>
#include <gpxe/netdevice.h>
#ifndef SFC_IPXE
#include <gpxe/dhcp.h>
#include <config/general.h>
#include <drivers/net/mii.h>
#include "tlv_layout.h"
#endif	/* SFC_IPXE */
#include "efx_common.h"
#include "efx_bitfield.h"
#include "mc_driver_pcol.h"

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/*******************************************************************************
 *
 *
 * Low-level hardware access
 *
 *
 ******************************************************************************/
#ifndef SFC_IPXE
void
efx_write(struct efx_nic *efx, efx_oword_t *value, unsigned int reg)
{
	EFX_REGDUMP("Writing register %x with " EFX_OWORD_FMT "\n",
		    reg, EFX_OWORD_VAL(*value));

	_efx_writel(efx, value->u32[0], reg + 0);
	_efx_writel(efx, value->u32[1], reg + 4);
	_efx_writel(efx, value->u32[2], reg + 8);
	wmb();
	_efx_writel(efx, value->u32[3], reg + 12);
	wmb();
}

void
efx_read(struct efx_nic *efx, efx_oword_t *value, unsigned int reg)
{
	value->u32[0] = _efx_readl(efx, reg + 0);
	wmb();
	value->u32[1] = _efx_readl(efx, reg + 4);
	value->u32[2] = _efx_readl(efx, reg + 8);
	value->u32[3] = _efx_readl(efx, reg + 12);
	EFX_REGDUMP("Read from register %x, got " EFX_OWORD_FMT "\n",
		    reg, EFX_OWORD_VAL(*value));
}
#endif	/* SFC_IPXE */

void
efx_writel(struct efx_nic *efx, efx_dword_t *value, unsigned int reg)
{
	DBGCIO(efx, "Writing partial register %x with " EFX_DWORD_FMT "\n",
	       reg, EFX_DWORD_VAL(*value));
	_efx_writel(efx, value->u32[0], reg);
}

void
efx_readl(struct efx_nic *efx, efx_dword_t *value, unsigned int reg)
{
	value->u32[0] = _efx_readl(efx, reg);
	DBGCIO(efx, "Read from register %x, got " EFX_DWORD_FMT "\n",
	       reg, EFX_DWORD_VAL(*value));
}

/*******************************************************************************
 *
 *
 * Inititialization and Close
 *
 *
 ******************************************************************************/
void efx_probe(struct net_device *netdev, enum efx_revision revision)
{
	struct efx_nic *efx = netdev_priv(netdev);
	struct pci_device *pci = container_of(netdev->dev,
					      struct pci_device, dev);
	unsigned int reg = PCI_BASE_ADDRESS_0;
	uint32_t bar_low;

	efx->netdev = netdev;
	efx->revision = revision;

	/* Find the memory bar to use */
	pci_read_config_dword(pci, reg, &bar_low);
	if ((bar_low & PCI_BASE_ADDRESS_IO_MASK) == PCI_BASE_ADDRESS_SPACE_IO)
		reg = PCI_BASE_ADDRESS_2;

	efx->mmio_start = pci_bar_start(pci, reg);
	efx->mmio_len = pci_bar_size(pci, reg);
	efx->membase = ioremap(efx->mmio_start, efx->mmio_len);

	DBGCP(efx, "BAR of %lx bytes at phys %lx mapped at %p\n",
	      efx->mmio_len, efx->mmio_start, efx->membase);

	/* Enable PCI access */
	adjust_pci_device(pci);
}

void efx_remove(struct net_device *netdev)
{
	struct efx_nic *efx = netdev_priv(netdev);

	iounmap(efx->membase);
	efx->membase = NULL;
}
#ifndef SFC_IPXE

void efx_suspend(struct net_device *netdev)
{
	struct efx_nic *efx = netdev_priv(netdev);

	EFX_LOG("suspend %p\n", netdev);
	netdev->op->close(netdev);
	if(efx->membase) {
		iounmap(efx->membase);
		efx->membase = NULL;
	}
	EFX_LOG("suspend %p done\n", netdev);
}

void efx_resume(struct net_device *netdev)
{
	struct efx_nic *efx = netdev_priv(netdev);

	EFX_LOG("resume %p\n", netdev);
	if(efx->membase == NULL) {
		efx->membase = ioremap(efx->mmio_start, efx->mmio_len);
		EFX_TRACE("BAR of %lx bytes at phys %lx re-mapped at %p\n",
			  efx->mmio_len, efx->mmio_start, efx->membase);
	}
	netdev->op->open(netdev);
	EFX_LOG("resume %p done\n", netdev);
}

/*******************************************************************************
 *
 *
 * SMASH CLP
 *
 *
 *******************************************************************************/
#ifdef __i386__
void efx_register_clp(struct pci_device *pci)
{
	struct efx_nic *efx = netdev_priv(pci_get_drvdata(pci));
	struct clp_device *clp_dev = &efx->clp_dev;

	clp_dev->device = efx;
	sfc_init_clp_dev(clp_dev, pci);
	register_clp_device(clp_dev);
}

void efx_unregister_clp(struct pci_device *pci)
{
	struct efx_nic *efx = netdev_priv(pci_get_drvdata(pci));
	struct clp_device *clp_dev = &efx->clp_dev;

	unregister_clp_device(clp_dev);
	sfc_fini_clp_dev(clp_dev);
}

int efx_get_boot_type(struct net_device *netdev)
{
	struct efx_nic *efx = netdev_priv(netdev);
	int rc;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CLP_IN_GET_BOOT_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_CLP_OUT_GET_BOOT_LEN);

	MCDI_SET_DWORD(inbuf, CLP_IN_OP, MC_CMD_CLP_OP_GET_BOOT);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_CLP,
				 inbuf, sizeof(inbuf),
				 outbuf, sizeof(outbuf),
				 NULL, false);

	/* Return the boot flag on success, or a (negative) error code */
	return rc ? rc : *MCDI_PTR(outbuf, CLP_OUT_GET_BOOT_FLAG);
}
#else
/* CLP is only supported in the i386 (ie. non-EFI) build */
void efx_register_clp(struct pci_device *pci __unused)
{
}

void efx_unregister_clp(struct pci_device *pci __unused)
{
}

int efx_get_boot_type(struct net_device *netdev __unused)
{
	return 0;
}
#endif

int efx_common_option_supported(struct net_device *netdev, int tag)
{
	(void) netdev;
	switch (tag) {
	case DHCP_EB_BOOT_IMAGE:
	case DHCP_EB_BOOT_TYPE:
	case DHCP_EB_TARGET_DHCP:
	case DHCP_RP_TARGETNAME:
	case DHCP_RP_SERVERNAME:
	case DHCP_RP_PORT:
	case DHCP_RP_LUN:
	case DHCP_EB_LUN_BUSY_RETRIES:
	case DHCP_EB_INITIATOR_DHCP:
	case DHCP_EB_YIADDR:
	case DHCP_SUBNET_MASK:
	case DHCP_ROUTERS:
	case DHCP_DNS_SERVERS:
	case DHCP_EB_INITIATOR_IQN_DHCP:
	case DHCP_ISCSI_INITIATOR_IQN:
	case DHCP_VENDOR_CLASS_ID:
	case DHCP_EB_CHAP:
	case DHCP_EB_USERNAME:
	case DHCP_EB_PASSWORD:
	case DHCP_EB_CHAP_TARGET:
	case DHCP_EB_REVERSE_USERNAME:
	case DHCP_EB_REVERSE_PASSWORD:
	case DHCP_EB_MPIO_PRIORITY:
	case DHCP_EB_LINK_SPEED:
	case DHCP_EB_LINKUP_DELAY:
	case DHCP_EB_MSIX_LIMIT:
		return 1;
	}
	return 0;
}

/*******************************************************************************
 *
 *
 * MAC statistics
 *
 *
 *******************************************************************************/

#ifdef IFMGMT_CMD
/* Adapted from v5 linux_net/mcdi_port.c */
#define STATS_PERIOD_MS_DEFAULT 1000
#define EFX_MC_STATS_GENERATION_INVALID ((__le64)(-1))

enum efx_stats_action {
	EFX_STATS_ENABLE,
	EFX_STATS_DISABLE,
	EFX_STATS_PULL,
	EFX_STATS_PERIOD,
};

static void efx_mcdi_mac_stats(struct efx_nic *efx,
			       enum efx_stats_action action, int clear)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_MAC_STATS_IN_LEN);
	int dma, change;
	dma_addr_t dma_addr = efx->stats_buffer.dma_addr;
	u32 dma_len;

	if (action == EFX_STATS_ENABLE)
		efx->stats_enabled = true;
	else if (action == EFX_STATS_DISABLE)
		efx->stats_enabled = false;

	if (action == EFX_STATS_PULL) {
		dma = 0;
		change = 0;
	} else {
		change = 1;
		dma = efx->stats_enabled;
	}

	dma_len = dma ? MC_CMD_MAC_NSTATS * sizeof(u64) : 0;

	BUILD_BUG_ON(MC_CMD_MAC_STATS_OUT_DMA_LEN != 0);

	MCDI_SET_QWORD(inbuf, MAC_STATS_IN_DMA_ADDR, dma_addr);
	MCDI_POPULATE_DWORD_7(inbuf, MAC_STATS_IN_CMD,
			      MAC_STATS_IN_DMA, dma,
			      MAC_STATS_IN_CLEAR, clear,
			      MAC_STATS_IN_PERIODIC_CHANGE, change,
			      MAC_STATS_IN_PERIODIC_ENABLE, efx->stats_enabled,
			      MAC_STATS_IN_PERIODIC_CLEAR, 0,
			      MAC_STATS_IN_PERIODIC_NOEVENT, 1,
			      MAC_STATS_IN_PERIOD_MS, STATS_PERIOD_MS_DEFAULT);
	MCDI_SET_DWORD(inbuf, MAC_STATS_IN_DMA_LEN, dma_len);
	MCDI_SET_DWORD(inbuf, MAC_STATS_IN_PORT_ID, EVB_PORT_ID_ASSIGNED);

	efx->type->mcdi_rpc(efx, MC_CMD_MAC_STATS, inbuf, sizeof(inbuf),
			    NULL, 0, NULL, false);
}

static void efx_mcdi_mac_start_stats(struct efx_nic *efx)
{
	__le64 *dma_stats = efx->stats_buffer.addr;

	dma_stats[MC_CMD_MAC_GENERATION_END] = EFX_MC_STATS_GENERATION_INVALID;

	efx_mcdi_mac_stats(efx, EFX_STATS_ENABLE, 0);
}

static void efx_mcdi_mac_stop_stats(struct efx_nic *efx)
{
	efx_mcdi_mac_stats(efx, EFX_STATS_DISABLE, 0);
}

#define EFX_MAC_STATS_WAIT_US 100
#define EFX_MAC_STATS_WAIT_ATTEMPTS 100

static void efx_mcdi_mac_pull_stats(struct efx_nic *efx)
{
	volatile __le64 *dma_stats = efx->stats_buffer.addr;
	int attempts = EFX_MAC_STATS_WAIT_ATTEMPTS;

	while (dma_stats[MC_CMD_MAC_GENERATION_END] !=
	       dma_stats[MC_CMD_MAC_GENERATION_START] &&
	       (attempts-- != 0))
		udelay(EFX_MAC_STATS_WAIT_US);
}

/* Adapted from v5 linux_net/nic.c */
static int efx_nic_alloc_buffer(struct efx_buffer *buffer, unsigned int len)
{
	buffer->addr = malloc_dma(len, EFX_BUF_ALIGN );
	if (!buffer->addr)
		return -ENOMEM;
	buffer->len = len;
	memset(buffer->addr, 0, len);
	buffer->dma_addr = virt_to_bus (buffer->addr);
	return 0;
}

static void efx_nic_free_buffer(struct efx_buffer *buffer)
{
	if (buffer->addr) {
		free_dma(buffer->addr, buffer->len);
		buffer->addr = NULL;
	}
}

/* Exported gPXE API for statistics */
void efx_stats_init(struct efx_nic *efx)
{
	/* Allocate buffer for stats */
	int rc = efx_nic_alloc_buffer(&efx->stats_buffer,
				      MC_CMD_MAC_NSTATS * sizeof(u64));
	if (rc)
		return;

	DBGC(efx, "stats buffer at %llx (virt %p phys %llx)\n",
	     (u64)efx->stats_buffer.dma_addr, efx->stats_buffer.addr,
	     (u64)virt_to_phys(efx->stats_buffer.addr));
	efx_mcdi_mac_stats(efx, EFX_STATS_DISABLE, 1);
	efx_mcdi_mac_start_stats(efx);
	efx_mcdi_mac_pull_stats(efx);
}

void efx_stats_fini(struct efx_nic *efx)
{
	if (!efx->stats_buffer.addr)
		return;

	efx_mcdi_mac_stop_stats(efx);
	efx_nic_free_buffer(&efx->stats_buffer);
}

void efx_stats_print(struct efx_nic *efx, const struct efx_hw_stat *dsc,
		     unsigned int dsc_count)
{
	struct efx_buffer *stats_buffer = &efx->stats_buffer;
	volatile __le64 *dma_stats = stats_buffer->addr;
	unsigned int	i;

	if (!dma_stats) {
		DBGC(efx, "%s: No stats\n", efx->netdev->name);
		return;
	}

	efx_mcdi_mac_pull_stats(efx);
	if (dma_stats[MC_CMD_MAC_GENERATION_END] !=
	    dma_stats[MC_CMD_MAC_GENERATION_START]) {
		DBGC(efx, "%s: Pulling stats incomplete\n", efx->netdev->name);
		//return;
	}

	/* To dump everything use:
	 * dbg_hex_dump_da(stats_buffer->dma_addr, dma_stats, stats_buffer->len);
	 */
	for (i=0; i < dsc_count; i++) {
		printf("%s = 0x%llx%c", dsc[i].name,
		       dma_stats[dsc[i].offset/sizeof(__le64)],
		       (i%2 ? '\n' : ' '));
	}
	if (i%2)
		printf("\n");
}
#endif

/*	MCDI access
 */
int efx_get_port_modes(struct efx_nic *efx, struct mcdi_port_modes *modes)
{
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_PORT_MODES_OUT_LEN);
	size_t outlen;
	int rc;

	BUILD_BUG_ON(MC_CMD_GET_PORT_MODES_IN_LEN != 0);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_GET_PORT_MODES,
				 NULL, 0,
				 outbuf, MC_CMD_GET_PORT_MODES_OUT_LEN,
				 &outlen, false);
	if (rc != 0)
		return rc;

	/* The output length changed from 4 to 8 and later to 12 bytes. We are
	 * only interested in the first 2 words here.
	 */
	if(outlen < (MC_CMD_GET_PORT_MODES_OUT_MODES_OFST + 4))
		return -EMSGSIZE;
	modes->modes = MCDI_DWORD(outbuf, GET_PORT_MODES_OUT_MODES);

	if(outlen >= (MC_CMD_GET_PORT_MODES_OUT_DEFAULT_MODE_OFST + 4))
		modes->dflt = MCDI_DWORD(outbuf,
					 GET_PORT_MODES_OUT_DEFAULT_MODE);
	else
		modes->dflt = TLV_PORT_MODE_DEFAULT;

	return 0;
}

#ifdef MISC_CMD

static MCDI_DECLARE_BUF(efx_mcdi_outbuf, 256);

int efx_read32(struct efx_nic *efx, unsigned int *value, unsigned int reg)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_READ32_IN_LEN);
	int rc;

	MCDI_SET_DWORD(inbuf, READ32_IN_ADDR, reg);
	MCDI_SET_DWORD(inbuf, READ32_IN_NUMWORDS, 1);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_READ32,
				 inbuf, sizeof(inbuf),
				 efx_mcdi_outbuf, sizeof(efx_mcdi_outbuf),
				 NULL, false);
	if (rc)
		return rc;

	*value = MCDI_DWORD(efx_mcdi_outbuf, READ32_OUT_BUFFER);
	return 0;
}

int efx_iread(struct efx_nic *efx, unsigned int *value, unsigned int reg)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_CSR_READ32_IN_LEN);
	int rc;

	MCDI_SET_DWORD(inbuf, CSR_READ32_IN_ADDR, reg);
	MCDI_SET_DWORD(inbuf, CSR_READ32_IN_STEP, 1);
	MCDI_SET_DWORD(inbuf, CSR_READ32_IN_NUMWORDS, 1);

	rc = efx->type->mcdi_rpc(efx, MC_CMD_CSR_READ32,
				 inbuf, sizeof(inbuf),
				 efx_mcdi_outbuf, sizeof(efx_mcdi_outbuf),
				 NULL, false);
	if (rc)
		return rc;

	*value = MCDI_DWORD(efx_mcdi_outbuf, CSR_READ32_OUT_BUFFER);
	return 0;
}

int efx_mccmd(struct efx_nic *efx, struct net_ioctl_mccmd *mccmd)
{
	size_t outlen_actual;
	int rc;

	rc = efx->type->mcdi_rpc(efx, mccmd->cmd,
				 (const efx_dword_t *) mccmd->buffer,
				 mccmd->inlen,
				 efx_mcdi_outbuf, sizeof(efx_mcdi_outbuf),
				 &outlen_actual, false);

	mccmd->buffer = (u32 *) efx_mcdi_outbuf;
	mccmd->outlen = outlen_actual;
	return rc;
}
#endif
#endif	/* SFC_IPXE */
