/**************************************************************************
 *
 * SPDX-License-Identifier: GPL-2.0+
 *
 * NVRAM driver for Solarflare network cards
 *
 * Copyright 2016 Solarflare Communications Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 (GPL), incorporated herein by
 * reference.  Drivers based on or derived from this code fall under
 * the GPL and must retain the authorship, copyright and license
 * notice.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 *
 ***************************************************************************/

#ifndef EFX_NVRAM_H
#define EFX_NVRAM_H

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdint.h>

#include "efx_common.h"
#include <config/general.h>

/** efx_nvram_info - get information about a partition
 *
 * efx  : NIC structure
 * type : Partition type
 *
 * Returns: NULL on failure, opaque pointer to partition info otherwise.
 */
struct efx_partition *efx_nvram_info(struct efx_nic *efx, uint32_t type);

/** efx_partition_free - free structure returned by @efx_nvram_info
 *
 * partition : Partition structure, returned by efx_nvram_info
 */
void efx_partition_free(struct efx_partition *partition);

/** efx_partition_erasesize - get erase block size for a partition
 *
 * partition : Partition structure, returned by efx_nvram_info
 */
int efx_partition_erasesize(struct efx_partition *partition);

/** efx_partition_size - get total size for a partition
 *
 * partition : Partition structure, returned by efx_nvram_info
 */
int efx_partition_size(struct efx_partition *partition);

/** efx_nvram_start - prepare a partition for erase or write.
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 *
 * Multiple calls to @efx_nvram_write_locked and @efx_nvram_erase_locked
 * can be made after this has succeeded.
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_start(struct efx_nic *efx, struct efx_partition *partition);

/** efx_nvram_finish - release a partition prepared with efx_nvram_start
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_finish(struct efx_nic *efx, struct efx_partition *partition);

/** efx_nvram_read - read from a partition
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 * offset    : Offset from start of partition
 * len       : Amount to read
 * buf       : Buffer to read in to; must be at least #len bytes
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_read(struct efx_nic *efx, struct efx_partition *partition,
		   size_t offset, size_t len, void *buf);

/** efx_nvram_erase_locked - erase a section of a partition with the lock held
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 * start     : Starting offset to erase
 * size      : Amount to erase
 *
 * @efx_nvram_start must be called before use, and @efx_nvram_finish after use.
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_erase_locked(struct efx_nic *efx, struct efx_partition *partition,
			   unsigned int start, size_t size);

/** efx_nvram_erase - erase a section of a partition
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 * start     : Starting offset to erase
 * size      : Amount to erase
 *
 * This calls @efx_nvram_start / @efx_nvram_finish around the operation.
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_erase(struct efx_nic *efx, struct efx_partition *partition,
		    unsigned int start, size_t size);

/** efx_nvram_write_locked - write to a partition with the lock held
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 * offset    : Offset from start of partition
 * len       : Amount to write
 * buf       : Buffer to write from; must be at least #len bytes
 *
 * @efx_nvram_start must be called before use, and @efx_nvram_finish after use.
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_write_locked(struct efx_nic *efx, struct efx_partition *partition,
			   size_t offset, size_t len, const void *buf);

/** efx_nvram_write - write to a partition
 *
 * efx       : NIC structure
 * partition : Partition structure, returned by efx_nvram_info
 * offset    : Offset from start of partition
 * len       : Amount to write
 * buf       : Buffer to write from; must be at least #len bytes
 *
 * This calls @efx_nvram_start / @efx_nvram_finish around the operation.
 *
 * Returns: 0 on success, errcode on failure.
 */
int efx_nvram_write(struct efx_nic *efx, struct efx_partition *partition,
		    size_t offset, size_t len, const void *buf);

/** efx_nvlog_init - set up logging to NVRAM
 *
 * efx            : NIC structure
 * partition_type : Partition type to log to
 *
 * The entire partition will be used.
 */
int efx_nvlog_init(struct efx_nic *efx, int partition_type);

/** efx_nvlog_packet - log a packet to NVRAM
 *
 * efx      : NIC structure
 * data     : Packet data
 * data_len : Length (bytes) of packet data
 */
void efx_nvlog_packet(struct efx_nic *efx, const u8 *data, u32 data_len);

#endif
