/**************************************************************************
 * SPDX-License-Identifier: GPL-2.0+
 *
 *
 * gPXE device driver for Solarflare Communications SFC90X0 devices
 *
 * Written by Steve Hodgson <shodgson@solarflare.com>
 *
 * Copyright Level 5 Networks Inc. 2005
 * Copyright 2006-2010 Solarflare Communications Inc.
 *
 * This software may be used and distributed according to the terms of
 * the GNU General Public License version 2 (GPL), incorporated herein by
 * reference.  Drivers based on or derived from this code fall under
 * the GPL and must retain the authorship, copyright and license
 * notice.
 *
 **************************************************************************
 */
FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <byteswap.h>
#include <gpxe/io.h>
#include <gpxe/pci.h>
#include <gpxe/malloc.h>
#include <gpxe/ethernet.h>
#include <gpxe/iobuf.h>
#include <gpxe/netdevice.h>
#include <gpxe/nvo.h>
#include <gpxe/nvs.h>
#include <drivers/net/mii.h>
#include <gpxe/dhcp.h>
#include "efx.h"
#include "efx_bitfield.h"
#include "efx_regs.h"
#include "mc_driver_pcol.h"

#include <gpxe/hp_ocsd.h>
#include "sfc/sfc_hp_ocsd.h"
#include <gpxe/uaccess.h>
#include <gpxe/instance.h>
#include <string.h>


struct siena_nic {
	struct efx_nic efx;

	/* NVO */
	struct nvo_fragment nvo_fragments[2];
	struct nvo_block nvo;
	struct nvs_device nvs;

	/* PHY information */
	unsigned int phy_cap_mask;
	unsigned int phy_cap;
	unsigned int phy_flags;
	char phy_name[20];
	int link_mode;
	unsigned long link_poll_timer;

	struct {
		/* Common payload for all MCDI requests */
		efx_dword_t payload[0xfc / sizeof(efx_dword_t)];

		unsigned int pdu;
		unsigned int doorbell;
		unsigned int seqno;
	} mcdi;
};

#define EFX_TRACE(...) DBGP(__VA_ARGS__)
#define EFX_LOG(...) DBG(__VA_ARGS__)
#define EFX_ERR(...) DBG(__VA_ARGS__)

#define MIN(a, b) ( ( a ) < ( b ) ? ( a ) : ( b ) )

#define SIENA_NVRAM_CHUNK_SIZE 0x80
#define SIENA_NVS_MAX_LENGTH 0x1000

/*******************************************************************************
 *
 *
 * MCDI transport
 *
 *
 *******************************************************************************/

static void siena_mcdi_init ( struct siena_nic *siena )
{
	siena->mcdi.pdu = FR_CZ_MC_TREG_SMEM + ( siena->efx.port
						 ? 0x108 : 0x8 );
	siena->mcdi.doorbell = FR_CZ_MC_TREG_SMEM + ( siena->efx.port
						      ? 0x4 : 0x0 );
}

static void siena_mcdi_copyin ( struct siena_nic *siena, unsigned cmd,
				uint8_t *inbuf, size_t inlen )
{
	unsigned int i;
	efx_dword_t hdr;
	uint32_t seqno;

	seqno = siena->mcdi.seqno & MCDI_SEQ_MASK;
	EFX_POPULATE_DWORD_6 ( hdr,
			       MCDI_HEADER_RESPONSE, 0,
			       MCDI_HEADER_RESYNC, 1,
			       MCDI_HEADER_CODE, cmd,
			       MCDI_HEADER_DATALEN, inlen,
			       MCDI_HEADER_SEQ, seqno,
			       MCDI_HEADER_XFLAGS, 0 );
	efx_writel ( &siena->efx, &hdr, siena->mcdi.pdu );

	for ( i = 0 ; i < inlen ; i += 4 )
		efx_writel ( &siena->efx,
			      ( efx_dword_t *) ( inbuf + i ),
			      siena->mcdi.pdu + 4 + i);

	wmb();
	efx_writel ( &siena->efx, &hdr, siena->mcdi.doorbell );
}

static void siena_mcdi_copyout ( struct siena_nic *siena,
				 uint8_t *outbuf, size_t outlen )
{
	unsigned int i;

	for ( i = 0 ; i < outlen ; i += 4 )
		efx_readl ( &siena->efx,
			     ( efx_dword_t * ) ( outbuf + i ),
			     siena->mcdi.pdu + 4 + i );
}

static int siena_mcdi_poll ( struct siena_nic *siena, size_t *resplenp )
{
	unsigned int resplen, respseq, respcmd __unused, error;
	unsigned long finish;
	efx_dword_t dword;
	int rc;

	/* Spin for up to 5s, polling every 10ms */
	finish = currticks ( ) + ( 5 * ticks_per_sec ( ) );
	while ( 1 ) {
		mdelay ( 10 );
		efx_readl ( &siena->efx, &dword, siena->mcdi.pdu );

		/* Check for reset or command pending */
		if ( EFX_DWORD_FIELD ( dword, EFX_DWORD_0 ) != 0xffffffff &&
		     EFX_DWORD_FIELD ( dword, MCDI_HEADER_RESPONSE ) )
			break;

		if ( currticks ( ) >= finish )
			return -ETIMEDOUT;
	}

	resplen = EFX_DWORD_FIELD ( dword, MCDI_HEADER_DATALEN );
	respseq = EFX_DWORD_FIELD ( dword, MCDI_HEADER_SEQ );
  respcmd = EFX_DWORD_FIELD ( dword,  MCDI_HEADER_CODE );
	error = EFX_DWORD_FIELD ( dword, MCDI_HEADER_ERROR );

	if ( error && resplen == 0 ) {
		EFX_ERR ( "MC rebooted" );
		return -EIO;
	} else if ( ( respseq ^ siena->mcdi.seqno ) & MCDI_SEQ_MASK ) {
		EFX_ERR ( "MC response mismatch rxseq 0x%x txseq 0x%x",
			  respseq, siena->mcdi.seqno );
		return -EIO;
	} else if ( error ) {
		efx_readl ( &siena->efx, &dword,
			    siena->mcdi.pdu + 4 + MC_CMD_ERR_CODE_OFST );
		rc = EFX_DWORD_FIELD ( dword, EFX_DWORD_0 );
		EFX_ERR ( "MC response rc %d", rc );

		switch ( rc ) {
		case MC_CMD_ERR_ENOENT:
			return -ENOENT;
		case MC_CMD_ERR_EINTR:
			return -EINTR;
		case MC_CMD_ERR_EACCES:
			return -EACCES;
		case MC_CMD_ERR_EBUSY:
			return -EBUSY;
		case MC_CMD_ERR_EINVAL:
			return -EINVAL;
		case MC_CMD_ERR_EDEADLK:
			return -EDEADLK;
		case MC_CMD_ERR_ENOSYS:
			return -ENOSYS;
		case MC_CMD_ERR_ETIME:
			return -ETIME;
		default:
			return -EIO;
		}
	}

	*resplenp = resplen;

	return 0;
}

int _siena_mcdi ( struct efx_nic *efx, unsigned cmd,
		  const efx_dword_t *inbuf, size_t inlen,
		  efx_dword_t *outbuf, size_t outlen,
		  size_t *outlen_actual, bool quiet )
{
	int rc;
	struct siena_nic *siena = (struct siena_nic *) efx;
	size_t local_outlen_actual;
	if (outlen_actual == NULL)
		outlen_actual = &local_outlen_actual;

	if (inlen > sizeof(siena->mcdi.payload))
		EFX_ERR ("Invalid inlen %zd for MCDI cmd 0x%x\n", inlen, cmd);

	++siena->mcdi.seqno;
	siena_mcdi_copyin ( siena, cmd, (uint8_t*)inbuf, inlen );

	if ( ( rc = siena_mcdi_poll ( siena, outlen_actual ) ) != 0 ) {
		if (!quiet)
			EFX_ERR ( " to cmd 0x%x inlen %zd\n", cmd, inlen );
		return rc;
	}

	siena_mcdi_copyout ( siena, ( uint8_t* )outbuf, outlen );
	return 0;
}

static int siena_mcdi ( struct siena_nic *siena, unsigned int cmd,
			size_t inlen, size_t min_outlen )
{
	size_t uninitialized_var(resplen);
	return _siena_mcdi ( &siena->efx, cmd,
			     ( const efx_dword_t * ) &siena->mcdi.payload, inlen,
			     ( efx_dword_t * ) &siena->mcdi.payload, min_outlen,
			     &resplen, false );
}

/*******************************************************************************
 *
 *
 * NVRAM support
 *
 *
 *******************************************************************************/

static int siena_nvram_types ( struct siena_nic *siena, unsigned int *types )
{
	int rc;

	rc = siena_mcdi ( siena, MC_CMD_NVRAM_TYPES,
			  MC_CMD_NVRAM_TYPES_IN_LEN,
			  MC_CMD_NVRAM_TYPES_OUT_LEN );
        if ( rc != 0 )
		return rc;

	*types = MCDI_DWORD( siena->mcdi.payload, NVRAM_TYPES_OUT_TYPES );
	return 0;
}

static int siena_nvram_read ( struct siena_nic *siena, unsigned int type,
			      unsigned int address, void *data, size_t len )
{
	unsigned int frag_len;
	int rc;

	while ( len ) {
		frag_len = MIN ( len, SIENA_NVRAM_CHUNK_SIZE );

		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_READ_IN_TYPE, type );
		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_READ_IN_OFFSET, address );
		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_READ_IN_LENGTH, frag_len );

		rc = siena_mcdi ( siena, MC_CMD_NVRAM_READ,
				  MC_CMD_NVRAM_READ_IN_LEN,
				  MC_CMD_NVRAM_READ_OUT_LEN ( frag_len ) );
		if ( rc != 0 )
			return rc;

		memcpy ( data, MCDI_PTR ( siena->mcdi.payload, NVRAM_READ_OUT_READ_BUFFER ),
			 frag_len );

		data += frag_len;
		address += frag_len;
		len -= frag_len;
	}

	return 0;
}

static int siena_nvram_update_start ( struct siena_nic *siena,
				      unsigned int type )
{
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_UPDATE_START_IN_TYPE, type );

	return siena_mcdi ( siena, MC_CMD_NVRAM_UPDATE_START,
			    MC_CMD_NVRAM_UPDATE_START_IN_LEN, 0 );
}

static int siena_nvram_erase ( struct siena_nic *siena, unsigned int type,
			       unsigned int length )
{
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_ERASE_IN_TYPE, type );
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_ERASE_IN_OFFSET, 0 );
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_ERASE_IN_LENGTH, length );

	return siena_mcdi ( siena, MC_CMD_NVRAM_ERASE,
			    MC_CMD_NVRAM_ERASE_IN_LEN, 0 );
}

static int siena_nvram_write ( struct siena_nic *siena, unsigned int type,
			       unsigned int address, const void *data,
			       size_t len )
{
	unsigned int frag_len;
	int rc;

	while ( len ) {
		frag_len = MIN ( len, SIENA_NVRAM_CHUNK_SIZE );

		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_WRITE_IN_TYPE, type );
		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_WRITE_IN_OFFSET, address );
		MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_WRITE_IN_LENGTH, frag_len );

		memcpy ( MCDI_PTR ( siena->mcdi.payload,
				    NVRAM_WRITE_IN_WRITE_BUFFER ),
			 data, frag_len );

		rc = siena_mcdi ( siena, MC_CMD_NVRAM_WRITE,
				  MC_CMD_NVRAM_WRITE_IN_LEN ( frag_len ), 0 );
		if ( rc != 0 )
			return rc;

		data += frag_len;
		address += frag_len;
		len -= frag_len;
	}

	return 0;
}

static void siena_nvram_update_finish ( struct siena_nic *siena, unsigned int type )
{
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_UPDATE_FINISH_IN_TYPE, type );

	( void ) siena_mcdi ( siena, MC_CMD_NVRAM_UPDATE_FINISH,
			      MC_CMD_NVRAM_UPDATE_FINISH_IN_LEN, 0 );
}

static int siena_nvram_test ( struct siena_nic *siena, unsigned int type )
{
	unsigned int result;
	int rc;

	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_TEST_IN_TYPE, type );

	rc = siena_mcdi ( siena, MC_CMD_NVRAM_TEST,
			  MC_CMD_NVRAM_TEST_IN_LEN,
			  MC_CMD_NVRAM_TEST_OUT_LEN );
	if ( rc != 0 ) {
		EFX_LOG ( "nvram_test(%d) %s\n", type, strerror ( rc ) );
		return rc;
	}

	result = MCDI_DWORD ( siena->mcdi.payload, NVRAM_TEST_OUT_RESULT );

	if ( result == MC_CMD_NVRAM_TEST_PASS ) {
		EFX_LOG ( "nvram_test(%d) PASS\n", type );
		return 0;
	}

	if ( result == MC_CMD_NVRAM_TEST_FAIL ) {
		EFX_LOG ( "nvram_test(%d) FAIL\n", type );
		return -EFAULT;
	}

	if ( result == MC_CMD_NVRAM_TEST_NOTSUPP ) {
		EFX_LOG ( "nvram_test(%d) NOT SUPPORTED\n", type );
		return -ENOTSUP;
	}

	EFX_LOG ( "nvram_test(%d) EINVAL\n", type );
	return -EINVAL;
}

/*******************************************************************************
 *
 *
 * NVS transport
 *
 *
 *******************************************************************************/

static int siena_nvs_read ( struct nvs_device *nvs, unsigned int address,
			    void *data, size_t len )
{
	struct siena_nic *siena = container_of ( nvs, struct siena_nic, nvs );
	unsigned int type;

	type = siena->efx.port
	    ? MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT1
	    : MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT0;

	return siena_nvram_read ( siena, type, address, data, len );
}

static int siena_nvs_write ( struct nvs_device *nvs, unsigned int address,
			     const void *data, size_t len )
{
	struct siena_nic *siena = container_of ( nvs, struct siena_nic, nvs );
	unsigned int type;
	int rc;

	/* siena_nvo_probe() guarantees to setup siena->nvs_device such
	 * that nvs_write() is passed the entire configuration block.
	 */
	assert ( address == 0 );

	type = siena->efx.port
	    ? MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT1
	    : MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT0;

	if ( ( rc = siena_nvram_update_start ( siena, type ) ) != 0 )
		goto fail1;

	if ( ( rc = siena_nvram_erase ( siena, type, nvs->size ) ) != 0 )
		goto fail2;

	if ( ( rc = siena_nvram_write ( siena, type, address,
					data, len ) ) != 0 )
		goto fail3;

	siena_nvram_update_finish ( siena, type );

	return 0;

  fail3:
  fail2:
	siena_nvram_update_finish ( siena, type );
  fail1:
	return rc;
}

/*******************************************************************************
 *
 *
 * Hardware initialization
 *
 *
 *******************************************************************************/

static int siena_mac_addr ( struct siena_nic *siena, uint8_t *ll_addr )
{
	unsigned int offset;
	int rc;

	rc = siena_mcdi ( siena, MC_CMD_GET_BOARD_CFG, 0,
			  MC_CMD_GET_BOARD_CFG_OUT_LENMIN );
	if ( rc != 0 )
		return rc;

	offset = siena->efx.port
	    ? MC_CMD_GET_BOARD_CFG_OUT_MAC_ADDR_BASE_PORT1_OFST
	    : MC_CMD_GET_BOARD_CFG_OUT_MAC_ADDR_BASE_PORT0_OFST;
	
	memcpy ( ll_addr, _MCDI_PTR(siena->mcdi.payload, offset ), ETH_ALEN );

	return 0;
}

static int siena_probe_phy ( struct siena_nic *siena )
{
	int rc;

	rc = siena_mcdi ( siena, MC_CMD_GET_PHY_CFG, 0,
			  MC_CMD_GET_PHY_CFG_OUT_LEN );
	if ( rc != 0 )
		return rc;

	siena->phy_flags =
	    MCDI_DWORD ( siena->mcdi.payload, GET_PHY_CFG_OUT_FLAGS );
	siena->phy_cap_mask = siena->phy_cap =
	    MCDI_DWORD ( siena->mcdi.payload, GET_PHY_CFG_OUT_SUPPORTED_CAP );
	memcpy ( siena->phy_name,
		 MCDI_PTR ( siena->mcdi.payload, GET_PHY_CFG_OUT_NAME ),
		 MIN ( sizeof ( siena->phy_name ) - 1,
		       MC_CMD_GET_PHY_CFG_OUT_NAME_LEN ) );

	return 0;
}

static int siena_nvo_probe ( struct siena_nic *siena )
{
	unsigned int type, length, partn_length;
	struct nvs_device *nvs = &siena->nvs;
	struct nvo_block *nvo = &siena->nvo;
	struct nvo_fragment *fragments = siena->nvo_fragments;
	struct net_device *netdev = siena->efx.netdev;
	int rc;

	/* Determine length of BOOTCFG partn */
	type = siena->efx.port
	    ? MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT1
	    : MC_CMD_NVRAM_TYPE_EXP_ROM_CFG_PORT0;
	MCDI_SET_DWORD ( siena->mcdi.payload, NVRAM_INFO_IN_TYPE, type );

	rc = siena_mcdi ( siena, MC_CMD_NVRAM_INFO,
			  MC_CMD_NVRAM_INFO_IN_LEN,
			  MC_CMD_NVRAM_INFO_OUT_LEN );
	if ( rc != 0 )
		return rc;

	memset ( siena->nvo_fragments, '\0', sizeof ( siena->nvo_fragments ) );
	partn_length = MCDI_DWORD ( siena->mcdi.payload, NVRAM_INFO_OUT_SIZE );
	length = MIN ( partn_length, SIENA_NVS_MAX_LENGTH );

	/* Initialise the nvo_fragments */
	fragments[0].len = length;

	/* Initialise the nvs device such that siena_nvs_read() and
	 * siena_nvs_write() receive the entire configuration block.
	 * Store the total partition length so that we erase the
	 * entire sector.
	 */
	nvs->word_len_log2 = 0;
	nvs->block_size = nvs->size = partn_length;
	nvs->read = siena_nvs_read;
	nvs->write = siena_nvs_write;

	nvo_init ( nvo, nvs, fragments, &netdev->refcnt );

	return 0;
}

static int siena_attach ( struct siena_nic *siena, int attach )
{
	/* Set the PREBOOT flag to indicate later instances of attach should
	 * force an ENTITY RESET
	 */
	if (attach)
		attach |= 1 << MC_CMD_DRV_PREBOOT_LBN;

	MCDI_SET_DWORD ( siena->mcdi.payload,
			 DRV_ATTACH_IN_NEW_STATE, attach );
	MCDI_SET_DWORD ( siena->mcdi.payload, DRV_ATTACH_IN_UPDATE, 1 );

	return siena_mcdi ( siena, MC_CMD_DRV_ATTACH,
			    MC_CMD_DRV_ATTACH_IN_LEN, 0 );
}

static int siena_reset ( struct siena_nic *siena )
{
	MCDI_POPULATE_DWORD_1 ( siena->mcdi.payload, ENTITY_RESET_IN_FLAG,
				ENTITY_RESET_IN_FUNCTION_RESOURCE_RESET, 1 );
	return siena_mcdi ( siena, MC_CMD_ENTITY_RESET, 0, 0 );
}

static void siena_reboot(struct siena_nic *siena, unsigned mode)
{
	int	rc;

	MCDI_SET_DWORD(siena->mcdi.payload, REBOOT_MODE_IN_VALUE, mode);
	rc = siena_mcdi(siena, MC_CMD_REBOOT_MODE, MC_CMD_REBOOT_MODE_IN_LEN,
			MC_CMD_REBOOT_MODE_OUT_LEN);
	if ( rc != 0 )
		EFX_ERR("MC_CMD_REBOOT_MODE Failed: %s\n", strerror(rc));

	MCDI_SET_DWORD(siena->mcdi.payload, REBOOT_IN_FLAGS, 0);
	/* This will timeout. */
	(void) siena_mcdi(siena, MC_CMD_REBOOT, MC_CMD_REBOOT_IN_LEN, 0);
}

static void siena_setup_nic ( struct siena_nic *siena )
{
	struct efx_nic *efx = &siena->efx;
	efx_oword_t oword;

	/* Set descriptor caches (low enough to support FPGA and ASIC) */
	EFX_POPULATE_OWORD_1 ( oword, FRF_AZ_SRM_TX_DC_BASE_ADR, 0x37d0 );
	efx_write ( efx, &oword, FR_AZ_SRM_TX_DC_CFG );
	EFX_POPULATE_OWORD_1 ( oword, FRF_AZ_TX_DC_SIZE, 1 /* 16 descriptors */ );
	efx_write ( efx, &oword, FR_AZ_TX_DC_CFG );
	EFX_POPULATE_OWORD_1 ( oword, FRF_AZ_SRM_RX_DC_BASE_ADR, 0x1810 );
	efx_write ( efx, &oword, FR_AZ_SRM_RX_DC_CFG );
	EFX_POPULATE_OWORD_1 ( oword, FRF_AZ_RX_DC_SIZE, 2 /* 32 descriptors */ );
	efx_write ( efx, &oword, FR_AZ_RX_DC_CFG );

	/* Disable the ugly timer-based TX DMA backoff and allow TX DMA to be
	 * controlled by the RX FIFO fill level. Set arbitration to one pkt/Q.
	 * Squash TX of packets of 16 bytes or less
	 */
	efx_read ( efx, &oword, FR_AZ_TX_RESERVED );
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_RX_SPACER, 0xfe );
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_RX_SPACER_EN, 1 );
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_ONE_PKT_PER_Q, 1);
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_PUSH_EN, 0 );
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_DIS_NON_IP_EV, 1 );
	EFX_SET_OWORD_FIELD ( oword, FRF_AZ_TX_PREF_THRESHOLD, 2 );
	EFX_SET_OWORD_FIELD ( oword, FRF_BZ_TX_FLUSH_MIN_LEN_EN, 1 );
	efx_write ( efx, &oword, FR_AZ_TX_RESERVED );

	/* Setup RX.  Wait for descriptor is broken and must be disabled.
	 * RXDP recovery shouldn't be needed, but is. Enable RX_INGR_EN
	 * always.
	 */
	efx_read ( efx, &oword, FR_AZ_RX_CFG );
	EFX_SET_OWORD_FIELD ( oword, FRF_BZ_RX_USR_BUF_SIZE, 0x3000 >> 5 );
	EFX_SET_OWORD_FIELD ( oword, FRF_BZ_RX_DESC_PUSH_EN, 0 );
	EFX_SET_OWORD_FIELD ( oword, FRF_BZ_RX_INGR_EN, 1 );
	efx_write ( efx, &oword, FR_AZ_RX_CFG );
}

static int siena_cfg_mac ( struct siena_nic *siena )
{
	struct net_device *netdev = siena->efx.netdev;
	unsigned int fcntl;

	memset ( &siena->mcdi.payload, 0, MC_CMD_SET_MAC_IN_LEN );
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_MAC_IN_MTU,
			 EFX_MAC_FRAME_LEN ( ETH_FRAME_LEN ) );
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_MAC_IN_DRAIN, 0 );
	memcpy ( MCDI_PTR ( siena->mcdi.payload, SET_MAC_IN_ADDR ),
		 netdev->ll_addr, ETH_ALEN );
	MCDI_POPULATE_DWORD_1 ( siena->mcdi.payload, SET_MAC_IN_REJECT,
				SET_MAC_IN_REJECT_UNCST, 1 );

	/* If the PHY supports autnegotiation, then configure the MAC to
	 * match the negotiated settings. Otherwise force the MAC to
	 * TX and RX flow control. */
	if ( siena->phy_cap_mask & ( 1 << MC_CMD_PHY_CAP_AN_LBN ) )
		fcntl = MC_CMD_FCNTL_AUTO;
	else
		fcntl = MC_CMD_FCNTL_BIDIR;
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_MAC_IN_FCNTL, fcntl );

	return siena_mcdi ( siena, MC_CMD_SET_MAC,
			    MC_CMD_SET_MAC_IN_LEN, 0 );
}

#define MCDI_PORT_SPEED_CAPS   ((1 << MC_CMD_PHY_CAP_10HDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_10FDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_100HDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_100FDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_1000HDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_1000FDX_LBN) | \
				(1 << MC_CMD_PHY_CAP_10000FDX_LBN))

static int siena_cfg_phy ( struct siena_nic *siena )
{
	int rc, state;

	if (!(siena->phy_cap & MCDI_PORT_SPEED_CAPS)) {
		siena->phy_cap |= siena->phy_cap_mask & MCDI_PORT_SPEED_CAPS;
		DBGC(siena, "Corrected link capabilies to %x\n", siena->phy_cap);
	}

	MCDI_SET_DWORD ( siena->mcdi.payload, SET_LINK_IN_CAP, siena->phy_cap );
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_LINK_IN_FLAGS, 0 );
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_LINK_IN_LOOPBACK_MODE, 0 );
	MCDI_SET_DWORD ( siena->mcdi.payload, SET_LINK_IN_LOOPBACK_SPEED, 0 );

	if ( ( rc = siena_mcdi ( siena, MC_CMD_SET_LINK,
				 MC_CMD_SET_LINK_IN_LEN, 0 ) ) != 0 )
		return rc;

	if ( ( rc = siena_mcdi ( siena, MC_CMD_GET_PHY_STATE,
				 0, MC_CMD_GET_PHY_STATE_OUT_LEN ) ) != 0 )
		return rc;

	
	state = MCDI_DWORD ( siena->mcdi.payload, GET_PHY_STATE_OUT_STATE );
	if ( state != MC_CMD_PHY_STATE_OK ) {
		EFX_ERR ( "PHY has failed to start. Run sfupdate.\n" );
		return -ECONNREFUSED;
	}

	return 0;
}

/*******************************************************************************
 *
 *
 * Diagnostic test support
 *
 *
 *******************************************************************************/

static int siena_bist ( struct siena_nic *siena, unsigned int type )
{
	unsigned int result;
	int rc;

	MCDI_SET_DWORD ( siena->mcdi.payload, START_BIST_IN_TYPE, type );

	rc = siena_mcdi ( siena, MC_CMD_START_BIST,
			  MC_CMD_START_BIST_IN_LEN,
			  MC_CMD_START_BIST_OUT_LEN );
        if ( rc != 0 )
		return rc;

	do {
		rc = siena_mcdi ( siena, MC_CMD_POLL_BIST,
				  MC_CMD_POLL_BIST_IN_LEN, 4 );
		if ( rc != 0 ) {
			EFX_LOG ( "bist(%d) %s\n", type, strerror ( rc ) );
			return rc;
		}

		result = MCDI_DWORD ( siena->mcdi.payload, POLL_BIST_OUT_RESULT );

	} while ( result == MC_CMD_POLL_BIST_RUNNING );

	if ( result == MC_CMD_POLL_BIST_PASSED ) {
		EFX_LOG ( "bist(%d) PASSED\n", type );
		return 0;
	}
	if ( result == MC_CMD_POLL_BIST_FAILED ) {
		EFX_LOG ( "bist(%d) FAILED\n", type );
		return -EFAULT;
	}
	if ( result == MC_CMD_POLL_BIST_TIMEOUT ) {
		EFX_LOG ( "bist(%d) TIMEOUT\n", type );
		return -ETIMEDOUT;
	}
	EFX_LOG ( "bist(%d) EINVAL\n", type );
	return -EINVAL;
}

/*******************************************************************************
 *
 *
 * Link state handling
 *
 *
 *******************************************************************************/

static void
siena_update_link ( struct siena_nic *siena, unsigned int speed,
		    unsigned int flags )
{
	struct net_device *netdev = siena->efx.netdev;
	unsigned int fd, up;
	int link_mode;

	fd = flags & ( 1 << MC_CMD_GET_LINK_OUT_FULL_DUPLEX_LBN );
	up = flags & ( 1 << MC_CMD_GET_LINK_OUT_LINK_UP_LBN );

	link_mode = 0;
	if ( up ) {
		if ( speed == 10 )
			link_mode = fd ? LPA_10FULL : LPA_10HALF;
		else if ( speed == 100 )
			link_mode = fd ? LPA_100FULL : LPA_100HALF;
		else if ( speed == 1000 )
			link_mode = fd ? LPA_1000FULL : LPA_1000HALF;
		else if ( speed == 10000 && fd )
			link_mode = LPA_10000FULL;
	}

	siena->link_mode = link_mode;
	if ( link_mode != 0 )
		netdev_link_up ( netdev );
	else
		netdev_link_down ( netdev );
}

static int siena_poll_link ( struct siena_nic *siena )
{
	unsigned int speed, flags;
	int rc;

	rc = siena_mcdi ( siena, MC_CMD_GET_LINK,
			  MC_CMD_GET_LINK_IN_LEN,
			  MC_CMD_GET_LINK_OUT_LEN );
	if ( rc != 0 )
		return rc;

	speed = MCDI_DWORD ( siena->mcdi.payload, GET_LINK_OUT_LINK_SPEED );
	flags = MCDI_DWORD ( siena->mcdi.payload, GET_LINK_OUT_FLAGS );

	siena_update_link ( siena, speed, flags );

	return 0;
}

/*******************************************************************************
 *
 *
 * Netdevice operations
 *
 *
 *******************************************************************************/

static int
siena_open ( struct net_device *netdev )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	int rc;

	if ( ( rc = siena_attach ( siena, 1 ) ) != 0 )
		goto fail1;

	siena_setup_nic ( siena );

	/* Initialize data path */
	rc = efx_open ( netdev );
	if ( rc )
		goto fail3;

	if ( ( rc = siena_cfg_mac ( siena ) ) != 0 )
		goto fail4;

	if ( ( rc = siena_cfg_phy ( siena ) ) != 0 )
		goto fail5;

	/* Mark the link as down because we don't know when it went down */
	netdev_link_down ( netdev );

	efx_stats_init( &siena->efx );
	return 0;

  fail5:
  fail4:
	efx_close ( netdev );
  fail3:
	( void ) siena_attach ( siena, 0 );
  fail1:
	return rc;
}

static void
siena_poll ( struct net_device *netdev )
{
	struct siena_nic *siena = netdev_priv ( netdev );

	/* If called while already polling, return immediately */
	if (siena->efx.state & EFX_STATE_POLLING)
		return;
	siena->efx.state |= EFX_STATE_POLLING;

	/* Poll link state */
	if ( siena->link_poll_timer + ticks_per_sec() < currticks() ) {
		siena->link_poll_timer = currticks();
		siena_poll_link ( siena );
	}

	/* Poll data path */
	efx_poll ( netdev );

	siena->efx.state &= ~EFX_STATE_POLLING;
}

extern int reboot_needed;

static void
siena_close ( struct net_device *netdev )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	struct pci_device *pci = (struct pci_device *) netdev->dev;

	efx_stats_fini( &siena->efx );

	/* Stop datapath */
	efx_close ( netdev );

	/* Some menu changes require a cold reboot of the PCI adapter.
	 * Devices are removed in reverse order of their creation, so reset
	 * the PCI device and then the MC for function 0.
	 */
	if (reboot_needed && (PCI_FUNC(pci->devfn) == 0)) {
		siena_reboot(siena, MC_CMD_REBOOT_MODE_POR);
		pci_dev_reset(pci);
	} else {
		/* Warm reset hardware and detach */
		(void) siena_reset(siena);
		(void) siena_attach(siena, 0);
	}
}

static void
siena_flash_leds ( struct net_device *netdev, enum sf_flash_state state )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	static const uint8_t states[] = {
		[FLASH_NONE] = MC_CMD_LED_OFF,
		[FLASH_BLINK_ON] = MC_CMD_LED_OFF,
		[FLASH_BLINK_OFF] = MC_CMD_LED_DEFAULT,
	};

	MCDI_SET_DWORD ( siena->mcdi.payload, SET_ID_LED_IN_STATE, states[state] );
	( void ) siena_mcdi ( siena, MC_CMD_SET_ID_LED,
			      MC_CMD_SET_ID_LED_IN_LEN, 0 );
}

#define SIENA_MC_STATIC_CONFIG_MAGIC 0xbdcf5555
#define SIENA_MC_DYNAMIC_CONFIG_MAGIC 0xbdcfdddd

#define MAX( _a, _b )				\
	( ( _a ) > ( _b ) ? ( _a ) : ( _b ) )

static int
siena_find_vpd ( struct siena_nic *siena, unsigned int partn,
		 uint32_t magic, enum sf_vpd_tag tag,
		 uint16_t keyword, uint8_t *buf, size_t len )
{
	uint8_t *contents;
	uint8_t csum;
	unsigned int pos;
	int rc;

	/* Conviniently the important bits of the static and
	 * dynamic configuration areas are the same */
	struct {
		uint32_t	magic;
		uint16_t	length;
		uint8_t		version;
		uint8_t		csum;
		uint32_t	vpd_offset;
		uint32_t	vpd_length;
	} hdr;

	/* Select the correct port */
	if ( siena->efx.port )
		++partn;

	/* Read the hdr in */
	rc = siena_nvram_read ( siena, partn, 0, &hdr, sizeof ( hdr ) );
	if ( rc < 0 )
		return rc;
	if ( hdr.magic != magic )
		return -ENOENT;

	/* Allocate a buffer big enough for the entire header and vpd sector */
	contents = malloc ( MAX ( hdr.length, hdr.vpd_length ) );
	if ( contents == NULL )
		return -ENOMEM;

	/* Verify the header */
	rc = siena_nvram_read ( siena, partn, 0, contents, hdr.length );
	if ( rc < 0 )
		goto out;
	for ( csum = 0, pos = 0; pos < hdr.length; pos++ )
		csum += contents[pos];
	if ( csum != 0 ) {
		rc = -EFAULT;
		goto out;
	}

	/* Read the VPD */
	if ( ( rc = siena_nvram_read ( siena, partn, hdr.vpd_offset, contents,
				       hdr.vpd_length ) ) < 0 )
		goto out;

	rc = efx_vpd ( contents, hdr.vpd_length, tag, keyword, buf, len );

  out:
	free ( contents );

	return rc;
}

static int
siena_vpd ( struct net_device *netdev, enum sf_vpd_tag tag,
	    uint16_t keyword, uint8_t *buf, size_t len )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	int rc;

	if ( ( rc = siena_find_vpd ( siena, MC_CMD_NVRAM_TYPE_STATIC_CFG_PORT0,
				     SIENA_MC_STATIC_CONFIG_MAGIC, tag,
				     keyword, buf, len ) ) > 0 )
		return rc;

	return siena_find_vpd ( siena, MC_CMD_NVRAM_TYPE_DYNAMIC_CFG_PORT0,
				SIENA_MC_DYNAMIC_CONFIG_MAGIC, tag, keyword,
				buf, len );
}

static int
siena_has_option_rom ( struct net_device *netdev )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	uint32_t magic;
	int rc;

	if ( ( rc = siena_nvram_read ( siena, MC_CMD_NVRAM_TYPE_EXP_ROM,
				       0, &magic, sizeof ( magic ) ) ) != 0 )
		/* Assume there is a ROM */
		return 1;

	if ( magic == 0xB0070102 || (uint16_t)magic == 0xaa55 )
		return 1;

	return 0;
}

static int
siena_run_diagnostic ( struct net_device *netdev, unsigned int diagnostic )
{
	struct siena_nic *siena = netdev_priv ( netdev );
	unsigned int type, types;
	int rc;

	if ( diagnostic == 0 ) {
		/* BIST */
		efx_dword_t phy_flags = { { siena->phy_flags } };

		EFX_LOG ( "phy_flags = %X\n", siena->phy_flags );

		types = ( ( 1 << MC_CMD_BPX_SERDES_BIST  ) |
			  ( 1 << MC_CMD_MC_LOOPBACK_BIST ) );

		if ( EFX_DWORD_FIELD ( phy_flags, MC_CMD_GET_PHY_CFG_OUT_BIST_CABLE_SHORT ) )
			types |= ( 1 << MC_CMD_PHY_BIST_CABLE_SHORT );

		if ( EFX_DWORD_FIELD ( phy_flags, MC_CMD_GET_PHY_CFG_OUT_BIST_CABLE_LONG ) )
			types |= ( 1 << MC_CMD_PHY_BIST_CABLE_LONG );

		if ( EFX_DWORD_FIELD ( phy_flags, MC_CMD_GET_PHY_CFG_OUT_BIST ) )
			types |= ( 1 << MC_CMD_PHY_BIST );

		if ( ! EFX_DWORD_FIELD ( phy_flags, MC_CMD_GET_PHY_CFG_OUT_PRESENT ) ) {
			EFX_LOG ( "bist: PHY not present\n" );
		}

		for ( type = 0; types != 0; types >>= 1, type++ ) {

			if ( ( types & 1 ) == 0 )
				continue;

			rc = siena_bist ( siena, type );
			if ( ( rc != 0 ) && ( rc != -EINVAL ) ) {
				EFX_LOG ( "bist(%d) = %s\n",
					  type, strerror ( rc ) );
				return rc;
			}
		}
		return 0;
	}
	if ( diagnostic == 1 ) {
		/* NVRAM_TEST */
		unsigned int nvram_no_test_types =
		    ( 1 << MC_CMD_NVRAM_TYPE_MC_FW_BACKUP ) |
		    ( 1 << MC_CMD_NVRAM_TYPE_LOG );

		/* Query for supported NVRAM partitions */
		rc = siena_nvram_types ( siena, &types );
		if ( rc != 0 )
			return -ENOTSUP;

		EFX_LOG ( "nvram_types() = %X\n", types );

		/* Do not test some partitions even if present */
		types  &= ~nvram_no_test_types;

		EFX_LOG ( "nvram test types = %X\n", types );

		/* Test each supported partition */
		for ( type = 0; types != 0; types >>= 1, type++ ) {

			if ( ( types & 1 ) == 0 )
				continue;

			rc = siena_nvram_test ( siena, type );

			if ( ( rc != 0 ) && ( rc != -EINVAL ) ) {
				EFX_LOG ( "nvram_test(%d) = %s\n",
					  type, strerror ( rc ) );
				return rc;
			}
		}
		return 0;
	}

	/* unsupported diagnostic test id */
	return -ENOTSUP;
}

static int
siena_option_supported ( struct net_device *netdev, int tag )
{
	switch (tag) {
	case DHCP_EB_SRIOV_ENABLED:
	case DHCP_EB_VF_COUNT:
	case DHCP_EB_VF_MSIX_LIMIT:
		return 1;
	}
	return efx_common_option_supported(netdev, tag);
}

#define SIENA_DMA_STAT(mcdi_name)	\
	{ # mcdi_name, 64, 8 * MC_CMD_MAC_ ## mcdi_name }

static const struct efx_hw_stat siena_stats[] = {
	SIENA_DMA_STAT(TX_PKTS),
	SIENA_DMA_STAT(TX_BAD_BYTES),
	SIENA_DMA_STAT(RX_PKTS),
	SIENA_DMA_STAT(RX_GOOD_PKTS),
	SIENA_DMA_STAT(RX_PAUSE_PKTS),
	SIENA_DMA_STAT(RX_CONTROL_PKTS),
	SIENA_DMA_STAT(RX_UNICAST_PKTS),
	SIENA_DMA_STAT(RX_MULTICAST_PKTS),
	SIENA_DMA_STAT(RX_BROADCAST_PKTS),
	SIENA_DMA_STAT(RX_UNDERSIZE_PKTS),
	SIENA_DMA_STAT(RX_BAD_BYTES),
	SIENA_DMA_STAT(RX_UNDERSIZE_PKTS),
	SIENA_DMA_STAT(RX_GTJUMBO_PKTS),
	SIENA_DMA_STAT(RX_BAD_FCS_PKTS),
	SIENA_DMA_STAT(RX_OVERFLOW_PKTS),
	SIENA_DMA_STAT(RX_FALSE_CARRIER_PKTS),
	SIENA_DMA_STAT(RX_SYMBOL_ERROR_PKTS),
	SIENA_DMA_STAT(RX_ALIGN_ERROR_PKTS),
	SIENA_DMA_STAT(RX_LENGTH_ERROR_PKTS),
	SIENA_DMA_STAT(RX_INTERNAL_ERROR_PKTS),
	SIENA_DMA_STAT(RX_JABBER_PKTS),
	SIENA_DMA_STAT(RX_NODESC_DROPS),
	SIENA_DMA_STAT(RX_MATCH_FAULT),
};

static int siena_ioctl ( struct net_device *netdev, enum net_ioctl_cmd cmd,
			 unsigned long arg )
{
	struct siena_nic *siena = netdev_priv(netdev);
#ifdef MISC_CMD
	struct net_ioctl_peek *argp = (struct net_ioctl_peek *) arg;
#endif
	int rc = -ENOTSUP;
	int *intp;

	switch (cmd) {
	case NET_GET_PORT_NUMBER:
		intp = (int *) arg;
		*intp = siena->efx.port;
		rc = 0;
		break;
#ifdef MISC_CMD
	case NET_PEEK:
		argp->value = _efx_readl(&siena->efx, argp->address);
		rc = 0;
		break;
	case NET_READ32:
		rc = efx_read32(&siena->efx, &argp->value, argp->address);
		break;
	case NET_IREAD:
		rc = efx_iread(&siena->efx, &argp->value, argp->address);
		break;
	case NET_MCCMD:
		rc = efx_mccmd(&siena->efx, (struct net_ioctl_mccmd *) arg);
		break;
#endif
	case NET_STATS:
		efx_stats_print(&siena->efx, siena_stats,
				sizeof(siena_stats)/sizeof(siena_stats[0]));
	default:
		break;
	}
	return rc;
}

/*******************************************************************************
 *
 *
 * Public operations
 *
 *
 *******************************************************************************/


static struct net_device_operations siena_operations = {
        .open			= siena_open,
        .close			= siena_close,
        .transmit		= efx_transmit,
        .poll			= siena_poll,
        .irq			= efx_irq,
	.ioctl			= siena_ioctl,
	.sf_flash_leds		= siena_flash_leds,
	.sf_vpd			= siena_vpd,
	.sf_suspend		= efx_suspend,
	.sf_resume		= efx_resume,
	.sf_has_option_rom	= siena_has_option_rom,
	.sf_run_diagnostic	= siena_run_diagnostic,
	.sf_get_boot_type       = efx_get_boot_type,
	.sf_option_supported	= siena_option_supported,
};


static int
siena_probe ( struct pci_device *pci,
	      const struct pci_device_id *id __unused )
{
	struct net_device *netdev;
	struct siena_nic *siena;
	struct efx_nic *efx;
	efx_oword_t oword;
	unsigned int port;
	int rc;

	/* Create the network adapter */
	netdev = alloc_etherdev ( sizeof ( struct siena_nic ) );
	if ( ! netdev ) {
		rc = -ENOMEM;
		goto fail1;
	}

	/* Initialise the network adapter, and initialise private storage */
	netdev_init ( netdev, &siena_operations );
	pci_set_drvdata ( pci, netdev );
	netdev->dev = &pci->dev;
	netdev->state |= NETDEV_IRQ_UNSUPPORTED;

	siena = netdev_priv ( netdev );
	memset ( siena, 0, sizeof ( *siena ) );
	efx = &siena->efx;

	efx->type = &siena_nic_type;

	/* Initialise efx datapath */
	efx_probe ( netdev, EFX_SIENA );

	/* Initialise MCDI */
	efx_read ( efx, &oword, FR_AZ_CS_DEBUG );
	port = EFX_OWORD_FIELD ( oword, FRF_CZ_CS_PORT_NUM );
	if ( port == 0 )
		siena->efx.port = 0;
	else
		siena->efx.port = port - 1;
	siena_mcdi_init ( siena );

	rc = siena_reset ( siena );
	if ( rc != 0 )
		goto fail2;

#ifdef __i386__
	if ( is_hp ( pci ) )
		hp_ocsd_start( pci );
#endif

	if ( ( rc = siena_mac_addr ( siena, netdev->ll_addr ) ) != 0 )
		goto fail2;

	if ( ( rc = siena_probe_phy ( siena ) ) != 0 )
		goto fail3;

	if ( ( rc = siena_nvo_probe ( siena ) ) != 0 )
		goto fail4;

	if ( ( rc = register_netdev ( netdev ) ) != 0 )
		goto fail5;

	rc = register_nvo ( &siena->nvo, netdev_settings ( netdev ), NULL, 0 );
	if ( rc != 0 )
		goto fail6;

	efx_register_clp ( pci );

	return 0;

  fail6:
	unregister_netdev ( netdev );
  fail5:
  fail4:
  fail3:
  fail2:
	efx_remove ( netdev );
	netdev_put ( netdev );
  fail1:
	return rc;
}

static void
siena_remove ( struct pci_device *pci )
{
	struct net_device *netdev = pci_get_drvdata ( pci );
	struct siena_nic *siena = netdev_priv ( netdev );

	efx_unregister_clp ( pci );

	unregister_nvo ( &siena->nvo );

	/* The netdevice might still be open, so unregister it now
	 * before ripping stuff out from underneath */
	unregister_netdev ( netdev );

	/* Destroy data path */
	efx_remove ( netdev );

	netdev_nullify ( netdev );
	netdev_put ( netdev );
}

const struct efx_nic_type siena_nic_type = {
	.mcdi_rpc = _siena_mcdi,
};

static struct pci_device_id siena_nics[] = {
	PCI_ROM(0x1924, 0x0803, "SFC9020", "Solarflare"),
	PCI_ROM(0x1924, 0x0813, "SFL9021", "Solarflare"),
};

struct pci_driver siena_driver __pci_driver = {
	.ids = siena_nics,
	.id_count = sizeof ( siena_nics ) / sizeof ( siena_nics[0] ),
	.probe = siena_probe,
	.remove = siena_remove,
};
