/**************************************************************************
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Device driver for Solarflare Communications EF10 devices
 *
 * Written by Shradha Shah <sshah@solarflare.com>
 *
 * Copyright 2012-2017 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 *
 ***************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <byteswap.h>
#include <gpxe/io.h>
#include <gpxe/pci.h>
#include <gpxe/malloc.h>
#include <gpxe/ethernet.h>
#include <gpxe/iobuf.h>
#include <gpxe/netdevice.h>
#ifndef SFC_IPXE
#include <gpxe/nvo.h>
#include <gpxe/nvs.h>
#include <gpxe/dyncfg.h>
#include <gpxe/uaccess.h>
#include <gpxe/dhcp.h>
#endif	/* SFC_IPXE */
#include "efx_hunt.h"
#include "efx_bitfield.h"
#include "ef10_regs.h"
#include "mc_driver_pcol.h"
#ifdef SFC_IPXE
#include <ipxe/if_ether.h>
#else
#include <gpxe/hp_ocsd.h>
#include "sfc/sfc_hp_ocsd.h"
#include <gpxe/instance.h>
#include "sfc/ef10/dyncfg_accessors.h"
#include "tlv_partition.h"
#include "efx_nvram.h"
#include "pci_ids.h"
#endif	/* SFC_IPXE */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#define	HUNTINGTON_NVRAM_CHUNK 0x80
#define HUNTINGTON_NVS_MAX_LENGTH 0x1000

#ifdef SFC_IPXE
#define EMCDI_IO(code)	EUNIQ(EINFO_EIO, (code))
#else
#define EMCDI_IO(code)	EUNIQ(EIO, (code))
#endif

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(_a, _b) ((_a) > (_b) ? (_a) : (_b))
#endif

struct hunt_nic *primary_nics = NULL;

struct hunt_nic {
	struct efx_nic efx;

#ifndef SFC_IPXE
	/* NVO */
	struct nvo_fragment nvo_fragments[2];
	struct nvo_block nvo;
	struct nvs_device expcfg_nvs;

	/* dynamic config */
	struct dyncfg dyncfg;
#endif	/* SFC_IPXE */

	/* PHY information */
	unsigned int phy_cap_mask;
	unsigned int phy_cap;
	unsigned long link_poll_timer;

	/* resource housekeeping */
	uint64_t uc_filter_id;
	uint64_t mc_filter_id;
	u8 mac[ETH_ALEN];

	struct {
		/* Common payload for all MCDI requests */
		unsigned int seqno;

		size_t resp_hdr_len;
		size_t resp_data_len;

		struct io_buffer *iob;
		uint64_t dma_addr;
	} mcdi;

	struct hunt_nic *primary;
	struct hunt_nic *next_primary;
	u32 flags;
#ifndef SFC_IPXE
	unsigned int pf_index;
	unsigned int multiple_expcfg;
	u32 datapath_caps;
	u32 datapath_caps2;
#endif	/* SFC_IPXE */
};

#ifndef SFC_IPXE
#define BIT_ULL(nr)		(1ULL << (nr))
#define hunt_has_cap(caps, flag) \
	(!!((caps) & BIT_ULL(MC_CMD_GET_CAPABILITIES_V2_OUT_ ## flag ## _LBN)))

static int hunt_has_multiple_port_modes ( struct net_device *netdev );
#endif	/* SFC_IPXE */

static int hunt_nic_is_primary(struct hunt_nic *hunt)
{
	return (hunt->flags & (1 << MC_CMD_DRV_ATTACH_EXT_OUT_FLAG_PRIMARY));
}

#ifndef SFC_IPXE
static int hunt_nic_is_linkctrl ( struct hunt_nic *hunt )
{
	return (hunt->flags & (1 << MC_CMD_DRV_ATTACH_EXT_OUT_FLAG_LINKCTRL));
}

static inline int pci_nic_is_medford2_or_newer(struct pci_device *pci)
{
	return (SFU_DEVICE_TYPE(pci->device) >= SFC_MEDFORD2_PF_DEVICE_ID);
}
#endif	/* SFC_IPXE */

/*******************************************************************************
 *
 *
 * MCDI transport
 *
 * This has been based on the implementation of MCDI in the common code driver.
 *
 *
 ******************************************************************************/

static int hunt_mcdi_init(struct hunt_nic *hunt)
{
	size_t max_msg_size;
	int rc;

	/* The MCDI message has two 32-bit headers (the MCDI header and the
	 * MCDI v2 extended command) and then up to MCDI_CTL_SDU_LEN_MAX_V2
	 * bytes of payload
	 */
	max_msg_size = 2 * sizeof(efx_dword_t) + MCDI_CTL_SDU_LEN_MAX_V2;

	hunt->mcdi.iob = alloc_iob(max_msg_size);
	if (!hunt->mcdi.iob) {
		rc = -ENOMEM;
		return rc;
	}
	return 0;
}

static void hunt_mcdi_copyin(struct hunt_nic *hunt,
			     unsigned int cmd,
			     uint8_t *inbuf,
			     size_t inlen)
{
	efx_dword_t hdr[2];
	uint32_t seqno;
	unsigned int xflags;
	size_t hdr_len;
	u8 *pdu = hunt->mcdi.iob->data;

	seqno = hunt->mcdi.seqno & MCDI_SEQ_MASK;

	xflags = 0;

	EFX_POPULATE_DWORD_7(hdr[0],
			     MCDI_HEADER_CODE, MC_CMD_V2_EXTN,
			     MCDI_HEADER_RESYNC, 1,
			     MCDI_HEADER_DATALEN, 0,
			     MCDI_HEADER_SEQ, seqno,
			     MCDI_HEADER_ERROR, 0,
			     MCDI_HEADER_RESPONSE, 0,
			     MCDI_HEADER_XFLAGS, xflags);
	EFX_POPULATE_DWORD_2(hdr[1],
			     MC_CMD_V2_EXTN_IN_EXTENDED_CMD, cmd,
			     MC_CMD_V2_EXTN_IN_ACTUAL_LEN, inlen);

	hdr_len = sizeof(hdr);

	memcpy(pdu, &hdr, hdr_len);
	assert(inlen <= MCDI_CTL_SDU_LEN_MAX_V2);
	memcpy(pdu + hdr_len, inbuf, inlen);

	wmb();	/* Sync the data before ringing the doorbell */

	/* Ring the doorbell to post the command DMA address to the MC */
	hunt->mcdi.dma_addr = virt_to_bus(hunt->mcdi.iob->data);

	assert((hunt->mcdi.dma_addr & 0xFF) == 0);

	_efx_writel(&hunt->efx,
		   cpu_to_le32((u64)hunt->mcdi.dma_addr >> 32),
		   ER_DZ_MC_DB_LWRD);

	_efx_writel(&hunt->efx,
		   cpu_to_le32((u32)hunt->mcdi.dma_addr),
		   ER_DZ_MC_DB_HWRD);
}

static void hunt_mcdi_copyout(struct hunt_nic *hunt,
			      uint8_t *outbuf, size_t outlen)
{
	size_t offset;
	const u8 *pdu = hunt->mcdi.iob->data;

	offset =  hunt->mcdi.resp_hdr_len;

	if (outlen > 0)
		memcpy(outbuf, pdu+offset, outlen);
}

static int hunt_mcdi_request_poll(struct hunt_nic *hunt, bool quiet)
{
	unsigned int resplen, respseq, error;
	unsigned long finish;
	efx_dword_t errdword;
	efx_qword_t qword;
	const efx_dword_t *pdu = hunt->mcdi.iob->data;
	const u8 *pdu1 = hunt->mcdi.iob->data;
	int delay, rc;

	/* Spin for up to 5s, polling at intervals of 10us, 20us, ... ~100ms  */
	finish = currticks() + (5 * TICKS_PER_SEC);
	delay = 10;
	while (1) {
		udelay(delay);

		/* Check for an MCDI response */
		if (EFX_DWORD_FIELD(*pdu, MCDI_HEADER_RESPONSE))
			break;

		if (currticks() >= finish)
			return -ETIMEDOUT;

		if (delay < 100000)
			delay *= 2;
	}

	memcpy(&qword, pdu1, 8);

	/* qword.dword[0] is the MCDI header; qword.dword[1] is the MCDI v2
	 * extended command
	 */
	respseq = EFX_DWORD_FIELD(qword.dword[0], MCDI_HEADER_SEQ);
	error = EFX_DWORD_FIELD(qword.dword[0], MCDI_HEADER_ERROR);
	resplen = EFX_DWORD_FIELD(qword.dword[1], MC_CMD_V2_EXTN_IN_ACTUAL_LEN);

	if (error && resplen == 0) {
		if (!quiet)
			DBGC(hunt, "MC rebooted\n");
		return -EIO;
	} else if ((respseq ^ hunt->mcdi.seqno) & MCDI_SEQ_MASK) {
		if (!quiet)
			DBGC(hunt, "MC response mismatch rxseq 0x%x txseq "
			     "0x%x\n", respseq, hunt->mcdi.seqno);
		return -EIO;
	} else if (error) {
		memcpy(&errdword, pdu1 + 8, 4);
		rc = EFX_DWORD_FIELD(errdword, EFX_DWORD_0);
		switch (rc) {
		case MC_CMD_ERR_ENOENT:
			return -ENOENT;
		case MC_CMD_ERR_EINTR:
			return -EINTR;
		case MC_CMD_ERR_EACCES:
			return -EACCES;
		case MC_CMD_ERR_EBUSY:
			return -EBUSY;
		case MC_CMD_ERR_EINVAL:
			return -EINVAL;
		case MC_CMD_ERR_EDEADLK:
			return -EDEADLK;
		case MC_CMD_ERR_ENOSYS:
			return -ENOSYS;
		case MC_CMD_ERR_ETIME:
			return -ETIME;
		case MC_CMD_ERR_EPERM:
			return -EPERM;
		default:
			/* Return the MC error in an I/O error. */
			return EMCDI_IO(rc & 0xff);
		}
	}
	hunt->mcdi.resp_hdr_len = 8;
	hunt->mcdi.resp_data_len = resplen;

	return 0;
}

static void hunt_mcdi_fini(struct hunt_nic *hunt)
{
	free_iob(hunt->mcdi.iob);
}

int _hunt_mcdi(struct efx_nic *efx, unsigned int cmd,
	       const efx_dword_t *inbuf, size_t inlen,
	       efx_dword_t *outbuf, size_t outlen,
	       size_t *outlen_actual, bool quiet)
{
	int rc;
	struct hunt_nic *hunt = (struct hunt_nic *) efx;
	size_t local_outlen_actual;

	if (outlen_actual == NULL)
		outlen_actual = &local_outlen_actual;

	++hunt->mcdi.seqno;
	hunt_mcdi_copyin(hunt, cmd, (uint8_t *) inbuf, inlen);

	rc = hunt_mcdi_request_poll(hunt, quiet);
	if (rc != 0) {
		if (!quiet)
#ifdef SFC_IPXE
			DBGC(hunt, "MC response to cmd 0x%x: %s\n",
			     cmd, strerror(rc));
#else
			DBGC(hunt, "MC response rc %x to cmd 0x%x: %s\n",
			     (PXENV_STATUS(rc) == EIO? (rc >> 8) & 0xff :
			      -rc),
			     cmd, strerror(rc));
#endif
		return rc;
	}

	*outlen_actual = hunt->mcdi.resp_data_len;

	hunt_mcdi_copyout(hunt, (uint8_t *) outbuf, outlen);

	return 0;
}

static int hunt_mcdi(struct hunt_nic *hunt, struct efx_mcdi_req_s *req)
{
	return _hunt_mcdi(&hunt->efx, req->emr_cmd,
			  (const efx_dword_t *) req->emr_in_buf,
			  req->emr_in_length,
			  (efx_dword_t *) req->emr_out_buf, req->emr_out_length,
			  &req->emr_out_length_used, false);
}

static int hunt_mcdi_quiet(struct hunt_nic *hunt, struct efx_mcdi_req_s *req)
{
	return _hunt_mcdi(&hunt->efx, req->emr_cmd,
			  (const efx_dword_t *) req->emr_in_buf,
			  req->emr_in_length,
			  (efx_dword_t *) req->emr_out_buf, req->emr_out_length,
			  &req->emr_out_length_used, true);
}

#ifndef SFC_IPXE
/*******************************************************************************
 *
 *
 * NVRAM support
 *
 *
 ******************************************************************************/
static int hunt_has_option_rom ( struct net_device *netdev )
{
	struct hunt_nic *hunt = netdev_priv ( netdev );

	return hunt_nic_is_linkctrl ( hunt );
}

static int hunt_nvram_partn_size(struct hunt_nic *hunt_nic, uint32_t partn,
				 size_t *size)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_INFO_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_NVRAM_INFO_OUT_LEN);
	int rc;

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	req.emr_cmd = MC_CMD_NVRAM_INFO;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_NVRAM_INFO_IN_LEN;
	req.emr_out_buf = outbuf;
	req.emr_out_length = MC_CMD_NVRAM_INFO_OUT_LEN;

	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_INFO_IN_TYPE, partn);

	rc = hunt_mcdi(hunt_primary, &req);
	if(rc != 0)
		return rc;

	if(req.emr_out_length_used < MC_CMD_NVRAM_INFO_OUT_LEN)
		return -EMSGSIZE;

	*size = MCDI_DWORD(req.emr_out_buf, NVRAM_INFO_OUT_SIZE);

	return 0;
}

static int hunt_nvram_update_start(struct hunt_nic *hunt_nic, uint32_t partn)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_UPDATE_START_IN_LEN);

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	req.emr_cmd = MC_CMD_NVRAM_UPDATE_START;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_NVRAM_UPDATE_START_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_UPDATE_START_IN_TYPE, partn);

	return hunt_mcdi(hunt_primary, &req);
}

static int hunt_nvram_read(struct hunt_nic *hunt_nic, uint32_t partn,
			   uint32_t address, void *data, size_t len)
{
	size_t chunk;
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_READ_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_NVRAM_READ_OUT_LENMAX);
	int rc;

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	while(len > 0) {
		chunk = MIN(len, HUNTINGTON_NVRAM_CHUNK);

		BUILD_BUG_ON(HUNTINGTON_NVRAM_CHUNK >
			     MC_CMD_NVRAM_READ_OUT_LENMAX);

		req.emr_cmd = MC_CMD_NVRAM_READ;
		req.emr_in_buf = inbuf;
		req.emr_in_length = MC_CMD_NVRAM_READ_IN_LEN;
		req.emr_out_buf = outbuf;
		req.emr_out_length = MC_CMD_NVRAM_READ_OUT_LENMAX;

		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_READ_IN_TYPE, partn);
		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_READ_IN_OFFSET, address);
		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_READ_IN_LENGTH, chunk);

		rc = hunt_mcdi(hunt_primary, &req);
		if(rc != 0)
			return rc;

		if(req.emr_out_length_used < MC_CMD_NVRAM_READ_OUT_LEN(chunk))
			return -EMSGSIZE;

		memcpy(data,
		       MCDI_PTR(req.emr_out_buf, NVRAM_READ_OUT_READ_BUFFER),
		       chunk);

		len -= chunk;
		data += chunk;
		address += chunk;
	}

	return 0;
}

static int hunt_nvram_erase(struct hunt_nic *hunt_nic, uint32_t partn,
			    uint32_t offset, size_t length)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_ERASE_IN_LEN);

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	req.emr_cmd = MC_CMD_NVRAM_ERASE;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_NVRAM_ERASE_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_ERASE_IN_TYPE, partn);
	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_ERASE_IN_OFFSET, offset);
	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_ERASE_IN_LENGTH, length);

	return hunt_mcdi(hunt_primary, &req);
}

static int hunt_nvram_set_generation(void *data, size_t len,
				      unsigned int generation)
{
	nvram_partition_t nvram_partition;
	int rc;

	rc = tlv_init_partition_from_buffer(&nvram_partition, (void *) data,
					    len);
	if (rc) {
		DBG("The dynamic config default partition is corrupt, rc=%x\n",
		    rc);
		return rc;
	}

	return tlv_set_partition_generation(&nvram_partition, generation);
}

static int hunt_nvram_increment_generation(struct hunt_nic *hunt_nic,
					   const void *data, size_t len)
{
	struct tlv_partition_header *header;
	nvram_partition_t nvram_partition;
	int rc;

	rc = tlv_init_partition_from_buffer(&nvram_partition, (void *) data,
					    len);
	if (!rc)
		rc = tlv_increment_partition_generation(&nvram_partition);

	header = (void *) data;
	DBGC(hunt_nic, "Partition type %x rc=%d new generation=%d\n",
	     nvram_partition.type, rc, header->generation);
	return rc;
}

static int hunt_nvram_write(struct hunt_nic *hunt_nic, uint32_t partn,
			    uint32_t address, const void *data, size_t len)
{
	size_t chunk;
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_WRITE_IN_LENMAX);
	int rc;

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	while(len > 0) {
		chunk = MIN(len, HUNTINGTON_NVRAM_CHUNK);

		BUILD_BUG_ON(HUNTINGTON_NVRAM_CHUNK >
			     MC_CMD_NVRAM_WRITE_IN_LENMAX);

		req.emr_cmd = MC_CMD_NVRAM_WRITE;
		req.emr_in_buf = inbuf;
		req.emr_in_length = MC_CMD_NVRAM_WRITE_IN_LEN(chunk);
		req.emr_out_buf = NULL;
		req.emr_out_length = 0;

		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_WRITE_IN_TYPE, partn);
		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_WRITE_IN_OFFSET, address);
		MCDI_SET_DWORD(req.emr_in_buf, NVRAM_WRITE_IN_LENGTH, chunk);

		memcpy(MCDI_PTR(req.emr_in_buf, NVRAM_WRITE_IN_WRITE_BUFFER),
		       data, chunk);

		rc = hunt_mcdi(hunt_primary, &req);
		if(rc != 0)
			return rc;

		len -= chunk;
		data += chunk;
		address += chunk;
	}

	return 0;
}

static void hunt_nvram_update_finish(struct hunt_nic *hunt_nic, uint32_t partn)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_NVRAM_UPDATE_FINISH_IN_LEN);

	struct hunt_nic *hunt_primary = hunt_nic->primary;

	req.emr_cmd = MC_CMD_NVRAM_UPDATE_FINISH;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_NVRAM_UPDATE_FINISH_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_UPDATE_FINISH_IN_TYPE, partn);
	MCDI_SET_DWORD(req.emr_in_buf, NVRAM_UPDATE_FINISH_IN_REBOOT, 0);

	(void)hunt_mcdi(hunt_primary, &req);
}

static int hunt_vpd_id(struct efx_nic *efx, uint8_t *outbuf, size_t outlen)
{
	struct tlv_partition_header sconfig_hdr;
	nvram_partition_t nvram_partition;
	tlv_cursor_t *cursor = &nvram_partition.tlv_cursor;
	struct efx_partition *partition;
	uint8_t	*sconfig = NULL;
	int	rc, size;

	partition = efx_nvram_info(efx, NVRAM_PARTITION_TYPE_STATIC_CONFIG);
	if (!partition)
		return -EINVAL;

	rc = efx_nvram_read(efx, partition, 0, sizeof(sconfig_hdr),
			    &sconfig_hdr);
	if (rc)
		goto out;
	if (sconfig_hdr.tag != TLV_TAG_PARTITION_HEADER) {
		rc = -ENOENT;
		goto out;
	}

	size = sconfig_hdr.total_length;
	sconfig = malloc(size);
	if (!sconfig)
		return -ENOMEM;

	rc = efx_nvram_read(efx, partition, 0, size, sconfig);
	if (rc)
		goto out;

	rc = tlv_init_partition_from_buffer(&nvram_partition, sconfig, size);
	if (rc)
		goto out;

	/* The device identity is stored in the static config. For newer NICs
	 * (like Medford) it part of tag GLOBAL_STATIC_VPD, and for older ones
	 * it is inside PF_STATIC_VPD.
	 */
	rc = tlv_find(cursor, TLV_TAG_GLOBAL_STATIC_VPD);
	if (rc != TLV_OK)
		rc = tlv_find(cursor, TLV_TAG_PF_STATIC_VPD(0));
	if (rc != TLV_OK) {
		rc = -ENOENT;
		goto out;
	}

	rc = efx_vpd(tlv_value(cursor), tlv_length(cursor), VPD_ID, 0,
		     outbuf, outlen);
out:
	if (sconfig)
		free(sconfig);
	efx_partition_free(partition);
	return rc;
}

static void hunt_get_vpd_ro(uint16_t keyword, unsigned char *buf,
			    unsigned size, uint8_t *out_value, size_t out_len)
{
	nvram_partition_t nvram_partition;
	tlv_cursor_t *cursor = &nvram_partition.tlv_cursor;
	int rc;

	out_value[0] = '\0';

	rc = tlv_init_partition_from_buffer(&nvram_partition, buf, size);
	if (rc)
		return;

	/* The device identity is stored in the dynamic config. For newer NICs
	 * (like Medford) it part of tag GLOBAL_DYNAMIC_VPD, and for older ones
	 * it is inside PF_DYNAMIC_VPD.
	 */
	rc = tlv_find(cursor, TLV_TAG_GLOBAL_DYNAMIC_VPD);
	if (rc != TLV_OK)
		rc = tlv_find(cursor, TLV_TAG_PF_DYNAMIC_VPD(0));
	if (rc != TLV_OK)
		return;

	(void) efx_vpd(tlv_value(cursor), tlv_length(cursor), VPD_RO,
		       keyword, out_value, out_len);
}

static int hunt_vpd(struct net_device *netdev, enum sf_vpd_tag tag,
		    uint16_t keyword, uint8_t *buf, size_t len)
{
	struct hunt_nic *hunt = netdev_priv(netdev);
	struct efx_nic *efx = &hunt->efx;
	int rc = -EINVAL;

	switch (tag) {
	case VPD_ID:
		rc = hunt_vpd_id(efx, buf, len);
		break;
	case VPD_RO:
		if (keyword == SF_VPD_KEYWORD('V', '0')) {
			memcpy(buf, &hunt->dyncfg.vpd_v0,
			       MIN(len, MAX_VPD_V0));
			rc = len;
		}
		break;
	}
	return rc;
}

/*******************************************************************************
 *
 *
 * NVS transport
 *
 *
 ******************************************************************************/
static int hunt_nvs_read(unsigned partn_type,
			 struct hunt_nic *hunt,
			 unsigned int address, void *data, size_t len)
{
	return hunt_nvram_read(hunt, partn_type, address, data, len);
}

static int hunt_nvs_write(unsigned partn_type,
			  struct hunt_nic *hunt, struct nvs_device *nvs,
			  unsigned int address, const void *data, size_t len)
{
	int rc;

	if (hunt_has_cap(hunt->datapath_caps2, TSA_BOUND)) {
		printf("Cannot write to flash, adapter is bound to a trusted controller\n");
		return -EPERM;
	}

	if((rc = hunt_nvram_update_start(hunt, partn_type)) != 0)
		goto fail1;

	if((rc = hunt_nvram_erase(hunt, partn_type, address, nvs->size)) != 0)
		goto fail2;

	if((rc = hunt_nvram_write(hunt, partn_type, address, data, len)) != 0)
		goto fail3;

	hunt_nvram_update_finish(hunt, partn_type);

	return 0;

fail3:
fail2:
	hunt_nvram_update_finish(hunt, partn_type);
fail1:
	if (rc)
		DBGC(hunt, "%s NVS write of type %d length %d failed: %d\n",
		     __func__, partn_type, len, rc);
	return rc;
}

/* Find out if there are multiple EXPROM_CONFIG partitions.  This assumes
 * that a partition exists for EXPROM_CONFIG on port 0, so return whether
 * or not there are any others.
 *
 * Return 0 if there is only a single EXPROM_CONFIG partition, 1 if there
 * are multiple partitions, or negative for an error.
 */
static int hunt_have_multiple_expcfg(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_NVRAM_PARTITIONS_OUT_LENMAX);
	unsigned int num_partns, idx;
	int rc;

	/* All PFs on a NIC will have the same value for multiple_expcfg so
	 * return the value from the primary PF to avoid permissions-related
	 * restrictions on MC_CMD_NVRAM_PARTITIONS.
	 */
	if (!hunt_nic_is_primary(hunt))
		return hunt->primary->multiple_expcfg;

	req.emr_cmd = MC_CMD_NVRAM_PARTITIONS;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc < 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_NVRAM_PARTITIONS_OUT_LENMIN)
		return -EMSGSIZE;

	num_partns = MCDI_DWORD(outbuf, NVRAM_PARTITIONS_OUT_NUM_PARTITIONS);
	for (idx = 0; idx < num_partns; ++idx) {
		unsigned int type;

		type = MCDI_ARRAY_DWORD(outbuf,
					NVRAM_PARTITIONS_OUT_TYPE_ID, idx);

		if (type == NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT1 ||
		    type == NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT2 ||
		    type == NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT3)
			return 1;
	}

	return 0;
}

static int hunt_expcfg_nvs_read(struct nvs_device *nvs, unsigned int address,
				void *data, size_t len)
{
	struct hunt_nic *hunt = container_of(nvs, struct hunt_nic, expcfg_nvs);
	unsigned int partn_type;

	if (hunt->multiple_expcfg)
		partn_type = NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT0 +
			     hunt->efx.port;
	else
		partn_type = NVRAM_PARTITION_TYPE_EXPROM_CONFIG;

	return hunt_nvs_read(partn_type, hunt, address, data, len);
}

static int hunt_expcfg_nvs_read_dummy(struct nvs_device *nvs __unused,
				      unsigned int address __unused,
				      void *data __unused, size_t len __unused)
{
	return 0;
}

static int hunt_expcfg_nvs_write(struct nvs_device *nvs, unsigned int address,
				 const void *data, size_t len)
{
	struct hunt_nic *hunt = container_of(nvs, struct hunt_nic, expcfg_nvs);
	unsigned int partn_type;

	if (hunt->multiple_expcfg)
		partn_type = NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT0 +
			     hunt->efx.port;
	else
		partn_type = NVRAM_PARTITION_TYPE_EXPROM_CONFIG;

	return hunt_nvs_write(partn_type, hunt, nvs, address, data, len);
}

static int hunt_expcfg_nvs_write_primary(struct nvs_device *nvs,
					 unsigned int address __unused,
					 const void *data __unused,
					 size_t len __unused)
{
	struct hunt_nic *hunt = container_of(nvs, struct hunt_nic, expcfg_nvs);
	struct hunt_nic *hunt_primary = hunt->primary;
	struct nvs_device *nvs_primary = &hunt_primary->expcfg_nvs;
	struct nvo_block *nvo_primary = &hunt_primary->nvo;

	/* Pretend to be the primary PF and write the whole block, so start
	 * from address offset 0.
	 */
	return hunt_nvs_write(NVRAM_PARTITION_TYPE_EXPROM_CONFIG, hunt_primary,
			      nvs_primary, 0, nvo_primary->data,
			      nvs_primary->size);
}

static int hunt_dyncfg_nvs_read(struct nvs_device *nvs, unsigned int address,
				void *data, size_t len)
{
	struct dyncfg *dyncfg = container_of(nvs, struct dyncfg, nvs);
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);

	unsigned partn_type =
		NVRAM_PARTITION_TYPE_DYNAMIC_CONFIG;

	return hunt_nvs_read(partn_type, hunt, address, data, len);
}

static int hunt_dyncfg_nvs_write(struct nvs_device *nvs, unsigned int address,
				 const void *data, size_t len)
{
	struct dyncfg *dyncfg = container_of(nvs, struct dyncfg, nvs);
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);

	unsigned partn_type =
		NVRAM_PARTITION_TYPE_DYNAMIC_CONFIG;

	return hunt_nvs_write(partn_type, hunt, nvs, address, data, len);
}

static int hunt_nvo_probe_multiple_expcfg(struct hunt_nic *hunt)
{
	struct nvo_fragment *fragments = hunt->nvo_fragments;
	struct nvs_device *nvs = &hunt->expcfg_nvs;
	size_t partn_length = 0;
	unsigned int type;
	int rc;

	type = NVRAM_PARTITION_TYPE_EXPROM_CONFIG_PORT0 + hunt->efx.port;

	if ((rc = hunt_nvram_partn_size(hunt, type, &partn_length)) != 0)
		return rc;

	nvs->block_size = partn_length;
	nvs->size = partn_length;
	nvs->read = hunt_expcfg_nvs_read;
	nvs->write = hunt_expcfg_nvs_write;
	fragments[0].len = MIN(partn_length, HUNTINGTON_NVS_MAX_LENGTH);

	return 0;
}

/* The 64kB flash block is shared between 16 PFs and the CMAC.
 * Each PF gets 2kB.
 */
#define PF_EXPCFG_SIZE 2048

static int hunt_nvo_probe_single_expcfg(struct hunt_nic *hunt)
{
	struct nvo_fragment *fragments = hunt->nvo_fragments;
	struct nvs_device *nvs = &hunt->expcfg_nvs;
	int rc;

	if (hunt_nic_is_primary(hunt)) {
		if ((rc = hunt_nvo_probe_multiple_expcfg(hunt)) != 0)
			return rc;
		/* Primary PF accesses the entire flash block */
		fragments[0].len = nvs->size;
	} else {
		nvs->block_size = PF_EXPCFG_SIZE;
		nvs->size = PF_EXPCFG_SIZE;
		/* The ->read() function pointer isn't expected to be called
		 * for this function, but use a dummy function just in case
		 * this changes because the caller doesn't check for a NULL
		 * function pointer.
		 */
		nvs->read = hunt_expcfg_nvs_read_dummy;
		nvs->write = hunt_expcfg_nvs_write_primary;
		fragments[0].len = PF_EXPCFG_SIZE;
	}

	return 0;
}

/* Recover the expansion ROM config from the default partition */
static bool expcfg_recover_from_partition(struct settings *settings)
{
	struct nvo_block *nvo = container_of(settings, struct nvo_block,
					     settings);
	struct hunt_nic *hunt = container_of(nvo, struct hunt_nic, nvo);
	struct hunt_nic *hunt_primary = hunt->primary;
	size_t defaults_size;
	unsigned char *new_buf;
	struct nvs_device *nvs;
	int rc;

	rc = hunt_nvram_partn_size(hunt,
				   NVRAM_PARTITION_TYPE_ROMCONFIG_DEFAULTS,
				   &defaults_size);
	if (rc)
		return false;

	/* The primary PF is used to store the whole partition */
	nvo = &hunt_primary->nvo;
	if (defaults_size > nvo->total_len) {
		new_buf = realloc(nvo->data, defaults_size);
		if (!new_buf)
			return false;
		nvo->data = new_buf;
	} else {
		new_buf = nvo->data;
	}

	rc = hunt_nvs_read(NVRAM_PARTITION_TYPE_ROMCONFIG_DEFAULTS, hunt_primary,
			   0, new_buf, defaults_size);
	if (rc) {
		DBGC(hunt_primary,
		     "Error %d reading the exprom config default partition\n",
		     rc);
		return false;
	}

	nvs = &hunt_primary->expcfg_nvs;
	nvs->block_size = nvs->size = defaults_size;
	rc = nvs->write(nvs, 0, new_buf, defaults_size);
	if (rc)
		return false;
	return true;
}

static int hunt_nvo_clear(struct settings *settings)
{
	if (expcfg_recover_from_partition(settings))
		return 0;

	return nvo_clear(settings);
}

static struct settings_operations hunt_nvo_settings_operations = {
	.applies = nvo_applies,
	.store = nvo_store,
	.fetch = nvo_fetch,
	.clear = hunt_nvo_clear,
};

static int hunt_nvo_probe(struct hunt_nic *hunt)
{
	struct nvs_device *nvs = &hunt->expcfg_nvs;
	struct nvo_block *nvo = &hunt->nvo;
	struct net_device *netdev = hunt->efx.netdev;
	int rc;

	/* Initialise the nvo_fragments */
	memset(hunt->nvo_fragments, '\0', sizeof(hunt->nvo_fragments));

	nvs->word_len_log2 = 0;

	if (hunt->multiple_expcfg)
		rc = hunt_nvo_probe_multiple_expcfg(hunt);
	else
		rc = hunt_nvo_probe_single_expcfg(hunt);

	if (rc != 0)
		return rc;

	nvo->nvs = nvs;
	nvo->fragments = hunt->nvo_fragments;
	settings_init(&nvo->settings, &hunt_nvo_settings_operations,
		      &netdev->refcnt, 0);

	return 0;
}

/*******************************************************************************
 *
 *
 * Dynamic Config 
 *
 *
 ******************************************************************************/
int hunt_has_dyncfg(struct net_device *netdev)
{
	(void)netdev;
	return 1;
}

static enum sfc_port_mode to_generic_port_mode(uint16_t device_id,
					       uint32_t port_mode)
{
	enum sfc_port_mode ret = port_mode;

	switch (device_id) {
	case SFC_GREENPORT_PF_DEVICE_ID:
	case SFC_GREENPORT_VF_DEVICE_ID:
	case SFC_FARMINGDALE_PF_DEVICE_ID:
	case SFC_FARMINGDALE_VF_DEVICE_ID:
		switch (port_mode) {
		case TLV_PORT_MODE_DEFAULT:
			ret = SFC_PORT_MODE_DEFAULT;
		case TLV_PORT_MODE_40G_40G:
		case TLV_PORT_MODE_10G_10G_10G_10G:
			break;
		default:
			ret = SFC_PORT_MODE_INVALID;
		}
		break;
	case SFC_MEDFORD_PF_DEVICE_ID:
	case SFC_MEDFORD_VF_DEVICE_ID:
		switch (port_mode) {
		case TLV_PORT_MODE_DEFAULT:
			ret = SFC_PORT_MODE_DEFAULT;
			break;
		case TLV_PORT_MODE_10G_10G_10G_10G_Q:
			ret = SFC_PORT_MODE_10G_10G_10G_10G;
			break;
		case TLV_PORT_MODE_10G_10G_10G_10G:
			ret = SFC_PORT_MODE_4x10G;
			break;
		default:
			if (port_mode < TLV_PORT_MODE_40G ||
			    port_mode > TLV_PORT_MODE_10G_10G_10G_10G_Q)
				ret = SFC_PORT_MODE_INVALID;
		}
		break;
	case SFC_MEDFORD2_PF_DEVICE_ID:
	case SFC_MEDFORD2_VF_DEVICE_ID:
		switch (port_mode) {
		case TLV_PORT_MODE_DEFAULT:
			return SFC_PORT_MODE_DEFAULT;
		case TLV_PORT_MODE_40G:
			return SFC_PORT_MODE_100G;
		case TLV_PORT_MODE_10G_10G:
			return SFC_PORT_MODE_25G_25G;
		case TLV_PORT_MODE_1x2_1x2:
			return SFC_PORT_MODE_50G_50G_Q;
		case TLV_PORT_MODE_10G_10G_10G_10G:
			return SFC_PORT_MODE_4x25G;
		case TLV_PORT_MODE_10G_10G_10G_10G_Q:
			return SFC_PORT_MODE_25G_25G_25G_25G;
		case TLV_PORT_MODE_2x2_NA:
			return SFC_PORT_MODE_50G_50G;
		default:
			ret = SFC_PORT_MODE_INVALID;
		}
		break;
	default:
		ret = SFC_PORT_MODE_INVALID;
	}

	if (ret == SFC_PORT_MODE_INVALID)
		DBG("Unsupport port mode %d for PCI device 0x%x\n", port_mode,
		    device_id);
	return ret;
}

static unsigned int to_generic_port_modes(struct pci_device *pci,
					  unsigned int port_modes)
{
	unsigned int ret = 0, bit = 0;
	enum sfc_port_mode generic_port_mode;

	for (bit=0; port_modes; bit++)
	{
		if (port_modes & 1) {
			generic_port_mode = to_generic_port_mode(pci->device,
								 bit);
			if ((generic_port_mode != SFC_PORT_MODE_DEFAULT) &&
			    (generic_port_mode != SFC_PORT_MODE_INVALID))
				ret |= BIT(generic_port_mode);
		}
		port_modes >>= 1;
	}
	return ret;
}

static void set_generic_port_mode(struct dyncfg *dyncfg)
{
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	struct net_device *netdev = hunt->efx.netdev;
	struct pci_device *pci = (struct pci_device *) netdev->dev;

	dyncfg->generic_port_mode = to_generic_port_mode(pci->device,
							 dyncfg->port_mode);
}

static void set_nic_port_mode(struct dyncfg *dyncfg)
{
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	struct net_device *netdev = hunt->efx.netdev;
	struct pci_device *pci = (struct pci_device *) netdev->dev;

	if (dyncfg->generic_port_mode == SFC_PORT_MODE_INVALID)
		return;	/* Do not modify port_mode */

	switch (pci->device) {
	case SFC_GREENPORT_PF_DEVICE_ID:
	case SFC_GREENPORT_VF_DEVICE_ID:
	case SFC_FARMINGDALE_PF_DEVICE_ID:
	case SFC_FARMINGDALE_VF_DEVICE_ID:
		dyncfg->port_mode = dyncfg->generic_port_mode;
		break;
	case SFC_MEDFORD_PF_DEVICE_ID:
	case SFC_MEDFORD_VF_DEVICE_ID:
		switch (dyncfg->generic_port_mode) {
		case SFC_PORT_MODE_10G_10G_10G_10G:
			dyncfg->port_mode = TLV_PORT_MODE_10G_10G_10G_10G_Q;
			break;
		case SFC_PORT_MODE_4x10G:
			dyncfg->port_mode = TLV_PORT_MODE_10G_10G_10G_10G;
			break;
		default:
			dyncfg->port_mode = dyncfg->generic_port_mode;
		}
		break;
	case SFC_MEDFORD2_PF_DEVICE_ID:
	case SFC_MEDFORD2_VF_DEVICE_ID:
		switch (dyncfg->generic_port_mode) {
		case SFC_PORT_MODE_DEFAULT:
			dyncfg->port_mode = TLV_PORT_MODE_DEFAULT;
			break;
		case SFC_PORT_MODE_100G:
			dyncfg->port_mode = TLV_PORT_MODE_40G;
			break;
		case SFC_PORT_MODE_25G_25G:
			dyncfg->port_mode = TLV_PORT_MODE_10G_10G;
			break;
		case SFC_PORT_MODE_50G_50G_Q:
			dyncfg->port_mode = TLV_PORT_MODE_1x2_1x2;
			break;
		case SFC_PORT_MODE_4x25G:
			dyncfg->port_mode = TLV_PORT_MODE_10G_10G_10G_10G;
			break;
		case SFC_PORT_MODE_25G_25G_25G_25G:
			dyncfg->port_mode = TLV_PORT_MODE_10G_10G_10G_10G_Q;
			break;
		case SFC_PORT_MODE_50G_50G:
			dyncfg->port_mode = TLV_PORT_MODE_2x2_NA;
			break;
		default:
			DBG("Unsupport port mode %d for PCI device 0x%x\n",
			    dyncfg->generic_port_mode, pci->device);
			dyncfg->port_mode = TLV_PORT_MODE_DEFAULT;
		}
		break;
	default:
		DBGC(hunt, "Unsupport PCI device 0x%x\n", pci->device);
	}
}

/**
 * Store value of dynamic config setting
 *
 * @v settings          Settings block
 * @v setting           Setting to store
 * @v data              Setting data, or NULL to clear setting
 * @v len               Length of setting data
 * @ret rc              Return status code
 */
int hunt_dyncfg_store(struct settings *settings, struct setting *setting,
		      const void *data, size_t len)
{
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);

#define CASE_SETTING(_DHCP_TAG, _dyncfg_name, _value_type, _bswap_helper)      \
	case _DHCP_TAG : {                                                     \
		_value_type value;                                             \
		if (len != sizeof(value))                                      \
			return -EINVAL;                                        \
		value = _bswap_helper(*((_value_type*)data));                  \
		dyncfg->settings_modified += (value != dyncfg->_dyncfg_name);  \
		dyncfg->_dyncfg_name = value;                                  \
		break;                                                         \
	}
	switch (setting->tag) {
	CASE_SETTING ( DHCP_EB_MSIX_LIMIT,    msix_limit,       uint16_t, ntohs)
	CASE_SETTING ( DHCP_EB_VF_MSIX_LIMIT, vf_msix_limit,    uint16_t, ntohs)
	CASE_SETTING ( DHCP_EB_VF_COUNT,      vf_count,         uint8_t,       )
	CASE_SETTING ( DHCP_FW_VARIANT,       firmware_variant, uint32_t, ntohl)
	CASE_SETTING ( DHCP_EB_PF_COUNT,      pf_count,         int16_t, ntohs)
	CASE_SETTING ( DHCP_EB_PORT_MODE,     generic_port_mode,uint32_t, ntohl)
	CASE_SETTING ( DHCP_EB_PFIOV_ENABLED, pfiov,            uint8_t,       )
	case DHCP_INSECURE_FILTERS : {
		/* special case, we read-modify-write the privilege mask */
		uint32_t privilege_mask = dyncfg->privilege_mask;
		uint8_t value;
		if (len != sizeof(value))
			return -EINVAL;
		value = *((uint8_t*)data);
		privilege_mask &= ~MC_CMD_PRIVILEGE_MASK_IN_GRP_INSECURE_FILTERS;
		if (value)
			privilege_mask |= MC_CMD_PRIVILEGE_MASK_IN_GRP_INSECURE_FILTERS;
		dyncfg->settings_modified += dyncfg->privilege_mask != privilege_mask;
		dyncfg->privilege_mask = privilege_mask;
		break;
	}
	case DHCP_VLAN_ENABLED: {
		uint8_t value;
		if (len != sizeof(value))
			return -EINVAL;
		value = *((uint8_t*)data);
		dyncfg->vlan_enabled = value;
		break;
	}
	case DHCP_EB_VLAN_TAG: {
		uint32_t value;
		if (len != sizeof(value))
			return -EINVAL;
		if (dyncfg->vlan_enabled)
			value = ntohl(*((uint32_t*) data));
		else
			value = DYNCFG_SETTING_DEFAULT_VLAN_TAG;
		dyncfg->settings_modified += (value != dyncfg->vlan_tag);
		dyncfg->vlan_tag = value;
		break;
	}
	default:
		return -EINVAL;
	}
#undef CASE_SETTING

        return 0;
}

/**
 *  * Fetch value of dynamic config setting
 *
 * @v settings          Settings block
 * @v setting           Setting to fetch
 * @v data              Buffer to fill with setting data
 * @v len               Length of buffer
 * @ret len             Length of setting data, or negative error
 *
 * The actual length of the setting will be returned even if
 * the buffer was too small.
 */
int hunt_dyncfg_fetch(struct settings *settings, struct setting *setting,
		      void *data, size_t len)
{
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);

#define CASE_SETTING(_DHCP_TAG, _dyncfg_name, _value_type, _bswap_helper)      \
	case _DHCP_TAG : {                                                     \
		_value_type value;                                             \
		if (len < sizeof(value))                                       \
			return -EINVAL;                                        \
		value = _bswap_helper(dyncfg->_dyncfg_name);                   \
		memcpy(data, &value, sizeof(value));                           \
		return sizeof(value);                                          \
	}
	switch (setting->tag) {
	CASE_SETTING ( DHCP_EB_MSIX_LIMIT,    msix_limit,       uint16_t, htons)
	CASE_SETTING ( DHCP_EB_VF_MSIX_LIMIT, vf_msix_limit,    uint16_t, htons)
	CASE_SETTING ( DHCP_EB_VF_COUNT,      vf_count,         uint8_t,       )
	CASE_SETTING ( DHCP_FW_VARIANT,       firmware_variant, uint32_t, htonl)
	CASE_SETTING ( DHCP_EB_PF_COUNT,      pf_count,         int16_t,  htons)
	CASE_SETTING ( DHCP_EB_PORT_MODE,     generic_port_mode,uint32_t, htonl)
	CASE_SETTING ( DHCP_EB_PFIOV_ENABLED, pfiov,            uint8_t,       )
	case DHCP_INSECURE_FILTERS : {
		uint8_t value;
		if (len < sizeof(value))
			return -EINVAL;
		value = !!(dyncfg->privilege_mask & MC_CMD_PRIVILEGE_MASK_IN_GRP_INSECURE_FILTERS);
		memcpy(data, &value, sizeof(value));
		return sizeof(value);
	}
	case DHCP_VLAN_ENABLED: {
		uint8_t value;
		if (len < sizeof(value))
			return -EINVAL;
		value = dyncfg->vlan_enabled;
		memcpy(data, &value, sizeof(value));
		return sizeof(value);
	}
	case DHCP_EB_VLAN_TAG: {
		uint32_t value;
		if (len < sizeof(value))
			return -EINVAL;
		if (dyncfg->vlan_enabled)
			value = htonl(dyncfg->vlan_tag);
		else
			value = 0;
		memcpy(data, &value, sizeof(value));
		return sizeof(value);
	}
	default:
		return -EINVAL;
	}
#undef CASE_SETTING
}

static int hunt_register_dyncfg(struct hunt_nic *hunt, struct settings *parent)
{
	return register_settings(&hunt->dyncfg.settings, parent, "sfc_dyncfg");
}

static void hunt_unregister_dyncfg(struct hunt_nic *hunt)
{
	unregister_settings(&hunt->dyncfg.settings);
}

static void hunt_dyncfg_remove(struct hunt_nic *hunt)
{
	struct dyncfg *dyncfg = &hunt->dyncfg;
	dyncfg->nvs.size = 0;
	dyncfg->nvs.block_size = 0;
	dyncfg->nvs.read = NULL;
	dyncfg->nvs.write = NULL;
}


int dyncfg_changed(struct settings *settings)
{
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);
	return dyncfg->settings_modified != 0;
}

static inline void init_dev_info_from_hunt(struct sfgpxe_dev_info *dev_info,
					   struct hunt_nic *hunt)
{
	struct net_device *netdev = hunt->efx.netdev;
	struct pci_device *pcidev = (struct pci_device *)netdev->dev;

	dev_info->pf_index = hunt->pf_index;
	dev_info->pci_device_id =
		((uint32_t)pcidev->vendor << 16) | pcidev->device;
	dev_info->port = hunt->efx.port;
}

/* Recover the dynamic config block from the default partition */
static bool dyncfg_recover_from_partition(struct dyncfg *dyncfg,
					  unsigned char **dyncfg_buf,
					  size_t dyncfg_buf_size)
{
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	size_t defaults_size = 0;
	unsigned char *new_buf;
	struct nvs_device *nvs;
	unsigned int generation;
	int rc;

	rc = hunt_nvram_partn_size(hunt,
				   NVRAM_PARTITION_TYPE_DYNCONFIG_DEFAULTS,
				   &defaults_size);
	if (rc)
		return false;

	generation = tlv_get_partition_generation(*dyncfg_buf);

	if (defaults_size > dyncfg_buf_size) {
		new_buf = realloc(*dyncfg_buf, defaults_size);
		if (!new_buf)
			return false;	/* dyncfg_buf contents is still ok */
		*dyncfg_buf = new_buf;
	} else {
		new_buf = *dyncfg_buf;
	}

	rc = hunt_nvs_read(NVRAM_PARTITION_TYPE_DYNCONFIG_DEFAULTS, hunt, 0,
			   new_buf, defaults_size);
	if (rc) {
		DBGC(hunt,
		     "Error %x reading the dynamic config default partition\n",
		     rc);
		return false;
	}

	rc = hunt_nvram_set_generation(new_buf, defaults_size, ++generation);
	if (rc) {
		DBGC(hunt,
		     "Error %x setting dynamic config new generation %d\n", rc,
		     generation);
		return false;
	}

	nvs = &dyncfg->nvs;
	nvs->block_size = nvs->size = defaults_size;
	rc = nvs->write(nvs, 0, new_buf, defaults_size);
	if (rc) {
		DBGC(hunt,
		     "Error %x writing the new dynamic config partition\n",
		     rc);
		return false;
	}

	dyncfg->settings_modified = 0;	/* Ignore changes to these settings */
	return true;
}

static int dyncfg_wipe ( struct settings *settings );

/* Recover from a corrupt dynamic config partition. */
static int dyncfg_recover(struct settings *settings, unsigned char *dyncfg_buf,
			  size_t dyncfg_buf_size)
{
	struct nvs_device *nvs;
	struct dyncfg *dyncfg;
	int rc;

	dyncfg = container_of(settings, struct dyncfg, settings);
	if (dyncfg_recover_from_partition(dyncfg, &dyncfg_buf, dyncfg_buf_size))
		return 0;

	rc = tlv_init_dynamic_config_buf((char *) dyncfg_buf, dyncfg_buf_size);
	if (rc)
		return rc;

	/* Write the empty valid config to flash now, since dyncfg_wipe below
	 * will re-read it from flash.
	 */
	nvs = &dyncfg->nvs;
	rc = nvs->write(nvs, 0, dyncfg_buf, dyncfg_buf_size);
	if (rc)
		return rc;

	rc = dyncfg_wipe(settings);
	return rc;
}

static int dyncfg_validate(struct settings *settings, unsigned char *dyncfg_buf,
			   size_t dyncfg_buf_size)
{
	nvram_partition_t partition;
	int rc = tlv_init_partition_from_buffer(&partition, dyncfg_buf,
						dyncfg_buf_size);

	if (rc == 0)	/* The dynamic config is OK. */
		return rc;

	printf("The dynamic config partition is not valid.\n");
	return dyncfg_recover(settings, dyncfg_buf, dyncfg_buf_size);
}

/* Fetch the whole dynamic config block. */
int dyncfg_read(struct settings *settings)
{
	int rc;
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	struct nvs_device *nvs = &dyncfg->nvs;
	struct sfgpxe_dev_info dev_info;
	unsigned char *buf;
	unsigned int port_count;
	uint32_t port_mode;

	init_dev_info_from_hunt(&dev_info, hunt);

	buf = malloc(nvs->size);
	if (buf == NULL)
		return -ENOMEM;

	rc = nvs->read(nvs, 0, buf, nvs->size);
	if (rc)
		goto out;
	rc = dyncfg_validate(settings, buf, nvs->size);
	if (rc)
		goto out;

	gpxe_dyncfg_get_msix_limit(
	    buf, nvs->size, &dev_info,
	    &dyncfg->msix_limit, DYNCFG_SETTING_DEFAULT_MSIX_LIMIT);
	gpxe_dyncfg_get_vf_msix_limit(
	    buf, nvs->size, &dev_info,
	    &dyncfg->vf_msix_limit, DYNCFG_SETTING_DEFAULT_VF_MSIX_LIMIT);
	gpxe_dyncfg_get_vf_count(
	    buf, nvs->size, &dev_info,
	    &dyncfg->vf_count, DYNCFG_SETTING_DEFAULT_VF_COUNT);
	gpxe_dyncfg_get_firmware_variant(
	    buf, nvs->size, &dev_info,
	    &dyncfg->firmware_variant, DYNCFG_SETTING_DEFAULT_FIRMWARE_VARIANT);
	gpxe_dyncfg_get_pf_count(
	    buf, nvs->size, &dev_info,
	    &dyncfg->pf_count, DYNCFG_SETTING_DEFAULT_PF_COUNT);
	/* The value read is max_pf_number, which is zero-based */
	dyncfg->pf_count++;

	gpxe_dyncfg_get_port_mode(
	    buf, nvs->size, &dev_info,
	    &dyncfg->port_mode, DYNCFG_SETTING_DEFAULT_PORT_MODE);
	set_generic_port_mode(dyncfg);
	port_mode = dyncfg->port_mode;
	if (port_mode == TLV_PORT_MODE_DEFAULT) {
		struct mcdi_port_modes modes;

		rc = efx_get_port_modes(&hunt->efx, &modes);
		if (rc)
			goto out;
		port_mode = modes.dflt;
	}
	port_count = get_port_count(port_mode);
	if (!port_count) {
		rc = -EIO;
		goto out;
	}

	dyncfg->pf_count /= port_count;
	gpxe_dyncfg_get_privilege_mask(
	    buf, nvs->size, &dev_info,
	    &dyncfg->privilege_mask, DYNCFG_SETTING_DEFAULT_PRIVILEGE_MASK);
	gpxe_dyncfg_get_pfiov(buf, nvs->size, &dev_info, &dyncfg->pfiov,
			      DYNCFG_SETTING_DEFAULT_PFIOV);
	gpxe_dyncfg_get_vlan_tag(buf, nvs->size, &dev_info, &dyncfg->vlan_tag,
				 DYNCFG_SETTING_DEFAULT_VLAN_TAG);

	gpxe_dyncfg_get_vswitch_type(buf, nvs->size, &dev_info,
				     &dyncfg->vswitch_type,
				     TLV_VSWITCH_TYPE_NONE);

	if (dyncfg->vlan_tag == DYNCFG_SETTING_DEFAULT_VLAN_TAG)
		dyncfg->vlan_enabled = 0;
	else
		dyncfg->vlan_enabled = 1;
	dyncfg->settings_modified = 0;

	/* V0 holds the MCFW version */
	hunt_get_vpd_ro(SF_VPD_KEYWORD('V', '0'), buf, nvs->size,
			dyncfg->vpd_v0, MAX_VPD_V0);
out:
	free(buf);
	return rc;
}

static inline int flush_dyncfg_to_buf(struct dyncfg *dyncfg, unsigned char *buf)
{
	int rc;
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	struct nvs_device *nvs = &dyncfg->nvs;
	struct net_device *netdev = hunt->efx.netdev;
	struct pci_device *pcidev = (struct pci_device *)netdev->dev;
	int pci_function = PCI_FUNC(pcidev->devfn);
	struct sfgpxe_dev_info dev_info;
	unsigned int port_count;
	uint32_t port_mode, old_port_mode = dyncfg->port_mode;

	init_dev_info_from_hunt(&dev_info, hunt);

	set_nic_port_mode(dyncfg);
	port_mode = dyncfg->port_mode;
	if (port_mode == TLV_PORT_MODE_DEFAULT) {
		struct mcdi_port_modes modes;

		rc = efx_get_port_modes(&hunt->efx, &modes);
		if (rc)
			return rc;
		port_mode = modes.dflt;
	}

	if ((rc = gpxe_dyncfg_set_msix_limit(
		    buf, nvs->size, &dev_info, dyncfg->msix_limit)))
		return rc;
	if ((rc = gpxe_dyncfg_set_vf_msix_limit(
		    buf, nvs->size, &dev_info, dyncfg->vf_msix_limit)))
		return rc;
	if ((rc = gpxe_dyncfg_set_vf_count(
		    buf, nvs->size, &dev_info, dyncfg->vf_count)))
		return rc;
	/* Only function 0 is allowed to set the shared firmware variant
	 * setting.
	 */
	if (pci_function == 0 &&
	    (rc = gpxe_dyncfg_set_firmware_variant(
		    buf, nvs->size, &dev_info, dyncfg->firmware_variant)))
		return rc;

	/* Only function 0 can set the (per-adapter) PF count.  This setting
	 * is stored as max_pf_number, which is zero-based.
	 * Do not update if multiple PFs have been configured, the menu
	 * prevents any changes in this case.
	 */
	port_count = get_port_count(port_mode);
	if (!port_count)
		return -EIO;

	if (pci_function == 0 && (dyncfg->port_mode != old_port_mode)) {
		dyncfg->pf_count = port_count - 1;
		rc = gpxe_dyncfg_set_pf_count(buf, nvs->size, &dev_info,
					      dyncfg->pf_count);
		if (rc)
			return rc;
	}

	if (pci_function == 0 && hunt_has_multiple_port_modes(netdev)) {
		rc = gpxe_dyncfg_set_port_mode(buf, nvs->size, &dev_info,
					       dyncfg->port_mode);
		if (rc)
			return rc;
	}

	if (pci_function == 0 &&
	    (rc = gpxe_dyncfg_set_privilege_mask(
		    buf, nvs->size, &dev_info, dyncfg->privilege_mask)))
		return rc;

	if ((rc = gpxe_dyncfg_set_pfiov(
		     buf, nvs->size, &dev_info, dyncfg->pfiov)))
		return rc;

	if ((rc = gpxe_dyncfg_set_vlan_tag(
		     buf, nvs->size, &dev_info, dyncfg->vlan_tag)))
		return rc;
	if (dyncfg->vlan_tag == DYNCFG_SETTING_DEFAULT_VLAN_TAG)
		dyncfg->vswitch_type = TLV_VSWITCH_TYPE_NONE;
	else
		dyncfg->vswitch_type = TLV_VSWITCH_TYPE_VLAN;

	if ((rc = gpxe_dyncfg_set_vswitch_type(
		     buf, nvs->size, &dev_info, dyncfg->vswitch_type)))
		return rc;

	if ((rc = update_per_pf_pcie_config(buf, nvs->size,
					    dev_info.pci_device_id & 0xffff,
					    port_mode)))
		return rc;

	return update_vf_aperture(buf, nvs->size, &dev_info, port_mode);
}

/* Write the whole dynamic config block. */
int dyncfg_flush(struct settings *settings)
{
	int rc = 0;
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);
	struct nvs_device *nvs = &dyncfg->nvs;
	struct hunt_nic *hunt = container_of(dyncfg, struct hunt_nic, dyncfg);
	unsigned char *buf;

	if (!dyncfg->settings_modified)
		return 0;

	buf = malloc(nvs->size);
	if (buf == NULL)
		return -ENOMEM;

	if ((rc = nvs->read(nvs, 0, buf, nvs->size)))
		goto out;

	if ((rc = flush_dyncfg_to_buf(dyncfg, buf)))
		goto out;

	rc = hunt_nvram_increment_generation(hunt, buf, nvs->size);
	if (rc)
		goto out;

	rc = nvs->write(nvs, 0, buf, nvs->size);

	dyncfg->settings_modified = 0;

out:
	free(buf);
	if (rc)
		DBGC(dyncfg, "%s rc=%x\n", __func__, rc);
	return rc;
}

static int dyncfg_wipe ( struct settings *settings )
{
	struct dyncfg *dyncfg =
		container_of(settings, struct dyncfg, settings);

	dyncfg->msix_limit = DYNCFG_SETTING_DEFAULT_MSIX_LIMIT;
	dyncfg->vf_msix_limit = DYNCFG_SETTING_DEFAULT_VF_MSIX_LIMIT;
	dyncfg->vf_count = DYNCFG_SETTING_DEFAULT_VF_COUNT;
	dyncfg->firmware_variant = DYNCFG_SETTING_DEFAULT_FIRMWARE_VARIANT;
	dyncfg->pf_count = DYNCFG_SETTING_DEFAULT_PF_COUNT;
	dyncfg->port_mode = DYNCFG_SETTING_DEFAULT_PORT_MODE;
	dyncfg->generic_port_mode = SFC_PORT_MODE_DEFAULT;
	dyncfg->pfiov = DYNCFG_SETTING_DEFAULT_PFIOV;
	dyncfg->vlan_tag = DYNCFG_SETTING_DEFAULT_VLAN_TAG;
	dyncfg->vswitch_type = TLV_VSWITCH_TYPE_NONE;
	dyncfg->settings_modified += 9;
	dyncfg->vlan_enabled = DYNCFG_SETTING_DEFAULT_VLAN_ENABLED;

	return dyncfg_flush(settings);
}

static int dyncfg_clear(struct settings *settings)
{
	struct dyncfg *dyncfg =	container_of(settings, struct dyncfg, settings);
	struct nvs_device *nvs = &dyncfg->nvs;
	unsigned char *buf;
	int rc;

	buf = malloc(nvs->size);
	if (buf == NULL)
		return -ENOMEM;

	/* Must read the existing buffer to get the current generation count */
	rc = nvs->read(nvs, 0, buf, nvs->size);
	if (rc)
		DBGC(dyncfg, "%s rc=%x\n", __func__, rc);

	rc = dyncfg_recover(settings, buf, nvs->size);
	free(buf);
	return rc;
}

static struct settings_operations dyncfg_settings_operations = {
	.store = hunt_dyncfg_store,
	.fetch = hunt_dyncfg_fetch,
	.clear = dyncfg_clear,
};

static int hunt_dyncfg_probe(struct hunt_nic *hunt)
{
	struct dyncfg *dyncfg = &hunt->dyncfg;
	struct nvs_device *nvs = &dyncfg->nvs;
	size_t dyncfg_size = 0;
	int rc;

	rc = hunt_nvram_partn_size(hunt, NVRAM_PARTITION_TYPE_DYNAMIC_CONFIG,
				   &dyncfg_size);
	if (rc)
		return rc;

	memset(dyncfg, 0, sizeof(*dyncfg));

	nvs->word_len_log2 = 0;
	nvs->block_size = nvs->size = dyncfg_size;
	nvs->read = hunt_dyncfg_nvs_read;
	nvs->write = hunt_dyncfg_nvs_write;

	settings_init(&hunt->dyncfg.settings, &dyncfg_settings_operations,
		      &hunt->efx.netdev->refcnt, 0);
	return 0;
}
#endif	/* SFC_IPXE */

/*******************************************************************************
 *
 *
 * Hardware initialization
 *
 *
 ******************************************************************************/
static int hunt_get_workarounds(struct hunt_nic *hunt, uint32_t *implemented,
				uint32_t *enabled)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_WORKAROUNDS_OUT_LEN);
	int rc;

	*implemented = *enabled = 0;

	req.emr_cmd = MC_CMD_GET_WORKAROUNDS;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);

	if (rc)
		return rc;

	if (req.emr_out_length_used < MC_CMD_GET_WORKAROUNDS_OUT_LEN)
		return -EMSGSIZE;

	*implemented = MCDI_DWORD(outbuf, GET_WORKAROUNDS_OUT_IMPLEMENTED);
	*enabled = MCDI_DWORD(outbuf, GET_WORKAROUNDS_OUT_ENABLED);
	return 0;
}

static int hunt_enable_workaround_35388(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(payload, MC_CMD_WORKAROUND_IN_LEN);

	req.emr_cmd = MC_CMD_WORKAROUND;
	req.emr_in_buf = payload;
	req.emr_in_length = MC_CMD_WORKAROUND_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, WORKAROUND_IN_TYPE,
		       MC_CMD_WORKAROUND_BUG35388);
	MCDI_SET_DWORD(req.emr_in_buf, WORKAROUND_IN_ENABLED, 1);

	/* If the firmware doesn't support this workaround, hunt_mcdi() will
	 * return -EINVAL from hunt_mcdi_request_poll().
	 */
	return hunt_mcdi(hunt, &req);
}

static int hunt_workaround_35388(struct hunt_nic *hunt)
{
	uint32_t implemented, enabled;
	int rc = hunt_get_workarounds(hunt, &implemented, &enabled);

	if (rc < 0)
		return 0;
	if (!(implemented & MC_CMD_GET_WORKAROUNDS_OUT_BUG35388))
		return 0;
	if (enabled & MC_CMD_GET_WORKAROUNDS_OUT_BUG35388)
		return 1;

	rc = hunt_enable_workaround_35388(hunt);
	if (rc == 0)
		return 1; /* Workaround is enabled */
	else
		return 0;
}

static int hunt_get_port_assignment(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_PORT_ASSIGNMENT_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_GET_PORT_ASSIGNMENT;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	hunt->efx.port = MCDI_DWORD(req.emr_out_buf,
				    GET_PORT_ASSIGNMENT_OUT_PORT);
	return 0;
}

#ifndef SFC_IPXE
static int hunt_get_pf_index(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_FUNCTION_INFO_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_GET_FUNCTION_INFO;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc)
		return rc;

	hunt->pf_index = MCDI_DWORD(outbuf, GET_FUNCTION_INFO_OUT_PF);

	return 0;
}

static void hunt_init_datapath_caps(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_CAPABILITIES_V2_OUT_LEN);
	int rc;

	BUILD_BUG_ON(MC_CMD_GET_CAPABILITIES_IN_LEN != 0);

	req.emr_cmd = MC_CMD_GET_CAPABILITIES;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc || (req.emr_out_length_used < MC_CMD_GET_CAPABILITIES_OUT_LEN))
		return;

	hunt->datapath_caps = MCDI_DWORD(outbuf, GET_CAPABILITIES_OUT_FLAGS1);

	if (req.emr_out_length_used >= MC_CMD_GET_CAPABILITIES_V2_OUT_LEN) {
		hunt->datapath_caps2 = MCDI_DWORD(outbuf,
				GET_CAPABILITIES_V2_OUT_FLAGS2);
	}
}

#endif	/* SFC_IPXE */
static int hunt_mac_addr(struct hunt_nic *hunt, uint8_t *ll_addr)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_MAC_ADDRESSES_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_GET_MAC_ADDRESSES;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = MC_CMD_GET_MAC_ADDRESSES_OUT_LEN;

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_GET_MAC_ADDRESSES_OUT_LEN)
		return -EMSGSIZE;

	memcpy(ll_addr,
	       MCDI_PTR(req.emr_out_buf, GET_MAC_ADDRESSES_OUT_MAC_ADDR_BASE),
	       ETH_ALEN);

	return 0;
}

static int hunt_get_phy_cfg(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_PHY_CFG_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_GET_PHY_CFG;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_GET_PHY_CFG_OUT_LEN)
		return -EMSGSIZE;

	hunt->phy_cap_mask = hunt->phy_cap =
		MCDI_DWORD(req.emr_out_buf, GET_PHY_CFG_OUT_SUPPORTED_CAP);
	DBGC2(hunt, "GET_PHY_CFG: flags=%x, caps=%x\n", rc, hunt->phy_cap);
	return 0;
}

static int hunt_driver_attach(struct hunt_nic *hunt, int attach)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_DRV_ATTACH_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_DRV_ATTACH_EXT_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_DRV_ATTACH;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	/* Set the PREBOOT flag to indicate later instances of attach should
	 * force an ENTITY RESET
	 */
	if (attach)
		attach |= 1 << MC_CMD_DRV_PREBOOT_LBN;

	MCDI_SET_DWORD(req.emr_in_buf, DRV_ATTACH_IN_NEW_STATE, attach);
	MCDI_SET_DWORD(req.emr_in_buf, DRV_ATTACH_IN_UPDATE, 1);
	MCDI_SET_DWORD(req.emr_in_buf, DRV_ATTACH_IN_FIRMWARE_ID,
		       MC_CMD_FW_DONT_CARE);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_DRV_ATTACH_OUT_LEN)
		return -EMSGSIZE;

	hunt->flags = MCDI_DWORD(outbuf, DRV_ATTACH_EXT_OUT_FUNC_FLAGS);

	return 0;
}

static int hunt_reset(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_ENTITY_RESET_IN_LEN);

	req.emr_cmd = MC_CMD_ENTITY_RESET;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_POPULATE_DWORD_1(req.emr_in_buf, ENTITY_RESET_IN_FLAG,
			      ENTITY_RESET_IN_FUNCTION_RESOURCE_RESET, 1);
	return hunt_mcdi(hunt, &req);
}

#ifndef SFC_IPXE
static void hunt_reboot(struct hunt_nic *hunt, unsigned int mode)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_REBOOT_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_REBOOT_MODE_OUT_LEN);

	req.emr_cmd = MC_CMD_REBOOT_MODE;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	MCDI_SET_DWORD(req.emr_in_buf, REBOOT_MODE_IN_VALUE, mode);
	hunt_mcdi(hunt, &req);

	req.emr_cmd = MC_CMD_REBOOT;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, REBOOT_IN_FLAGS, 0);
	hunt_mcdi_quiet(hunt, &req);
	DBGC(hunt, "MC rebooting...\n");
	/* Delay for the MC reboot to complete. */
	mdelay(100);
}
#endif	/* SFC_IPXE */

static void hunt_clear_udp_tunnel_ports(struct hunt_nic *hunt)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS_IN_LENMAX);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS_OUT_LEN);
	struct efx_mcdi_req_s req;
	int rc;

	memset(inbuf, 0, MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS_IN_LENMAX);
	MCDI_SET_DWORD(inbuf, SET_TUNNEL_ENCAP_UDP_PORTS_IN_FLAGS,
		(1 << MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS_IN_UNLOADING_LBN));

	req.emr_cmd = MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi_quiet(hunt, &req);
	if (rc)
		return;

	if (MCDI_DWORD(outbuf, SET_TUNNEL_ENCAP_UDP_PORTS_OUT_FLAGS) &
	    (1 << MC_CMD_SET_TUNNEL_ENCAP_UDP_PORTS_OUT_RESETTING_LBN)) {
		DBGC(hunt,
		     "Rebooting MC due to clearing UDP tunnel port list\n");
		/* Delay for the MC reboot to complete. */
		mdelay(100);
	}
}

static int hunt_set_mac(struct hunt_nic *hunt)
{
	struct net_device *netdev = hunt->efx.netdev;
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(payload, MC_CMD_SET_MAC_IN_LEN);
	unsigned int fcntl;
	int rc;

	req.emr_cmd = MC_CMD_SET_MAC;
	req.emr_in_buf = payload;
	req.emr_in_length = MC_CMD_SET_MAC_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, SET_MAC_IN_MTU,
		       EFX_MAC_FRAME_LEN(ETH_FRAME_LEN));
	MCDI_SET_DWORD(req.emr_in_buf, SET_MAC_IN_DRAIN, 0);
	memcpy(MCDI_PTR(req.emr_in_buf, SET_MAC_IN_ADDR),
	       netdev->ll_addr, ETH_ALEN);
	MCDI_SET_DWORD(req.emr_in_buf, SET_MAC_IN_REJECT, 0);

	/* If the PHY supports autnegotiation, then configure the MAC to match
	 * the negotiated settings. Otherwise force the MAC to TX and RX flow
	 * control.
	 */
	if (hunt->phy_cap_mask & (1 << MC_CMD_PHY_CAP_AN_LBN))
		fcntl = MC_CMD_FCNTL_AUTO;
	else
		fcntl = MC_CMD_FCNTL_BIDIR;
	MCDI_SET_DWORD(req.emr_in_buf, SET_MAC_IN_FCNTL, fcntl);

	rc = hunt_mcdi(hunt, &req);
	/* Ignore failure for permissions reasons */
	if (rc == -EPERM)
		rc = 0;
	return rc;
}

static int hunt_alloc_vis(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_ALLOC_VIS_IN_LEN);

	req.emr_cmd = MC_CMD_ALLOC_VIS;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, ALLOC_VIS_IN_MIN_VI_COUNT, 1);
	MCDI_SET_DWORD(req.emr_in_buf, ALLOC_VIS_IN_MAX_VI_COUNT, 1);

	return hunt_mcdi(hunt, &req);
}

static void hunt_free_vis(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	int rc;

	req.emr_cmd = MC_CMD_FREE_VIS;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		DBGC(hunt, "MC_CMD_FREE_VIS Failed\n");
}

/*******************************************************************************
 *
 *
 * Link state handling
 *
 *
 ******************************************************************************/
static int hunt_check_link(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_LINK_OUT_LEN);
	unsigned int flags, speed;
	bool up;
	int rc;
	static bool link_state = false;

	req.emr_cmd = MC_CMD_GET_LINK;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_GET_LINK_OUT_LEN)
		return -EMSGSIZE;

	flags = MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_FLAGS);
	up = !!(flags & (1 << MC_CMD_GET_LINK_OUT_LINK_UP_LBN));
	speed = MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_LINK_SPEED);

	/* Set netdev_link_*() based on the link status from the MC */
	if (up && speed)
		netdev_link_up(hunt->efx.netdev);
	else
		netdev_link_down(hunt->efx.netdev);

	if (up != link_state) {
		DBGC(hunt, "Link %s, flags=%x, our caps=%x, lpa=%x, speed=%d, fcntl=%x, mac_fault=%x\n",
		     (up? "up": "down"), flags,
		     MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_CAP),
		     MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_LP_CAP),
		     speed,
		     MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_FCNTL),
		     MCDI_DWORD(req.emr_out_buf, GET_LINK_OUT_MAC_FAULT));
		link_state = up;
	}

	return 0;
}

/*******************************************************************************
 *
 *
 * TX
 *
 *
 ******************************************************************************/
static int
hunt_tx_init(struct net_device *netdev, struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	dma_addr_t dma_addr;
	efx_qword_t *addr;
	MCDI_DECLARE_BUF(inbuf,
			 MC_CMD_INIT_TXQ_IN_LEN(EFX_TXQ_NBUFS(EFX_TXD_SIZE)));
	int rc, npages;

	rc = efx_hunt_tx_init(netdev, &dma_addr);
	if (rc != 0)
		return rc;

	npages = EFX_TXQ_NBUFS(EFX_TXD_SIZE);

	req.emr_cmd = MC_CMD_INIT_TXQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_INIT_TXQ_IN_LEN(npages);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_SIZE, EFX_TXD_SIZE);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_TARGET_EVQ, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_LABEL, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_INSTANCE, 0);

	MCDI_POPULATE_DWORD_6(req.emr_in_buf, INIT_TXQ_IN_FLAGS,
			      INIT_TXQ_IN_FLAG_BUFF_MODE, 0,
			      INIT_TXQ_IN_FLAG_IP_CSUM_DIS, 1,
			      INIT_TXQ_IN_FLAG_TCP_CSUM_DIS, 1,
			      INIT_TXQ_IN_FLAG_TCP_UDP_ONLY, 0,
			      INIT_TXQ_IN_CRC_MODE, 0,
			      INIT_TXQ_IN_FLAG_TIMESTAMP, 0);

	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_OWNER_ID, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_TXQ_IN_PORT_ID,
		       EVB_PORT_ID_ASSIGNED);

	addr = (efx_qword_t *) MCDI_PTR(req.emr_in_buf, INIT_TXQ_IN_DMA_ADDR);

	EFX_POPULATE_QWORD_2(*addr,
			     EFX_DWORD_1, (uint32_t)(dma_addr >> 32),
			     EFX_DWORD_0, (uint32_t)(dma_addr & 0xffffffff));

	return hunt_mcdi(hunt, &req);
}

static void hunt_tx_fini(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_FINI_TXQ_IN_LEN);
	struct efx_nic *efx = &hunt->efx;
	struct efx_tx_queue *txq = &efx->txq;
	int rc;

	req.emr_cmd = MC_CMD_FINI_TXQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, FINI_TXQ_IN_INSTANCE, 0);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		DBGC(hunt, "MC_CMD_FINI_TXQ Failed\n");

	efx_hunt_free_special_buffer(txq->ring,
				     sizeof(efx_tx_desc_t) * EFX_TXD_SIZE);
	txq->ring = NULL;
}

/*******************************************************************************
 *
 *
 * RX
 *
 *
 ******************************************************************************/
static int hunt_rx_filter_insert(struct net_device *netdev,
				 struct hunt_nic *hunt,
				 int multicast)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_FILTER_OP_IN_LEN);
	MCDI_DECLARE_BUF(outbuf, MC_CMD_FILTER_OP_OUT_LEN);
	int rc;
	uint64_t filter_id;
	(void) netdev;

	req.emr_cmd = MC_CMD_FILTER_OP;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_OP,
		       multicast ? MC_CMD_FILTER_OP_IN_OP_SUBSCRIBE
				 : MC_CMD_FILTER_OP_IN_OP_INSERT);
	MCDI_POPULATE_DWORD_1(req.emr_in_buf, FILTER_OP_IN_MATCH_FIELDS,
			      FILTER_OP_IN_MATCH_DST_MAC, 1);
	if (multicast)
		memset(MCDI_PTR(req.emr_in_buf, FILTER_OP_IN_DST_MAC),
		       0xff, ETH_ALEN);
	else
		memcpy(MCDI_PTR(req.emr_in_buf, FILTER_OP_IN_DST_MAC),
		       hunt->mac, ETH_ALEN);

	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_PORT_ID,
		       EVB_PORT_ID_ASSIGNED);
	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_RX_DEST,
			MC_CMD_FILTER_OP_IN_RX_DEST_HOST);
	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_RX_QUEUE, 0);
	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_RX_MODE, 0);
	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_TX_DEST,
		       MC_CMD_FILTER_OP_IN_TX_DEST_DEFAULT);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used <  MC_CMD_FILTER_OP_OUT_LEN)
		return -EIO;

	filter_id = MCDI_QWORD(req.emr_out_buf, FILTER_OP_OUT_HANDLE);
	if (multicast)
		hunt->mc_filter_id = filter_id;
	else
		hunt->uc_filter_id = filter_id;

	return 0;
}

static int hunt_rx_filter_remove(struct hunt_nic *hunt,
				 int multicast)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_FILTER_OP_IN_LEN);

	req.emr_cmd = MC_CMD_FILTER_OP;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, FILTER_OP_IN_OP,
		       multicast ? MC_CMD_FILTER_OP_IN_OP_UNSUBSCRIBE :
		       MC_CMD_FILTER_OP_IN_OP_REMOVE);
	MCDI_SET_QWORD(req.emr_in_buf, FILTER_OP_IN_HANDLE,
			  multicast ? hunt->mc_filter_id :
				      hunt->uc_filter_id);
	return hunt_mcdi(hunt, &req);
}

static int hunt_get_mac(struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(outbuf, MC_CMD_GET_MAC_ADDRESSES_OUT_LEN);
	int rc;

	req.emr_cmd = MC_CMD_GET_MAC_ADDRESSES;
	req.emr_in_buf = NULL;
	req.emr_in_length = 0;
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_INIT_EVQ_OUT_LEN)
		return -EMSGSIZE;

	memcpy(hunt->mac, MCDI_PTR(outbuf, GET_MAC_ADDRESSES_OUT_MAC_ADDR_BASE),
	       ETH_ALEN);
	return 0;
}

static int hunt_rx_filter_init(struct net_device *netdev,
			       struct hunt_nic *hunt)
{
	int rc = hunt_get_mac(hunt);

	if (rc != 0)
		return rc;

	rc = hunt_rx_filter_insert(netdev, hunt, 0);
	if (rc != 0)
		return rc;

	rc = hunt_rx_filter_insert(netdev, hunt, 1);
	if (rc != 0)
		hunt_rx_filter_remove(hunt, 0);

	return rc;
}

static int
hunt_rx_init(struct net_device *netdev,
	     struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	dma_addr_t dma_addr;
	efx_qword_t *addr;
	MCDI_DECLARE_BUF(inbuf,
			 MC_CMD_INIT_RXQ_IN_LEN(EFX_RXQ_NBUFS(EFX_RXD_SIZE)));
	int rc, npages;

	rc = efx_hunt_rx_init(netdev, &dma_addr);
	if (rc != 0)
		return rc;

	npages = EFX_RXQ_NBUFS(EFX_RXD_SIZE);

	req.emr_cmd = MC_CMD_INIT_RXQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_INIT_RXQ_IN_LEN(npages);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_SIZE, EFX_RXD_SIZE);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_TARGET_EVQ, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_LABEL, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_INSTANCE, 0);
	MCDI_POPULATE_DWORD_5(req.emr_in_buf, INIT_RXQ_IN_FLAGS,
				 INIT_RXQ_IN_FLAG_BUFF_MODE, 0,
				 INIT_RXQ_IN_FLAG_HDR_SPLIT, 0,
				 INIT_RXQ_IN_FLAG_TIMESTAMP, 0,
				 INIT_RXQ_IN_CRC_MODE, 0,
				 INIT_RXQ_IN_FLAG_PREFIX, 1);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_OWNER_ID, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_RXQ_IN_PORT_ID,
		       EVB_PORT_ID_ASSIGNED);

	addr = (efx_qword_t *) MCDI_PTR(req.emr_in_buf, INIT_RXQ_IN_DMA_ADDR);

	EFX_POPULATE_QWORD_2(*addr,
			     EFX_DWORD_1, (uint32_t)(dma_addr >> 32),
			     EFX_DWORD_0, (uint32_t)(dma_addr & 0xffffffff));
	return hunt_mcdi(hunt, &req);
}

static void hunt_rx_filter_fini(struct hunt_nic *hunt)
{
	hunt_rx_filter_remove(hunt, 0);
	hunt_rx_filter_remove(hunt, 1);
}

static void hunt_rx_fini(struct hunt_nic *hunt)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_FINI_RXQ_IN_LEN);
	struct efx_mcdi_req_s req;
	struct efx_nic *efx = &hunt->efx;
	struct efx_rx_queue *rxq = &efx->rxq;
	int rc;

	req.emr_cmd = MC_CMD_FINI_RXQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_FINI_RXQ_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, FINI_RXQ_IN_INSTANCE, 0);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		DBGC(hunt, "MC_CMD_FINI_RXQ Failed\n");

	efx_hunt_free_special_buffer(rxq->ring,
				     sizeof(efx_rx_desc_t) * EFX_RXD_SIZE);
	rxq->ring = NULL;
}

/*******************************************************************************
 *
 *
 * Event queues and interrupts
 *
 *
 ******************************************************************************/
static int
hunt_ev_init(struct net_device *netdev,
	     struct hunt_nic *hunt)
{
	struct efx_mcdi_req_s req;
	dma_addr_t dma_addr;
	efx_qword_t *addr;
	MCDI_DECLARE_BUF(inbuf,
			 MC_CMD_INIT_EVQ_IN_LEN(EFX_EVQ_NBUFS(EFX_EVQ_SIZE)));
	MCDI_DECLARE_BUF(outbuf, MC_CMD_INIT_EVQ_OUT_LEN);
	int rc, npages;

	rc = efx_hunt_ev_init(netdev, &dma_addr);
	if (rc != 0)
		return rc;

	npages = EFX_EVQ_NBUFS(EFX_EVQ_SIZE);

	req.emr_cmd = MC_CMD_INIT_EVQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_INIT_EVQ_IN_LEN(npages);
	req.emr_out_buf = outbuf;
	req.emr_out_length = sizeof(outbuf);

	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_SIZE, EFX_EVQ_SIZE);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_INSTANCE, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_IRQ_NUM, 0);

	MCDI_POPULATE_DWORD_6(req.emr_in_buf, INIT_EVQ_IN_FLAGS,
				 INIT_EVQ_IN_FLAG_INTERRUPTING, 1,
				 INIT_EVQ_IN_FLAG_RPTR_DOS, 0,
				 INIT_EVQ_IN_FLAG_INT_ARMD, 0,
				 INIT_EVQ_IN_FLAG_CUT_THRU, 0,
				 INIT_EVQ_IN_FLAG_RX_MERGE, 0,
				 INIT_EVQ_IN_FLAG_TX_MERGE, 0);

	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_TMR_MODE,
			  MC_CMD_INIT_EVQ_IN_TMR_MODE_DIS);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_TMR_LOAD, 0);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_TMR_RELOAD, 0);

	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_COUNT_MODE,
			  MC_CMD_INIT_EVQ_IN_COUNT_MODE_DIS);
	MCDI_SET_DWORD(req.emr_in_buf, INIT_EVQ_IN_COUNT_THRSHLD, 0);

	addr = (efx_qword_t *) MCDI_PTR(req.emr_in_buf, INIT_EVQ_IN_DMA_ADDR);

	EFX_POPULATE_QWORD_2(*addr,
			     EFX_DWORD_1, (uint32_t)(dma_addr >> 32),
			     EFX_DWORD_0, (uint32_t)(dma_addr & 0xffffffff));
	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		return rc;

	if (req.emr_out_length_used < MC_CMD_INIT_EVQ_OUT_LEN)
		return -EMSGSIZE;

	return 0;
}

static void hunt_ev_fini(struct hunt_nic *hunt)
{
	MCDI_DECLARE_BUF(inbuf, MC_CMD_FINI_EVQ_IN_LEN);
	struct efx_mcdi_req_s req;
	struct efx_nic *efx = &hunt->efx;
	struct efx_ev_queue *evq = &efx->evq;
	int rc;

	req.emr_cmd = MC_CMD_FINI_EVQ;
	req.emr_in_buf = inbuf;
	req.emr_in_length = sizeof(inbuf);
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD(req.emr_in_buf, FINI_EVQ_IN_INSTANCE, 0);

	rc = hunt_mcdi(hunt, &req);
	if (rc != 0)
		DBGC(hunt, "MC_CMD_FINI_EVQ Failed\n");

	efx_hunt_free_special_buffer(evq->ring,
				     sizeof(efx_event_t) * EFX_EVQ_SIZE);
	evq->ring = NULL;
}

static void
hunt_poll(struct net_device *netdev)
{
	struct hunt_nic *hunt = netdev_priv(netdev);

	/* If called while already polling, return immediately */
	if (hunt->efx.state & EFX_STATE_POLLING)
		return;
	hunt->efx.state |= EFX_STATE_POLLING;

	/* Poll link state */
	if (hunt->link_poll_timer + TICKS_PER_SEC < currticks()) {
		hunt->link_poll_timer = currticks();
		hunt_check_link(hunt);
	}

	/* Poll data path */
	efx_hunt_poll(netdev);

	hunt->efx.state &= ~EFX_STATE_POLLING;
}

/*******************************************************************************
 *
 *
 * Netdevice operations
 *
 *
 ******************************************************************************/
static int hunt_open(struct net_device *netdev)
{
	struct hunt_nic *hunt = netdev_priv(netdev);
	int rc;

	/* Allocate VIs */
	rc = hunt_alloc_vis(hunt);
	if (rc != 0)
		goto fail2;

	/* Initialize data path */
	rc = hunt_ev_init(netdev, hunt);
	if (rc != 0)
		goto fail3;

	rc = hunt_rx_init(netdev, hunt);
	if (rc != 0)
		goto fail4;

	rc = hunt_rx_filter_init(netdev, hunt);
	if (rc != 0)
		goto fail5;

	rc = hunt_tx_init(netdev, hunt);
	if (rc != 0)
		goto fail6;

	rc = efx_hunt_open(netdev);
	if (rc)
		goto fail7;

	rc = hunt_set_mac(hunt);
	if (rc)
		goto fail8;

	/* Mark the link as down before checking the link state because the
	 * latter might fail.
	 */
	netdev_link_down(netdev);
	hunt_check_link(hunt);

	DBGC2(hunt, "%s: open ok\n", netdev->name);
	return 0;

fail8:
	efx_hunt_close(netdev);
fail7:
	hunt_tx_fini(hunt);
fail6:
	hunt_rx_filter_fini(hunt);
fail5:
	hunt_rx_fini(hunt);
fail4:
	hunt_ev_fini(hunt);
fail3:
	hunt_free_vis(hunt);
fail2:
	DBGC2(hunt, "%s: %s\n", netdev->name, strerror(rc));
	return rc;
}


static void hunt_close(struct net_device *netdev)
{
	struct hunt_nic *hunt = netdev_priv(netdev);

	/* Stop datapath */
	efx_hunt_close(netdev);

	hunt_tx_fini(hunt);
	hunt_rx_fini(hunt);
	hunt_rx_filter_fini(hunt);
	hunt_ev_fini(hunt);

	hunt_free_vis(hunt);

	/* Reset hardware and detach */
	hunt_reset(hunt);
}


/*******************************************************************************
 *
 *
 * Public operations
 *
 *
 ******************************************************************************/
#ifndef SFC_IPXE
static void hunt_flash_leds ( struct net_device *netdev, enum sf_flash_state state )
{
	struct hunt_nic *hunt = netdev_priv(netdev);
	static const uint8_t states[] = {
		[FLASH_NONE] = MC_CMD_LED_DEFAULT,
		[FLASH_BLINK_ON] = MC_CMD_LED_ON,
		[FLASH_BLINK_OFF] = MC_CMD_LED_OFF,
	};
	struct efx_mcdi_req_s req;
	MCDI_DECLARE_BUF(inbuf, MC_CMD_SET_ID_LED_IN_LEN);

	req.emr_cmd = MC_CMD_SET_ID_LED;
	req.emr_in_buf = inbuf;
	req.emr_in_length = MC_CMD_SET_ID_LED_IN_LEN;
	req.emr_out_buf = NULL;
	req.emr_out_length = 0;

	MCDI_SET_DWORD ( req.emr_in_buf, SET_ID_LED_IN_STATE, states[state] );
	( void ) hunt_mcdi ( hunt, &req );
}

/* Count the number of bits set
 *
 * GCC supplies __builtin_popcount(), but that does not work since this
 * compiles as i386 code so the popcnt instruction is not used.
 *
 * This is from Hacker's Delight, p. 66, Figure 5-2
 */
static int popcount(unsigned x)
{
	x = x - ((x >> 1) & 0x55555555);
	x = (x & 0x33333333) + ((x >> 2) & 0x33333333);
	x = (x + (x >> 4)) & 0x0F0F0F0F;
	x = x + (x >> 8);
	x = x + (x >> 16);
	return x & 0x0000003F;
}

static int hunt_has_multiple_port_modes ( struct net_device *netdev )
{
	struct hunt_nic *hunt = netdev_priv(netdev);
	struct pci_device *pci = (struct pci_device *) netdev->dev;
	struct mcdi_port_modes port_modes;
	unsigned int generic_port_modes;
	int rc;

	rc = efx_get_port_modes(&hunt->efx, &port_modes);
	if (rc)
		return 0;
	generic_port_modes = to_generic_port_modes(pci, port_modes.modes);
	DBGC(hunt, "port modes %x -> %x supported %x\n", port_modes.modes,
	     generic_port_modes, generic_port_modes & SFC_SUPPORTED_PORT_MODES);

	/* The current bootROM only supports some specific settings plus
	 * default, so check that more than 1 port mode is available.
	 */
	return (popcount(generic_port_modes & SFC_SUPPORTED_PORT_MODES) > 1);
}

static int hunt_option_supported ( struct net_device *netdev, int tag )
{
	struct hunt_nic *hunt = netdev_priv(netdev);

	switch (tag) {
	case DHCP_EB_VF_COUNT:
	case DHCP_EB_VF_MSIX_LIMIT:
	case DHCP_EB_PFIOV_ENABLED:
	case DHCP_VLAN_ENABLED:
	case DHCP_EB_VLAN_TAG:
		return 1;
	case DHCP_EB_PF_COUNT:
	case DHCP_FW_VARIANT:
	case DHCP_INSECURE_FILTERS:
		return hunt_nic_is_primary ( hunt );
	case DHCP_EB_PORT_MODE:
		return hunt_nic_is_primary ( hunt ) &&
		       hunt_has_multiple_port_modes( netdev );
	}
	return efx_common_option_supported(netdev, tag);
}

static int hunt_ioctl ( struct net_device *netdev, enum net_ioctl_cmd cmd,
			unsigned long arg )
{
	struct hunt_nic *hunt = netdev_priv(netdev);
	struct mcdi_port_modes port_modes;
#ifdef MISC_CMD
	struct net_ioctl_peek *argp = (struct net_ioctl_peek *) arg;
#endif
	struct pci_device *pci;
	int rc = -ENOTSUP;
	unsigned int *intp;

	switch (cmd) {
	case NET_GET_PORT_NUMBER:
		intp = (unsigned int *) arg;
		*intp = hunt->efx.port;
		rc = 0;
		break;
	case NET_GET_PF_INDEX:
		intp = (unsigned int *) arg;
		*intp = hunt->pf_index;
		rc = 0;
		break;
	case NET_GET_PORT_MODES:
		intp = (unsigned int *) arg;
		rc = efx_get_port_modes(&hunt->efx, &port_modes);
		if (rc == 0) {
			pci = (struct pci_device *) netdev->dev;
			*intp = to_generic_port_modes(pci, port_modes.modes);
		}
		break;
	case NET_NIC_IS_READONLY:
		intp = (unsigned int *) arg;
		*intp = hunt_has_cap(hunt->datapath_caps2, TSA_BOUND);
		rc = 0;
		break;
#ifdef MISC_CMD
	case NET_PEEK:
		argp->value = _efx_readl(&hunt->efx, argp->address);
		rc = 0;
		break;
	case NET_READ32:
		rc = efx_read32(&hunt->efx, &argp->value, argp->address);
		break;
	case NET_IREAD:
		rc = efx_iread(&hunt->efx, &argp->value, argp->address);
		break;
	case NET_MCCMD:
		rc = efx_mccmd(&hunt->efx, (struct net_ioctl_mccmd *) arg);
#endif
	default:
		break;
	}
	return rc;
}
#endif	/* SFC_IPXE */

static struct net_device_operations hunt_operations = {
	.open                  = hunt_open,
	.close                 = hunt_close,
	.transmit              = efx_hunt_transmit,
	.poll                  = hunt_poll,
	.irq                   = efx_hunt_irq,
#ifndef SFC_IPXE
	.ioctl		       = hunt_ioctl,
	.sf_vpd		       = hunt_vpd,
	.sf_suspend            = efx_suspend,
	.sf_resume             = efx_resume,
	.sf_has_option_rom     = hunt_has_option_rom,
	.sf_has_dyncfg         = hunt_has_dyncfg,
	.sf_flash_leds	       = hunt_flash_leds,
	.sf_get_boot_type      = efx_get_boot_type,
	.sf_option_supported   = hunt_option_supported,
#endif	/* SFC_IPXE */
};

static int
#ifdef SFC_IPXE
hunt_probe(struct pci_device *pci)
#else
hunt_probe(struct pci_device *pci,
	   const struct pci_device_id *id __unused)
#endif	/* SFC_IPXE */
{
	struct net_device *netdev;
	struct hunt_nic *hunt;
	struct efx_nic *efx;
	int rc = 0;

	/* Create the network adapter */
	netdev = alloc_etherdev(sizeof(struct hunt_nic));
	if (!netdev) {
		rc = -ENOMEM;
		goto fail1;
	}

	/* Initialise the network adapter, and initialise private storage */
	netdev_init(netdev, &hunt_operations);
	pci_set_drvdata(pci, netdev);
	netdev->dev = &pci->dev;
	netdev->state |= NETDEV_IRQ_UNSUPPORTED;

	hunt = netdev_priv(netdev);
	memset(hunt, 0, sizeof(*hunt));
	efx = &hunt->efx;

	efx->type = &hunt_nic_type;

	/* Initialise efx datapath */
	efx_probe(netdev, EFX_HUNTINGTON);

	/* Initialise MCDI.  In case we are recovering from a crash, first
	 * cancel any outstanding request by sending a special message using the
	 * least significant bits of the 'high' (doorbell) register.
	 */
	_efx_writel(efx, cpu_to_le32(1), ER_DZ_MC_DB_HWRD);
	rc = hunt_mcdi_init(hunt);
	if (rc != 0)
		goto fail2;

	/* Reset (most) configuration for this function */
	rc = hunt_reset(hunt);
	if (rc != 0)
		goto fail3;

	/* Medford has a list of UDP tunnel ports that is populated by the
	 * driver. Avoid dropping any unencapsulated packets. This may cause
	 * an MC reboot.
	 */
	hunt_clear_udp_tunnel_ports(hunt);

	/* Enable the workaround for bug35388, if supported */
	efx->workaround_35388 = hunt_workaround_35388(hunt);

	/* Set the RX packet prefix size */
	efx->rx_prefix_size = ES_DZ_RX_PREFIX_SIZE;

	rc = hunt_get_port_assignment(hunt);
	if (rc != 0)
		goto fail3;

#ifndef SFC_IPXE
	rc = hunt_get_pf_index(hunt);
	if (rc != 0)
		goto fail3;

	hunt_init_datapath_caps(hunt);

#endif	/* SFC_IPXE */
	rc = hunt_mac_addr(hunt, netdev->ll_addr);
	if (rc != 0)
		goto fail4;

	rc = hunt_get_phy_cfg(hunt);
	if (rc != 0)
		goto fail5;

	rc = hunt_driver_attach(hunt, 1);
	if (rc != 0)
		goto fail5;

	/* If not exposing this network device, return successfully here */
	if (hunt->flags & (1 << MC_CMD_DRV_ATTACH_EXT_OUT_FLAG_NO_ACTIVE_PORT))
		return 0;

	if (hunt_nic_is_primary(hunt)) {
		hunt->next_primary = primary_nics;
		primary_nics = hunt;
		hunt->primary = hunt;
	} else {
		struct hunt_nic *other_hunt = primary_nics;

		while (other_hunt && !hunt->primary) {
			struct pci_device *other_pci = (struct pci_device *)
				other_hunt->efx.netdev->dev;
#ifdef SFC_IPXE
			/* Check if the seg:bus:dev parts match. */
			if (PCI_FIRST_FUNC(other_pci->busdevfn) ==
			    PCI_FIRST_FUNC(pci->busdevfn))
#else
			/* This could go wrong on systems with multiple PCI
			 * root bridges, for machines that match all these
			 * criteria:
			 * - Have multiple SFC NICs in them.
			 * - Get the same bus on different PCI segments.
			 * - Use multiple PFs on the 2nd or later NIC.
			 * In this case a flash read or update will use the
			 * wrong primary on the 2nd or later NIC. The wrong
			 * values would show up in the per-PF screen of the
			 * Ctrl-B menu, but those are read-only any way.
			 */
			if (other_pci->bus == pci->bus)
#endif	/* SFC_IPXE */
				hunt->primary = other_hunt;

			other_hunt = other_hunt->next_primary;
		}
		if (!hunt->primary) {
			rc = -EIO;
			goto fail6;
		}
	}

#ifndef SFC_IPXE
	rc = hunt_have_multiple_expcfg(hunt);
	if (rc < 0)
		goto fail6;
	hunt->multiple_expcfg = !!rc;

	rc = hunt_nvo_probe(hunt);
	if (rc != 0)
		goto fail6;

	rc = hunt_dyncfg_probe(hunt);
	if (rc != 0)
		goto fail7;
#endif	/* SFC_IPXE */

	rc = register_netdev(netdev);
	if (rc != 0)
		goto fail8;

#ifndef SFC_IPXE
	/* Deferred work with NVRAM */
	if (hunt->multiple_expcfg) {
		rc = register_nvo(&hunt->nvo, netdev_settings(netdev), NULL, 0);
	} else if (hunt_nic_is_primary(hunt)) {
		rc = register_nvo(&hunt->nvo, netdev_settings(netdev), NULL,
				  PF_EXPCFG_SIZE);
	} else {
		/* Non-primary PF with single EXPCFG get config from the
		 * larger block loaded by the primary PF.
		 */
		struct hunt_nic *hunt_primary = hunt->primary;
		struct nvo_block *nvo_primary = &hunt_primary->nvo;
		struct nvs_device *nvs_primary = &hunt_primary->expcfg_nvs;
		unsigned int offset;

		/* EXPCFG slice must not exceed the end of the block */
		offset = pci->devfn * PF_EXPCFG_SIZE;
		if (offset + PF_EXPCFG_SIZE <= nvs_primary->size)
			rc = register_nvo(&hunt->nvo, netdev_settings(netdev),
					  nvo_primary->data + offset,
					  PF_EXPCFG_SIZE);
		else
			rc = -EINVAL;
	}
	if (rc != 0)
		goto fail9;

	rc = hunt_register_dyncfg(hunt, netdev_settings(netdev));
	if (rc != 0)
		goto fail10;

#ifdef __i386__
	if ( is_hp ( pci ) ) {
		struct dyncfg *dyncfg = &hunt->dyncfg;
		struct nvs_device *nvs = &dyncfg->nvs;
		unsigned int dyncfg_size = nvs->size;
		struct sfgpxe_dev_info dev_info;
		unsigned char *buf;
		uint32_t ocsd_mode;

		init_dev_info_from_hunt( &dev_info, hunt );

		buf = malloc( dyncfg_size );
		if ( buf == NULL ) {
			rc = -ENOMEM;
			goto fail8;
		}

		rc = nvs->read( nvs, 0, buf, dyncfg_size );
		/* If dyncfg is read successfully, get the OCSD mode */
		if ( rc == 0 )
			gpxe_dyncfg_get_ocsd_mode( buf, dyncfg_size, &dev_info,
						   &ocsd_mode,
						   TLV_OCSD_ENABLED );
		free( buf );

		if ( ocsd_mode & TLV_OCSD_ENABLED )
			hp_ocsd_start( pci );
	}
#endif

	efx_register_clp(pci);
#endif	/* SFC_IPXE */

	DBG2("%s " PCI_FMT " ok\n", __func__, PCI_ARGS(pci));
#ifndef SFC_IPXE
	if (DBG_EXTRA && (PCI_FUNC(pci->devfn) == 0))
		sleep(5);
#endif
	return 0;

#ifndef SFC_IPXE
fail10:
	if (hunt->multiple_expcfg || hunt_nic_is_primary(hunt))
		unregister_nvo(&hunt->nvo);
fail9:
	unregister_netdev(netdev);
fail8:
	hunt_dyncfg_remove(hunt);
fail7:
#else
fail8:
#endif	/* SFC_IPXE */
fail6:
	(void) hunt_driver_attach(hunt, 0);
fail5:
fail4:
fail3:
	hunt_mcdi_fini(hunt);
fail2:
	efx_remove(netdev);
	netdev_put(netdev);
fail1:
	DBG2("%s " PCI_FMT " rc=%d\n", __func__, PCI_ARGS(pci), rc);
#ifndef SFC_IPXE
	if (DBG_EXTRA && (PCI_FUNC(pci->devfn) == 0))
		sleep(5);
#endif
	return rc;
}

#ifndef SFC_IPXE
extern int reboot_needed;
#endif	/* SFC_IPXE */

static void hunt_remove(struct pci_device *pci)
{
	struct net_device *netdev = pci_get_drvdata(pci);
	struct hunt_nic *hunt = netdev_priv(netdev);

#ifndef SFC_IPXE
	DBGC(pci, "%s(%p) netdev=%p hunt_nic=%p\n", __FUNCTION__, pci, netdev,
	     hunt);
#endif
	if (!(hunt->flags &
	      (1 << MC_CMD_DRV_ATTACH_EXT_OUT_FLAG_NO_ACTIVE_PORT))) {
#ifndef SFC_IPXE
		efx_unregister_clp(pci);

		if (hunt->multiple_expcfg || hunt_nic_is_primary(hunt))
			unregister_nvo(&hunt->nvo);
		hunt_unregister_dyncfg(hunt);
#endif	/* SFC_IPXE */
		/* The netdevice might still be open, so unregister it now
		 * before ripping stuff out from underneath.
		 */
		unregister_netdev(netdev);
#ifndef SFC_IPXE
		hunt_dyncfg_remove(hunt);
#endif	/* SFC_IPXE */
	}

	(void)hunt_driver_attach(hunt, 0);
#ifndef SFC_IPXE
	/* Some menu changes require a cold reboot of the PCI adapter.
	 * Devices are removed in reverse order of their creation, so reset
	 * the MC and then the PCI device for function 0.
	 */
	if (reboot_needed && (PCI_FUNC(pci->devfn) == 0)) {
		/* Recent firmwares to not need a power-on reboot any more.
		 * There is no way to query this from the firmware.
		 */
		if (pci_nic_is_medford2_or_newer(pci))
			hunt_reboot(hunt, MC_CMD_REBOOT_MODE_NORMAL);
		else
			hunt_reboot(hunt, MC_CMD_REBOOT_MODE_POR);
		pci_dev_reset(pci);
	}
#endif	/* SFC_IPXE */
	hunt_mcdi_fini(hunt);

	/* Destroy data path */
	efx_remove(netdev);

	netdev_nullify(netdev);
	netdev_put(netdev);
#ifndef SFC_IPXE
	DBGC(pci, "%s done\n", __FUNCTION__);
	if (DBG_EXTRA && (PCI_FUNC(pci->devfn) == 0))
		sleep(5);
#endif
}

const struct efx_nic_type hunt_nic_type = {
	.mcdi_rpc = _hunt_mcdi,
};

static struct pci_device_id hunt_nics[] = {
#ifdef SFC_IPXE
	PCI_ROM(0x1924, 0x0a03, "SFC9220", "Solarflare SFN8xxx Adapter", 0),
	PCI_ROM(0x1924, 0x0b03, "SFC9250", "Solarflare X25xx Adapter", 0),
#else
	PCI_ROM(0x1924, 0x0903, "SFC9120", "Solarflare"),
	PCI_ROM(0x1924, 0x0923, "SFC9140", "Solarflare"),
	PCI_ROM(0x1924, 0x0a03, "SFC9220", "Solarflare"),
	PCI_ROM(0x1924, 0x0b03, "SFC9250", "Solarflare"),
#endif	/* SFC_IPXE */
};

struct pci_driver hunt_driver __pci_driver = {
	.ids = hunt_nics,
	.id_count = ARRAY_SIZE(hunt_nics),
	.probe = hunt_probe,
	.remove = hunt_remove,
};
