/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdio.h>
#include <stdlib.h>
#include <gpxe/netdevice.h>
#include <gpxe/command.h>
#include <gpxe/clp.h>

static int read_ioctl_exec ( enum net_ioctl_cmd cmd, int argc, char **argv )
{
	struct net_device *dev = last_opened_netdev();
	struct net_ioctl_peek buf;
	int rc;

	if (!dev || (argc != 2)) {
		printf("Usage:\n  %s <address>\nMust open a network device "
		       "before calling this command.\n\n", argv[0]);
		return 1;
	}

	buf.address = strtoul(argv[1], NULL, 16);
	rc = netdev_ioctl(dev, cmd, (unsigned long) &buf);
	if (rc)
		printf("%s 0x%x failed, rc=0x%x\n", argv[0], buf.address, rc);
	else
		printf("%s 0x%x = 0x%x\n", argv[0], buf.address, buf.value);
	
	/* Can never return success by definition */
	return 1;
}

static int peek_exec ( int argc, char **argv )
{
	return read_ioctl_exec ( NET_PEEK, argc, argv );
}

static int read32_exec ( int argc, char **argv )
{
	return read_ioctl_exec ( NET_READ32, argc, argv );
}

static int iread_exec ( int argc, char **argv )
{
	return read_ioctl_exec ( NET_IREAD, argc, argv );
}

static int mccmd_exec ( int argc, char **argv )
{
	struct net_device *dev = last_opened_netdev();
	struct net_ioctl_mccmd mccmd;
	u32 *inbuf;
	int rc, i;

	if (!dev || (argc < 2)) {
		printf("Usage:\n  %s <cmd> [word...]\nMust open a network "
		       "device before calling this command.\n\n", argv[0]);
		return 1;
	}

	mccmd.cmd = strtoul(argv[1], NULL, 16);
	mccmd.inlen = sizeof(u32) * (argc - 2);
	inbuf = malloc(mccmd.inlen);
	if (!inbuf) {
		printf("%s failed to malloc %d bytes\n", argv[0], mccmd.inlen);
		return 1;
	}
	for (i=2; i < argc; i++)
		inbuf[i-2] = strtoul(argv[i], NULL, 16);
	mccmd.buffer = inbuf;

	rc = netdev_ioctl(dev, NET_MCCMD, (unsigned long) &mccmd);
	if (rc)
		printf("%s 0x%x failed, rc=%d\n", argv[0], mccmd.cmd, rc);
	else {
		printf("%s 0x%x len %d result:\n", argv[0], mccmd.cmd,
		       mccmd.outlen);
		for (i=0; i < (int) mccmd.outlen; i += sizeof(u32))
			printf("0x%08x%c", mccmd.buffer[i],
			       (i && (i % (8*sizeof(u32)) == 0)? '\n': ' '));
		if (i % (8*sizeof(u32)) != 0)
			printf("\n");
	}
	free(inbuf);

	/* Can never return success by definition */
	return 1;
}

static void hex_dump(const char *data, unsigned int len)
{
	unsigned int offset;

	printf("data=");
	for (offset = 0 ; offset < len ; offset++)
		printf(" 0x%2x", data[offset] & 0xff);
	printf("\n");
}

static int clp_exec ( int argc, char **argv )
{
	struct clp_response response;
	char command[CLP_MAX_STR_LEN] = "";
	int i;

	for (i=1; i < argc; i++)
		snprintf(&command[strlen(command)], CLP_MAX_STR_LEN, "%s%c", argv[i],
			 (i == argc-1? '\0': ' '));

	/* This is what c_clp_entry() does */
	memset(&response, 0, sizeof(response));
	response.string = CLP_STRING_UNSUPPORTED;
	response.status = CLP_STATUS_UNSUPPORTED;

	clp_parse(&response, command);

	printf("%s response: status=0x%x flags=%d length=%d ", argv[0],
	       response.status, response.flags, response.length);
	if (response.flags & CLP_FORMAT_OEMHP_BINARY)
		hex_dump(response.string, response.length);
	else
		printf("string='%s'\n", response.string);

	return (response.status == CLP_STATUS_SUCCESS? 0: response.status);
}

struct command misc_command[] __command = {
	{
		.name = "peek",
		.exec = peek_exec,
	},
	{
		.name = "read32",
		.exec = read32_exec,
	},
	{
		.name = "iread",
		.exec = iread_exec,
	},
	{
		.name = "mccmd",
		.exec = mccmd_exec,
	},
	{
		.name = "clp",
		.exec = clp_exec,
	},
};
