/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Copyright 2008-2009: Solarflare Communications Inc,
 *                      9501 Jeronimo Road, Suite 250,
 *                      Irvine, CA 92618, USA
 *
 * Author of original gPXE implementation:
 * Michael Brown <mbrown@fensystems.co.uk>
 *
 *
 * Maintained by
 *          Solarflare Communications <linux-drivers@solarflare.com>
 *
 * Authors:
 *          Igor Baryshev <Igor.Baryshev@oktetlabs.ru>
 *          OKTET Labs Ltd, Russia,
 *          http://oktetlabs.ru, <info@oktetlabs.ru>
 *          by request of Solarflare Communications
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <gpxe/malloc.h>

#include <gpxe/pci.h>
#include <gpxe/device.h>
#include <gpxe/netdevice.h>
#include <gpxe/dhcp.h>

#include <realmode.h>

#include <usr/ifmgmt.h>
#include <usr/route.h>
#include <usr/dhcpmgmt.h>
#include <usr/autoboot.h>
#include <usr/bootconnect.h>

#include <gpxe/ibft.h>

#include <gpxe/image.h>
#include <gpxe/uri.h>
#include <usr/imgmgmt.h>

#include <gpxe/settings.h>
#include <gpxe/sf_settings.h>
#include <gpxe/net/sfc/pxe.h>
#include <gpxe/sanity_checks.h>
#include <gpxe/instance.h>
#include <gpxe/iscsi.h>

#include <console.h>
#include <gpxe/keys.h>
#include <gpxe/process.h>

/** Time to wait for link-up */
#define LINK_WAIT_S 30

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#ifdef BCV_PROTO_ISCSI
enum mpio_status {
	MPIO_NONE = 0,
	MPIO_DHCP_OK = 1,
	MPIO_ISCSI_OK = 2,
};

struct mpio_device {
	struct list_head	list;
	enum mpio_status	status;
	uint64_t 		priority;
	struct net_device	*net_dev;
};

static struct list_head mpio_devices = LIST_HEAD_INIT(mpio_devices);
static int mpio_devices_num = 0;
#endif

/* Wait for the device link to come up */
static int linkupwait(struct net_device *dev, long delay)
{
	int	rc = 0;

	if (delay > 0) {
		printf ( "Waiting %ld seconds for link up on %s...", delay,
			 dev->name );
		delay *= 1000;
		rc = iflinkwait ( dev, delay );
		if ( rc != 0 )
			printf ( " no link detected\n" );
		else
			printf ( " ok\n" );
	}

	return rc;
}

/* Wait for a period of time after the link has come up */
static int linkupdelay(struct net_device *dev, long delay)
{
	if (delay > 0) {
		printf("Waiting a further %ld seconds for %s post link up...",
		       delay, dev->name);
		delay *= 100;
		while (delay-- > 0) {
			if (iskey() && getchar() == CTRL_C) {
				printf(" cancelled\n");
				return -ECANCELED;
			}

			step();
			mdelay(10);
		}

		printf(" ok\n");
	}

	return 0;
}

/**
 * Handle link-up delay
 *
 * This is factored out from sf_netboot_prep() because the hardware restart
 * during delayed relocation may cause a need for this wait as well.  Ugly.
 *
 */
int sf_linkwait(struct net_device *dev)
{
	struct settings *settings = netdev_nvo_settings(dev);
	int delay, rc;
	long seconds, elapsed;

	/* Wait for LINK_WAIT_S since the device was opened */
	if (!netdev_link_ok(dev)) {
		elapsed = (currticks() - dev->open_ticks) / ticks_per_sec();
		delay = LINK_WAIT_S - elapsed;
		if ((rc = linkupwait(dev, delay)) != 0)
			return rc;

	} else if (sf_netdev_link_changed(dev))
		/* Ensure that we display something if the link up
		 * has already happened */
		printf("%s link up\n", dev->name);

	/* And ensure the link has been up for a sufficient period */
	seconds = sf_sett_get_intz(settings, &linkup_delay_setting);
	if (seconds > MAX_LINKUP_DELAY)
		seconds = MAX_LINKUP_DELAY;
	elapsed = (currticks() - dev->linkup_ticks) / ticks_per_sec();
	delay = seconds - elapsed;

	return linkupdelay(dev, delay);
}

/**
 * Check whether or not we have a usable PXE menu
 *
 * @v dev               Network device
 * @ret have_menu       A usable PXE menu is present
 */
static int sf_have_pxe_menu ( struct net_device *dev ) {
        struct setting vendor_class_id_setting
                = { .tag = DHCP_VENDOR_CLASS_ID };
        struct setting pxe_discovery_control_setting
                = { .tag = DHCP_PXE_DISCOVERY_CONTROL };
        struct setting pxe_boot_menu_setting
                = { .tag = DHCP_PXE_BOOT_MENU };
        char buf[ 10 /* "PXEClient" + NUL */ ];
        unsigned int pxe_discovery_control;
	struct settings *settings;

	/* The VCI string must be in a settings block from a stored packet,
	 * not from the EXPROM config settings block, so choose a suitable
	 * block here.  The other settings below can be fetched from any
	 * block as they do not conflict with the EXPROM config block.
	 */
	settings = netdev_pxebs_settings ( dev );
	if ( ! settings )
		settings = netdev_proxydhcp_settings ( dev );
	if ( ! settings )
		settings = netdev_dhcp_settings ( dev );

        fetch_string_setting ( settings, &vendor_class_id_setting,
                               buf, sizeof ( buf ) );
        pxe_discovery_control =
                fetch_uintz_setting ( NULL, &pxe_discovery_control_setting );

        return ( ( strcmp ( buf, "PXEClient" ) == 0 ) &&
                 setting_exists ( NULL, &pxe_boot_menu_setting ) &&
                 ( ! ( ( pxe_discovery_control & PXEBS_SKIP ) &&
                       setting_exists ( NULL, &filename_setting ) ) ) );
}

/**
 * Parse boot menu and perform Boot Server Discovery if required
 *
 * @v dev               Network device
 * @ret rc              Return status code
 */
static int sf_menu_pxebs ( struct net_device *dev ) {
	char buf_bs_menu[256], buf_bs_prompt[256];
	int len_bs_menu, len_bs_prompt;
	struct settings *settings;
	unsigned int pxe_type;

	settings = netdev_proxydhcp_settings ( dev );
	if ( ! settings )
		settings = netdev_dhcp_settings ( dev );
	if ( ! settings ) {
		DBG ( "%s: No DHCP block found in %s\n", __func__,
		      dev->name );
		return ( -1 );
	}

	len_bs_menu = fetch_string_setting ( settings, &boot_menu_setting,
					     buf_bs_menu,
					     sizeof ( buf_bs_menu ) );
	len_bs_prompt = fetch_string_setting ( settings,
					       &boot_menu_prompt_setting,
					       buf_bs_prompt,
					       sizeof ( buf_bs_prompt ) );

	pxe_type = pxe_get_bootserver_type ( buf_bs_menu, len_bs_menu,
					     buf_bs_prompt, len_bs_prompt );

	/* Return immediately if local boot selected */
	if ( ! pxe_type )
		return 0;

	/* Attempt PXE Boot Server Discovery */
	return pxebs ( dev, pxe_type );
}

/**
 * PXE Boot from a network device
 *
 * @v dev 		Network device
 * @ret rc		Return status code
 */
static int
sf_netboot_pxe(struct net_device *dev)
{
	char buf[DHCP_OPT_LEN_MAX];
	struct in_addr  next_server;
	struct settings  *settings = NULL;
	char *filename;
	int rc;

#ifdef BCV_PROTO_ISCSI
	ibft_clear();
#endif

	DBG("Initiating PXE boot over device: %s\n", dev->name);

	rc = ifopen(dev);
	if (rc != 0)
		return rc;

	if ((rc = sf_linkwait(dev)) != 0)
		return rc;

	booting_netdev = dev;

	if ((rc = dhcp(dev)) != 0) {
		printf("Error: DHCP completion status: %s\n",
		        strerror(rc));
		return rc;
	}

	route_dev(dev);

	if ( sf_have_pxe_menu ( dev ) ) {
		printf ( "Booting from PXE menu\n" );
		rc = sf_menu_pxebs ( dev );
		if ( rc )
			return rc;

		settings = netdev_pxebs_settings ( dev );
	}

	if ( ! settings ) {
		/* Prefer proxyDHCP settings */
		settings = netdev_proxydhcp_settings ( dev );
		if ( ! settings )
			settings = netdev_dhcp_settings ( dev );

		if ( ! settings ) {
			DBG ("netboot_pxe aborted: %s has no dhcp settings",
			     dev->name );
			return -ENOENT;
		}
	}

	if (fetch_ipv4_setting(settings, &next_server_setting,
			       &next_server) < 0 &&
	    fetch_ipv4_setting(settings, &dhcp_server_setting,
			       &next_server) < 0)
		/* Assume that filename is a URI */
		next_server.s_addr = 0;

	fetch_string_setting ( NULL, &filename_setting, buf, sizeof ( buf ) );
	if ( ! buf[0] ) {
		printf("Error: No filename specified.\n");
		return -ENOENT;
	}

	/* Expand filename setting */
	filename = expand_settings ( buf );
	if ( ! filename )
		filename = buf;

	printf ( "Booting from filename \"%s\"\n", filename );
	rc = boot_next_server_and_filename ( next_server, filename );

	free ( filename );
	return rc;
}

#ifdef BCV_PROTO_ISCSI
/**
 * If dhcp is configured for iSCSI, initialise the net_device and obtain a lease
 *
 * @v dev    		Network device
 * @ret rc   		Return status code
 */
static int
sf_netboot_iscsi_dhcp(struct net_device *dev)
{
	struct settings *settings = netdev_nvo_settings(dev);
	char buf[DHCP_OPT_LEN_MAX];
	int rc;

	if (sf_sett_get_intz(settings, &initiator_dhcp_setting) ||
	    sf_sett_get_intz(settings, &target_dhcp_setting)) {

		if ((rc = dhcp(dev)) != 0) {
			printf("Error: unable to obtain dhcp lease for %s: %s\n",
			       dev->name, strerror(rc));
			return rc;
		}
	}

	/* The dhcp lease is worthless without a root_path */
	rc = sf_root_path_fetch(dev, buf, sizeof(buf));
	if (rc <= 0) {
		printf("Error: %s has no root path\n", dev->name);
		return -ENOENT;
	}

	return 0;
}

/**
 * iSCSI connect to a single net_device. You must have called
 * sf_netboot_iscsi_dhcp() first
 *
 * @v dev    		Network device
 * @ret rc   		Return status code
 */
static int
sf_netboot_iscsi_connect(struct net_device *dev)
{
	char buf[DHCP_OPT_LEN_MAX];
	int rc;

	rc = sf_root_path_fetch(dev, buf, sizeof(buf));
	if (rc <= 0)
		return -ENOENT;

	if ((rc = connect_root_path(buf)) != 0) {
		printf("Error: %s failed to connect to %s: %s\n",
		       dev->name, buf, strerror(rc));
		return rc;
	}

	return 0;
}

/* Produce an ordered list of iscsi net_device's */
static int
sf_mpio_list(void) {
	struct net_device	*net_dev, *_nd;
	struct mpio_device	*mpio_dev, *next;
	int			inserted;
	uint64_t 		priority;


	/* Insert sorted mpio_devices into the mpio_devices list */
	list_for_each_entry_safe(net_dev, _nd, &net_devices, list) {
		priority = sf_mpio_priority_key(net_dev);
		if (priority == PRIORITY_KEY_NONE)
			continue;

		if ((mpio_dev = malloc(sizeof(*mpio_dev))) == NULL)
			return -ENOMEM;

		mpio_dev->status = MPIO_NONE;
		mpio_dev->net_dev = net_dev;
		mpio_dev->priority = priority;

		/* Find the insertion point */
		inserted = 0;
		list_for_each_entry(next, &mpio_devices, list) {
			if (next->priority > mpio_dev->priority) {
				__list_add(&mpio_dev->list,
					   next->list.prev,
					   &next->list);
				inserted = 1;
				break;
			}
		}
		if (!inserted)
			list_add_tail(&mpio_dev->list, &mpio_devices);
	}

	if (list_empty(&mpio_devices))
		return -ENODEV;

	return 0;
}

/**
 * Attempt a round of MPIO iSCSI boot
 *
 * @v mpio_devices	A list of devices without failed dhcp
 * @v rc		Return code
 */
static int
sf_netboot_iscsi_mpio(void)
{
	struct mpio_device	*mdev;
	struct net_device	*net_dev, *connected_dev = NULL;
	char			*iqn, *initiator_iqn = NULL;
	int			rc;

	list_for_each_entry(mdev, &mpio_devices, list) {
		net_dev = mdev->net_dev;
		if (!(net_dev->state & NETDEV_OPEN))
			continue;

		rc = sf_linkwait(net_dev);
		if (rc == -ETIMEDOUT) {
			/* Close this device since it's now useless */
			ifclose(net_dev);
			continue;
		} else if (rc != 0)
			return rc;

		booting_netdev = net_dev;

		if (mdev->status == MPIO_NONE) {
			rc = sf_netboot_iscsi_dhcp(net_dev);
			if (rc == 0) {
				mdev->status = MPIO_DHCP_OK;
				route_dev(net_dev);
			}
		}

		/* Warn the user if the Initiator IQN differs */
		iqn = iscsi_initiator_iqn(net_dev);
		if (initiator_iqn == NULL)
			initiator_iqn = iqn;
		else {
			if (strcmp(initiator_iqn, iqn) != 0)
				printf("Warning: The iSCSI Initiator IQN "
				       "differs between adapters.\n"
				       "         This can lead to "
				       "unexpected results.\n");
			free(iqn);
		}

		if (mdev->status == MPIO_DHCP_OK && connected_dev == NULL) {
			rc = sf_netboot_iscsi_connect(net_dev);
			if (rc == 0) {
				/* iBFT has already been filled in */
				mdev->status = MPIO_ISCSI_OK;
				connected_dev = net_dev;
			}
		}
	}

	free(initiator_iqn);

	if (connected_dev != NULL) {
		/* Fill out the ibft for any DHCP_OK devices */
		list_for_each_entry(mdev, &mpio_devices, list) {
			if (mdev->status == MPIO_DHCP_OK) {
				net_dev = mdev->net_dev;
				if ((rc = ibft_fill_netdev(net_dev)) != 0) {
					printf("Error: Failed to populate "
					       "iBFT with device %s: %s\n",
					       net_dev->name, strerror(rc));
				}
			}
		}

		/* Mark this net_device as the boot device */
		booting_netdev = connected_dev;
		return 0;
	} else
		return -EAGAIN;
}

/**
 * iSCSI connect to a network device
 *
 * @v net_dev 		Network device
 * @ret rc		Return status code
 */
static int
sf_netboot_iscsi(struct net_device *net_dev)
{
	struct mpio_device	*mpio_dev;
	int			i, rc;
	long			mpio_attempts;

	rc = sf_mpio_list();
	if (rc < 0) {
		DBG("Failed to initialize list of MPIO devices: %s\n",
		    strerror(rc));
		return rc;
	}

	/* Only attempt an iSCSI boot if we're the primary MPIO device */
	assert(!list_empty(&mpio_devices));
	mpio_dev = list_entry(mpio_devices.next, struct mpio_device, list);
	if (mpio_dev->net_dev != net_dev) {
		DBG("Current MPIO device is not primary: skipping\n" );
		return 0;
	}

	/* Prepare all net_devices */
	list_for_each_entry(mpio_dev, &mpio_devices, list) {
		rc = ifopen(mpio_dev->net_dev);
		if (rc != 0)
			printf("Unable to initialise %s: %s\n",
			       mpio_dev->net_dev->name, strerror(rc));
	}

	mpio_attempts = sf_sett_get_intz(netdev_nvo_settings(net_dev),
					 &mpio_attempts_setting);

	for (i = 0; i < mpio_attempts; i++) {
		rc = sf_netboot_iscsi_mpio();
		if (rc == 0)
			return 0;
	}

	return rc;
}
#endif

/** SFC replacement for original autoboot() */
int sf_autoboot(void)
{
	struct net_device	*netdev;
	int              	rc = 0;

	/* Find our netdevice */
	netdev = instance_netdevice();
	if (!netdev) {
		DBG("Failed to find boot device\n");
		goto cleanup;
	}

	if ((rc = sf_nvo_checks_netdev(netdev)) < 0) {
		printf("PXE boot failed: invalid boot configuration: %s\n",
		       sc_err_msgs[-rc]);
		goto cleanup;
	}

	/* Boot */
	rc = sf_netboot_pxe(netdev);
	if (rc != 0)
		printf("PXE boot failed: %s\n", strerror(rc));

cleanup:
	DBG("Autoboot done.\n");
	return rc;
}

#ifdef BCV_PROTO_ISCSI
void
sf_bootconnect(void)
{
	struct net_device	*netdev;
	int              	rc = 0;

	/* Find our netdevice */
	netdev = instance_netdevice();
	if (!netdev) {
		printf("Failed to find boot device\n");
		rc = -ENOENT;
		goto cleanup;
	}

	if ( (rc = sf_nvo_checks_netdev(netdev)) < 0) {
		printf("iSCSI boot failed: invalid boot configuration: %s\n",
		       sc_err_msgs[-rc]);
		goto cleanup;
	}

	/* Connect */
	rc = sf_netboot_iscsi(netdev);
	if (rc != 0)
		printf("iSCSI connect failed: %s\n", strerror(rc));

cleanup:
	while (!list_empty(&mpio_devices)) {
		struct mpio_device  *mdev;
		mdev = list_entry(mpio_devices.next, struct mpio_device, list);
		list_del(mpio_devices.next);
		free(mdev);
	}
	mpio_devices_num = 0;

	if (rc != 0)
		ibft_clear();

	DBG("Boot connection done.\n");
	return;
}
#endif
