/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <gpxe/malloc.h>
#include <gpxe/sf_settings.h>
#include <gpxe/dhcp.h>
#include <gpxe/netdevice.h>
#include <gpxe/nvo.h>
#include <gpxe/pci.h>
#include <drivers/net/mii.h>

/** @file
 *
 * Additional core settings for configuring Solarflare NICs
 *
 */

/** See settings descriptions in include/gpxe/sf_settings.h */
struct setting mpio_attempts_setting __setting = {
	.name = "mpio-attempts",
	.description = "MPIO boot attempts",
	.tag = DHCP_EB_MPIO_ATTEMPTS,
	.type = &setting_type_uint8,
};

struct setting banner_delay_setting __setting = {
	.name = "banner-delay",
	.description = "Banner delay time",
	.tag = DHCP_EB_BANNER_DELAY,
	.type = &setting_type_uint8,
};

struct setting bootskip_delay_setting __setting = {
	.name = "bootskip-delay",
	.description = "Boot skip delay time",
	.tag = DHCP_EB_BOOTSKIP_DELAY,
	.type = &setting_type_uint8,
};

struct setting linkup_delay_setting __setting = {
	.name = "linkup-delay",
	.description = "Link-up delay time",
	.tag = DHCP_EB_LINKUP_DELAY,
	.type = &setting_type_uint8,
};

struct setting lun_retries_setting __setting = {
	.name = "lun-retries",
	.description = "LUN busy retries",
	.tag = DHCP_EB_LUN_BUSY_RETRIES,
	.type = &setting_type_uint8,
};

struct setting dhcp_vendor_class_setting __setting = {
	.name = "dhcp-vendor",
	.description = "DHCP Vendor Class ID",
	.tag = DHCP_VENDOR_CLASS_ID,
	.type = &setting_type_string,
};

struct setting boot_type_setting __setting = {
	.name = "boot-type",
	.description = "Boot type",
	.tag = DHCP_EB_BOOT_TYPE,
	.type = &setting_type_uint8,
};

struct setting mpio_priority_setting __setting = {
	.name = "mpio-priority",
	.description = "MPIO priority",
	.tag = DHCP_EB_MPIO_PRIORITY,
	.type = &setting_type_uint8,
};

struct setting link_speed_setting __setting = {
	.name = "link-speed",
	.description = "Link speed",
	.tag = DHCP_EB_LINK_SPEED,
	.type = &setting_type_uint8,
};

struct setting chap_setting __setting = {
	.name = "chap",
	.description = "CHAP authentication",
	.tag = DHCP_EB_CHAP,
	.type = &setting_type_uint8,
};

struct setting chap_target_setting __setting = {
	.name = "chap-target",
	.description = "Mutual CHAP authentication",
	.tag = DHCP_EB_CHAP_TARGET,
	.type = &setting_type_uint8,
};

struct setting initiator_dhcp_setting __setting = {
	.name = "initiator-dhcp",
	.description = "Use DHCP for Initiator",
	.tag = DHCP_EB_INITIATOR_DHCP,
	.type = &setting_type_uint8,
};

struct setting target_dhcp_setting __setting = {
	.name = "target-dhcp",
	.description = "Use DHCP for Target",
	.tag = DHCP_EB_TARGET_DHCP,
	.type = &setting_type_uint8,
};

/** Vendor specific root path setting (encapsulated) */
struct setting  vendor_root_path_setting  __setting = {
	.name = "vendor-root-path",
	.description = "NFS/iSCSI vendor specific root path",
	.tag = DHCP_ISCSI_PRIMARY_TARGET_IQN,
	.type = &setting_type_string,
};

/**
 * This flag is a "suboption" of "Initiator DHCP" and have a meaning
 *  only if the main option is set to "Use DHCP". If the flag is not
 *  set, EEPROM value will be used. In any case, if IQN is a "null name",
 *  a value will be constructed using hostname, and if not available,
 *  a default IQN will be used.
 */
struct setting initiator_iqn_dhcp_setting __setting = {
	.name = "initiator-iqn-dhcp",
	.description = "Use DHCP for Initiator IQN",
	.tag = DHCP_EB_INITIATOR_IQN_DHCP,
	.type = &setting_type_uint8,
};

struct setting boot_image_setting __setting = {
	.name = "boot-image",
	.description = "Boot image",
	.tag = DHCP_EB_BOOT_IMAGE,
	.type = &setting_type_uint8,
};

struct setting server_name_setting __setting = {
	.name = "server-name",
	.description = "Target server",
	.tag = DHCP_RP_SERVERNAME,
	.type = &setting_type_string,
};

struct setting target_name_setting __setting = {
	.name = "target-name",
	.description = "Target IQN",
	.tag = DHCP_RP_TARGETNAME,
	.type = &setting_type_string,
};

struct setting msix_limit_setting __setting = {
	.name = "msix-limit",
	.description = "MSI-X interrupt limit",
	.tag = DHCP_EB_MSIX_LIMIT,
	.type = &setting_type_uint16_ro,
};

struct setting port_setting __setting = {
	.name = "port",
	.description = "Target port",
	.tag = DHCP_RP_PORT,
	.type = &setting_type_uint16,
};

struct setting lun_setting __setting = {
	.name = "lun",
	.description = "Target LUN",
	.tag = DHCP_RP_LUN,
	.type = &setting_type_lun,
};

struct setting sriov_setting __setting = {
	.name = "sriov",
	.description = "SR-IOV",
	.tag = DHCP_EB_SRIOV_ENABLED,
	.type = &setting_type_uint8_ro,
};

struct setting siena_vf_msix_limit_setting __setting = {
	.name = "vf-msix-limit",
	.description = "VF MSI-X interrupt limit",
	.tag = DHCP_EB_VF_MSIX_LIMIT,
	.type = &setting_type_uint8_ro,
};

struct setting hunt_vf_msix_limit_setting __setting = {
        .name = "vf-msix-limit",
        .description = "VF MSI-X interrupt limit",
        .tag = DHCP_EB_VF_MSIX_LIMIT,
	.type = &setting_type_uint16_ro,
};

struct setting pf_count_setting __setting = {
	.name = "pf-count",
	.description = "Number of Physical Functions on adapter",
	.tag = DHCP_EB_PF_COUNT,
	.type = &setting_type_int16_ro,
};

struct setting pfiov_setting __setting = {
	.name = "pfiov",
	.description = "Enable PFIOV mode",
	.tag = DHCP_EB_PFIOV_ENABLED,
	.type = &setting_type_uint8_ro,
};

struct setting vlan_setting __setting = {
	.name = "vlan_enabled",
	.description = "VPORT_VLAN enabled",
	.tag = DHCP_VLAN_ENABLED,
	.type = &setting_type_uint8_ro,
};

struct setting vlan_tag_setting __setting = {
	.name = "vlan_tag",
	.description = "VPORT VLAN tag",
	.tag = DHCP_EB_VLAN_TAG,
	.type = &setting_type_uint32_ro,
};

struct setting vf_count_setting __setting = {
	.name = "vf-count",
	.description = "Number of Virtual Functions",
	.tag = DHCP_EB_VF_COUNT,
	.type = &setting_type_uint8_ro,
};

struct setting firmware_variant_setting __setting = {
	.name = "firmware-variant",
	.description = "Firmware variant",
	.tag = DHCP_FW_VARIANT,
	.type = &setting_type_int32_ro,
};

/* This is read-write in single-PF mode. */
struct setting port_mode_setting __setting = {
	.name = "port-mode",
	.description = "Port mode",
	.tag = DHCP_EB_PORT_MODE,
	.type = &setting_type_int32_ro,
};

struct setting insecure_filters_setting __setting = {
	.name = "insecure-filters",
	.description = "Allow insecure filters",
	.tag = DHCP_INSECURE_FILTERS,
	.type = &setting_type_uint8_ro,
};

long
sf_sett_get_intz(struct settings *settings, struct setting *setting)
{
	long value;

	if (setting == NULL)
		return 0;

	if ( settings != NULL ) {
		if (fetch_int_setting(settings, setting, &value) >= 0)
			return value;
	}

	fetch_int_setting(sf_default_settings, setting, &value);
	return value;
};

/**
 * Fetch a string setting from the given settings node. If the string
 * setting doesn't exist, or is empty, then sf_default_settings is
 * searched instead.
 *
 * @v settings  	Settings object
 * @v setting		Setting to fetch
 * @v p_str  		Output pointer
 *
 * @ret      		Error Code or strlen of the resulting string.
 *			Guaranteed to never return rc = 0 (an empty
 *			string returns ENOENT)
 */
int
sf_sett_fetcha_string(struct settings *settings, struct setting *setting,
		      char **value)
{
	char buf[DHCP_OPT_LEN_MAX];
	int rc;

	if ((rc = sf_sett_fetch_string(settings, setting, buf,
				       sizeof(buf))) < 0)
		return rc;

	*value = strdup(buf);
	return rc;
}

/**
 * Fetch a string setting from the given settings node. if the string
 * setting doesn't exist, or is empty, then sf_default_settings is
 * searched instead.
 *
 * @v setts  		Settings object
 * @v setting		Setting to fetch
 * @v buf    		Storage for the result.
 * @v sz     		Size of the storage.
 * @ret rc   		Error Code or strlen of the resulting string.
 *			Guaranteed to never return rc = 0 (an empty
 *			string returns ENOENT)
 */
int
sf_sett_fetch_string(struct settings *settings, struct setting *setting,
                     char *buf, size_t sz)
{
	int rc;

	if ((rc = fetch_string_setting(settings, setting, buf, sz)) > 0)
		return rc;

	if ((rc = fetch_string_setting(sf_default_settings, setting,
				       buf, sz)) > 0)
		return rc;

	return -ENOENT;
}

/**
 * Fetch root-path for iSCSI boot.
 *
 * @v dev  		Network device
 * @v buf  		Storage for the result.
 * @v sz   		Size of the storage.
 * @ret rc 		Length of setting data, or negative error
 */
int
sf_root_path_fetch(struct net_device *dev, char *buf, size_t sz)
{
	struct settings *settings;
	char target_iqn[SF_TARGET_IQN_LEN_MAX + 1];
	char target_server[SF_TARGET_SERVER_LEN_MAX + 1];
	uint16_t port;
	uint8_t lun;
	int rc;

	memset (buf, 0, sz);
	settings = netdev_nvo_settings(dev);

	if (sf_sett_get_intz(settings, &target_dhcp_setting)) {
		settings = netdev_dhcp_settings(dev);
		if (settings == NULL) {
			DBG ("%s failed: %s has no dhcp settings\n",
			     __func__, dev->name);
			return -ENOENT;
		}
		/* Try Vendor encapsulated root-path first */
		rc = fetch_string_setting(settings,
		                          &vendor_root_path_setting,
		                          buf, sz);
		if (rc > 0)
			return rc;

		/* Then try option17 */
		return fetch_string_setting(settings,
					    &root_path_setting,
					    buf, sz);
	}

	/* From NVO, we only support constructing the iSCSI path out of
	 * component portions, or via defaults where appropriate */
	rc = sf_sett_fetch_string(settings, &server_name_setting,
				  target_server, sizeof(target_server));
	if (rc < 0)
		return rc;

	port = sf_sett_get_intz(settings, &port_setting);
	lun = sf_sett_get_intz(settings, &lun_setting);

	rc = sf_sett_fetch_string(settings, &target_name_setting,
				  target_iqn, sizeof(target_iqn));
	if (rc < 0)
		return rc;

	return snprintf(buf, sz, "iscsi:%s::%d:%x:%s",
			target_server, port, lun, target_iqn);
}

/**
 * Produce a key suitable for ordering mpio net_device's
 *
 * The key is 56:48=(priority) 48:0=mac.
 * No two devices should have the same key.
 *
 * This is the same algorithm as used in sfconfig, but
 * it's easier to reimplement it here, than share the
 * algorithm
 */
uint64_t __pure
sf_mpio_priority_key(struct net_device *net_dev)
{
	uint64_t 		priority = PRIORITY_KEY_NONE;
	struct settings	 	*settings;

	settings = netdev_nvo_settings(net_dev);
	if (settings == NULL)
		goto out;

	if (!sf_netdev_has_option_rom(net_dev))
		goto out;

	if (sf_sett_get_intz(settings, &boot_type_setting) != BOOTTYPE_ISCSI )
		goto out;

	/* Since sf_mpio_priority_key() gets run on other net_devices, ensure
	 * we ignore those not serving up legacy expansion rom images */
	if (sf_sett_get_intz(settings, &boot_image_setting) >= BOOTIMAGE_UEFI)
		goto out;

	priority = sf_sett_get_intz(settings, &mpio_priority_setting);
	if (priority > 0xff)
		priority = 0xff;
	priority <<= 48;

	memcpy(&priority, net_dev->ll_addr, 6);

out:
	return priority;
}
